/* ntdisp - NtD In System Programming.
 *
 * Copyright (C) 2012  Nicola Fontana <ntd at entidi.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */


#include "../config.h"
#include "../lib/ntdisp.h"
#include <stdlib.h>


typedef struct {
    gboolean     clear;
    gboolean     read;
    gboolean     write;
    gboolean     execute;
    GType        isp;
    gchar       *device;
    gchar       *from;
    gint         nbytes;
    gboolean     verbose;
    gboolean     initialize;
    gchar       *file;
} IspData;


static void
_isp_version(void)
{
    g_print("ntdisp " PACKAGE_VERSION "\n");
    exit(0);
}

static void
_isp_parse_args(gint *p_argc, gchar **p_argv[], IspData *data)
{
    GOptionContext *context;
    gchar *isp;
    GError *error;
    GOptionEntry entries[] = {
        {"command-clear", 'C', 0, G_OPTION_ARG_NONE,
         &data->clear, "Clear the flash memory", NULL},
        {"command-read", 'R', 0, G_OPTION_ARG_NONE,
         &data->read, "Read the program from the target", NULL},
        {"command-write", 'W', 0, G_OPTION_ARG_NONE,
         &data->write, "Write (download) the program to the target", NULL},
        {"command-execute", 'X', 0, G_OPTION_ARG_NONE,
         &data->execute, "Execute the program on the target", NULL},
        {"protocol", 'p', G_OPTION_FLAG_IN_MAIN, G_OPTION_ARG_STRING,
         &isp, "ISP protocol to use, mandatory", "PROTOCOL"},
        {"device", 'd', G_OPTION_FLAG_IN_MAIN, G_OPTION_ARG_STRING,
         &data->device, "Device to use, optional [/dev/ttyS0]", "DEVICE"},
        {"init", 'i', 0, G_OPTION_ARG_NONE,
         &data->initialize, "Perform the initialization step: autobaud, sync, ...", NULL},
        {"from", 'f', G_OPTION_FLAG_IN_MAIN, G_OPTION_ARG_STRING,
         &data->from, "Starting address, optional [flash begin address]", "ADDRESS"},
        {"nbytes", 'n', G_OPTION_FLAG_IN_MAIN, G_OPTION_ARG_INT,
         &data->nbytes, "Size, optional, used by -C and -R [flash size]", "BYTES"},
        {"verbose", 'v', 0, G_OPTION_ARG_NONE,
         &data->verbose, "Enable verbose output", NULL},
        {"version", 'V', G_OPTION_FLAG_NO_ARG, G_OPTION_ARG_CALLBACK,
         _isp_version, "Display version information", NULL},
        {NULL}
    };
    guint cmd_sum;
    guint8 return_code;

    context = g_option_context_new("COMMAND-OPTION PROTOCOL-OPTION [FILE]");
    isp = NULL;
    error = NULL;

    g_option_context_set_summary(context, "Multi-protocol utility for In System Programming support.");
    g_option_context_add_main_entries(context, entries, NULL);
    g_option_context_parse(context, p_argc, p_argv, &error);
    g_option_context_free(context);

    cmd_sum = data->clear + data->read + data->write + data->execute;
    return_code = 0;

    if (error != NULL) {
        return_code = error->code;
        g_printerr("%s\n", error->message);
        g_error_free(error);
    } else if (isp == NULL) {
        g_printerr("Mandatory protocol option (-p ...) not specified\n");
        return_code = 1;
    } else if (cmd_sum == FALSE + FALSE + FALSE + FALSE) {
        g_printerr("Mandatory command option (--command-...) not specified\n");
        return_code = 2;
    } else if (cmd_sum != TRUE + FALSE + FALSE + FALSE) {
        g_printerr("More than one command option (--command-...) has been used\n");
        return_code = 3;
    } else if (*p_argc > 1 && (data->clear || data->execute)) {
        g_printerr("Clear or execute does not need any input file\n");
        return_code = 4;
    } else if (*p_argc > 2) {
        g_printerr("Only one input file can be specified\n");
        return_code = 5;
    } else if (g_ascii_strcasecmp(isp, "stm") == 0) {
        data->isp = NTD_TYPE_STM;
    } else if (g_ascii_strcasecmp(isp, "lpc") == 0) {
        data->isp = NTD_TYPE_LPC;
    } else {
        g_printerr("Unknown protocol: `%s'\n", isp);
        return_code = 6;
    }

    g_free(isp);

    if (return_code != 0)
        exit(return_code);

    data->file = *p_argc == 2 ? g_strdup((*p_argv)[1]) : NULL;
    if (data->device == NULL)
        data->device = g_strdup("/dev/ttyS0");
}

static void
_isp_process(const IspData *data)
{
    NtdIsp *isp;
    NtdDevice *device;
    GError *error;

    isp = g_object_new(data->isp, NULL);
    ntd_isp_set_device(isp, data->device);
    device = (NtdDevice *) ntd_isp_get_serial(isp);
    g_return_if_fail(NTD_IS_SERIAL(device));
    error = NULL;

    ntd_set_verbose_mode(data->verbose);

    if (data->initialize) {
        /* The explicit connection is needed by ntd_device_flush() */
        ntd_isp_connect(isp);
    } else {
        ntd_device_open(device);
        /* Forcibly set NtdIsp::is-connected to skip the initialization */
        ntd_isp_set_connected(isp, TRUE);
        if (! ntd_device_is_opened(device)) {
            error = ntd_device_get_error(device);
            g_printerr("Device opening failed: %s\n", error->message);
            exit(10);
        }
    }

    /* Clearing any previous data in the queues, just in case */
    ntd_device_flush(device, TRUE, TRUE);

    /* Probing the device is common to all commands */
    if (! ntd_isp_probe(isp)) {
        g_printerr("Probe failed: unknown target?\n");
        exit(11);
    }

    if (data->clear) {
        if (! ntd_isp_wipe(isp, data->from, data->nbytes)) {
            g_printerr("Wiping failed\n");
            exit(12);
        }
    } else if (data->read) {
        if (! ntd_isp_dump_file(isp, data->from, data->nbytes, data->file, &error)) {
            g_printerr("ntd_isp_dump() failed: %s\n", error->message);
            exit(13);
        }
    } else if (data->write) {
        if (! ntd_isp_download_file(isp, data->from, data->file, &error)) {
            g_printerr("ntd_isp_download() failed: %s\n", error->message);
            exit(14);
        }
    } else if (data->execute) {
        if (! ntd_isp_run(isp, data->from)) {
            g_printerr("ntd_isp_run() failed\n");
            exit(15);
        }
    }

    g_object_unref(isp);
}


int
main(gint argc, gchar **argv)
{
    IspData data = { 0 };

#if GLIB_CHECK_VERSION(2, 36, 0)
    /* g_type_init() is deprecated since glib-2.36.0:
     * the type system is automatically initialized */
#else
    g_type_init();
#endif

    _isp_parse_args(&argc, &argv, &data);
    _isp_process(&data);

    /* Destroy dynamic data */
    g_free(data.device);
    g_free(data.file);
    g_free(data.from);

    return 0;
}
