/* ntdisp - NtD In System Programming.
 *
 * Copyright (C) 2012  Nicola Fontana <ntd at entidi.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * SECTION:ntd-regex-entry
 * @short_description: A #GtkEntry that automatically validates its content
 *                     with a regular expression.
 *
 * A simple entry with regular expressions validation similar to HRegExEntry
 * (http://code.google.com/p/holly-gtk-widgets/wiki/HRegExEntry) but written
 * in C. Has a small icon that changes if the content is valid or invalid.
 **/


/**
 * NtdRegexEntry:
 *
 * All fields are private and should not be used directly.
 * Use its public methods instead.
 **/


#include <gtk/gtk.h>

#include "ntd-regex-entry.h"
#include "ntd-regex-entry-private.h"

/* Fallbacks when no i18n support is provided */
#ifndef P_
#define P_(s)  (s)
#endif


#define _NTD_OLD_OBJECT_CLASS   ((GObjectClass *) ntd_regex_entry_parent_class)


enum {
    PROP_0,
    PROP_REGEX
};

enum {
    VALIDITY_CHANGED,
    LAST_SIGNAL
};


static void     _ntd_editable_init          (GtkEditableInterface *iface);
static void     _ntd_regex_free_and_nullify (GRegex        **p_regex);
static void     _ntd_dispose                (GObject        *object);
static void     _ntd_get_property           (GObject        *object,
                                             guint           prop_id,
                                             GValue         *value,
                                             GParamSpec     *pspec);
static void     _ntd_set_property           (GObject        *object,
                                             guint           prop_id,
                                             const GValue   *value,
                                             GParamSpec     *pspec);
static void     _ntd_editable_changed       (GtkEditable    *editable);
static const gchar *
                _ntd_get_regex              (NtdRegexEntry  *entry);
static void     _ntd_change_regex           (NtdRegexEntry  *entry,
                                             const gchar    *regex);
static void     _ntd_validate               (NtdRegexEntry  *entry);
static void     _ntd_set_validation         (NtdRegexEntry  *entry,
                                             gboolean        is_valid,
                                             gboolean        is_valided);


static guint    _ntd_signals[LAST_SIGNAL] = { 0 };
static GtkEditableInterface *
                _ntd_parent_editable_iface = NULL;


G_DEFINE_TYPE_WITH_CODE(NtdRegexEntry, ntd_regex_entry, GTK_TYPE_ENTRY,
                        G_IMPLEMENT_INTERFACE(GTK_TYPE_EDITABLE, _ntd_editable_init))


static void
ntd_regex_entry_class_init(NtdRegexEntryClass *klass)
{
    GObjectClass *gobject_class;
    GParamSpec *param;

    gobject_class = (GObjectClass *) klass;

    g_type_class_add_private(klass, sizeof(NtdRegexEntryPrivate));

    gobject_class->dispose = _ntd_dispose;
    gobject_class->get_property = _ntd_get_property;
    gobject_class->set_property = _ntd_set_property;

    klass->validity_changed = NULL;

    param = g_param_spec_string("regex",
                                P_("Regular Expression"),
                                P_("The Perl-compatible regular expressions to use for validation"),
                                NULL,
                                G_PARAM_READWRITE | G_PARAM_CONSTRUCT);
    g_object_class_install_property(gobject_class, PROP_REGEX, param);

    /**
     * NtdRegexEntry::validity-changed:
     * @entry: an #NtdRegexEntry
     *
     * Emitted when a valid @entry becomes invalid or when an invalid
     * @entry becomes valid.
     **/
    _ntd_signals[VALIDITY_CHANGED] =
        g_signal_new("validity-changed", NTD_TYPE_REGEX_ENTRY,
                     G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE,
                     G_STRUCT_OFFSET(NtdRegexEntryClass, validity_changed),
                     NULL, NULL,
                     g_cclosure_marshal_VOID__VOID,
                     G_TYPE_NONE, 0);
}

static void
ntd_regex_entry_init(NtdRegexEntry *entry)
{
    NtdRegexEntryPrivate *data = G_TYPE_INSTANCE_GET_PRIVATE(entry,
                                                             NTD_TYPE_REGEX_ENTRY,
                                                             NtdRegexEntryPrivate);

    data->regex = NULL;
    data->is_validated = FALSE;
    data->is_valid = FALSE;

    entry->data = data;
}

static void
_ntd_editable_init(GtkEditableInterface *iface)
{
    _ntd_parent_editable_iface = g_type_interface_peek_parent(iface);
    iface->changed = _ntd_editable_changed;
}

static void
_ntd_regex_free_and_nullify(GRegex **p_regex)
{
    if (*p_regex) {
        g_regex_unref(*p_regex);
        *p_regex = NULL;
    }
}

static void
_ntd_dispose(GObject *object)
{
    NtdRegexEntryPrivate *data = ((NtdRegexEntry *) object)->data;
    GObjectClass *parent_class = (GObjectClass *) ntd_regex_entry_parent_class;

    _ntd_regex_free_and_nullify(&data->regex);

    if (parent_class->dispose)
        parent_class->dispose(object);
}

static void
_ntd_get_property(GObject *object, guint prop_id,
                  GValue *value, GParamSpec *pspec)
{
    NtdRegexEntry *entry = (NtdRegexEntry *) object;

    switch (prop_id) {
    case PROP_REGEX:
        g_value_set_string(value, _ntd_get_regex(entry));
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
_ntd_set_property(GObject *object, guint prop_id,
                  const GValue *value, GParamSpec *pspec)
{
    NtdRegexEntry *entry = (NtdRegexEntry *) object;

    switch (prop_id) {
    case PROP_REGEX:
        _ntd_change_regex(entry, g_value_get_string(value));
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
_ntd_editable_changed(GtkEditable *editable)
{
    /* Chain up the parent method, if any */
    if (_ntd_parent_editable_iface->changed)
        _ntd_parent_editable_iface->changed(editable);

    _ntd_validate((NtdRegexEntry *) editable);
}


/**
 * ntd_regex_entry_new:
 *
 * Creates a new empty #NtdRegexEntry.
 *
 * Returns: the newly created widget
 **/
GtkWidget *
ntd_regex_entry_new(void)
{
    return g_object_new(NTD_TYPE_REGEX_ENTRY, NULL);
}

/**
 * ntd_regex_entry_new_with_regex:
 *
 * Creates a new #NtdRegexEntry widget with a specific regular expression
 * to be used as validator. This is conceptually equivalent to call
 * ntd_regex_entry_new() followed by ntd_regex_entry_set_regex().
 *
 * Returns: the newly created widget
 **/
GtkWidget *
ntd_regex_entry_new_with_regex(const gchar *regex)
{
    return g_object_new(NTD_TYPE_REGEX_ENTRY, "regex", regex, NULL);
}

/**
 * ntd_regex_entry_set_regex:
 * @entry: an #NtdRegexEntry
 * @regex: a Perl-compatible regular expression
 *
 * Sets the regular expression of @entry to @regex.
 * After this operation the validation is in an
 * unknown state: call ntd_regex_entry_validate()
 * to force the validation.
 *
 * %NULL is a valid @regex value, meaning the
 * regular expression is unset.
 **/
void
ntd_regex_entry_set_regex(NtdRegexEntry *entry, const gchar *regex)
{
    g_return_if_fail(NTD_IS_REGEX_ENTRY(entry));
    if (g_strcmp0(regex, _ntd_get_regex(entry)) != 0) {
        _ntd_change_regex(entry, regex);
        g_object_notify(G_OBJECT(entry), "regex");
    }
}

/**
 * ntd_regex_entry_get_regex:
 * @entry: an #NtdRegexEntry
 *
 * Returns the current regular expression bound to
 * @entry to be used for validation. The returned
 * string is owned by @entry and should not be
 * modified or freed.
 *
 * Returns: the regular expression or
 *          %NULL if it is not set.
 **/
const gchar *
ntd_regex_entry_get_regex(NtdRegexEntry *entry)
{
    g_return_val_if_fail(NTD_IS_REGEX_ENTRY(entry), NULL);
    return _ntd_get_regex(entry);
}

/**
 * ntd_regex_entry_validate:
 * @entry: an #NtdRegexEntry
 *
 * Performs the validation of the current text
 * against the regular expression bound to @entry.
 *
 * There are few reasons to use this function because
 * the validation is automatically triggered every
 * time the entry text has changed.
 **/
void
ntd_regex_entry_validate(NtdRegexEntry *entry)
{
    g_return_if_fail(NTD_IS_REGEX_ENTRY(entry));
    _ntd_validate(entry);
}

/**
 * ntd_regex_entry_set_validity:
 * @entry: an #NtdRegexEntry
 * @is_valid: the new value of the validity flag
 *
 * This function is only for #NtdRegexEntry implementations.
 *
 * Sets the validify flag of @entry to @is_valid.
 * If the previous validity state was different from
 * @is_valid, a #NtdRegexEntry::validity-changed
 * signal is emitted.
 **/
void
ntd_regex_entry_set_validity(NtdRegexEntry *entry, gboolean is_valid)
{
    g_return_if_fail(NTD_IS_REGEX_ENTRY(entry));
    _ntd_set_validation(entry, is_valid, TRUE);
}

/**
 * ntd_regex_entry_get_validity:
 * @entry: an #NtdRegexEntry
 * @fallback: what to return if @entry has never been
 *            validated.
 *
 * Gets the validation status of @entry. If @entry has
 * never been validated (here including a previous change
 * of regex with ntd_regex_entry_set_regex()), @fallback
 * is returned.
 *
 * Returns: the validation flag, or @fallback on no
 *          validation performed.
 **/
gboolean
ntd_regex_entry_get_validity(NtdRegexEntry *entry, gboolean fallback)
{
    NtdRegexEntryPrivate *data;

    g_return_val_if_fail(NTD_IS_REGEX_ENTRY(entry), FALSE);

    data = entry->data;
    return data->is_validated ? data->is_valid : fallback;
}

/**
 * ntd_regex_entry_validity_changed:
 * @entry: an #NtdRegexEntry
 *
 * This function is only for #NtdRegexEntry implementations.
 *
 * Emits the #NtdRegexEntry::validity-changed signal on @entry.
 **/
void
ntd_regex_entry_validity_changed(NtdRegexEntry *entry)
{
    g_return_if_fail(NTD_IS_REGEX_ENTRY(entry));
    g_signal_emit(entry, _ntd_signals[VALIDITY_CHANGED], 0);
}


static const gchar *
_ntd_get_regex(NtdRegexEntry *entry)
{
    NtdRegexEntryPrivate *data = entry->data;
    return data->regex ? g_regex_get_pattern(data->regex) : NULL;
}

static void
_ntd_change_regex(NtdRegexEntry *entry, const gchar *regex)
{
    NtdRegexEntryPrivate *data = entry->data;

    _ntd_regex_free_and_nullify(&data->regex);

    if (regex) {
        GError *error = NULL;

        data->regex = g_regex_new(regex, G_REGEX_OPTIMIZE, 0, &error);

        if (error) {
            g_error("%s", error->message);
            _ntd_regex_free_and_nullify(&data->regex);
            g_free(error);
        } else {
            _ntd_set_validation(entry, data->is_valid, FALSE);
        }
    }
}

static void
_ntd_validate(NtdRegexEntry *entry)
{
    NtdRegexEntryPrivate *data = entry->data;
    const gchar *string;
    gboolean is_valid;

    /* No regular expression set: don't do anything */
    if (data->regex == NULL)
        return;

    string = gtk_entry_get_text((GtkEntry *) entry);
    is_valid = g_regex_match(data->regex, string, 0, NULL);

    _ntd_set_validation(entry, is_valid, TRUE);
}

static void
_ntd_set_validation(NtdRegexEntry *entry,
                    gboolean is_valid, gboolean is_validated)
{
    NtdRegexEntryPrivate *data = entry->data;
    const gchar *icon_name;

    if (data->is_valid == is_valid && data->is_validated == is_validated)
        return;

    data->is_valid = is_valid;
    data->is_validated = is_validated;

    icon_name = is_validated ? (is_valid ? "gtk-yes" : "gtk-no") : NULL;
    gtk_entry_set_icon_from_icon_name((GtkEntry *) entry,
                                      GTK_ENTRY_ICON_SECONDARY,
                                      icon_name);

    g_signal_emit(entry, _ntd_signals[VALIDITY_CHANGED], 0);
}
