/* ntdisp - NtD In System Programming.
 *
 * Copyright (C) 2012  Nicola Fontana <ntd at entidi.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */


#include <gtk/gtk.h>
#include <stdlib.h>
#include <string.h>

#include "../config.h"
#include "../lib/ntdisp.h"
#include "ntd-regex-entry.h"


/* A wrapper around g_settings_set_string() that handles NULL gracefully,
 * that is consider them as empty strings */
#define ntd_settings_set_string(s,k,v) \
    g_settings_set_string((s), (k), (v) == NULL ? "" : (v));

/* Actually the protocol list is hardcoded. See issue #83 for details:
 * http://dev.entidi.com/p/ntdisp/issues/83/
 */
static gpointer _ntd_protocols[][2] = {
    { "",       "" },
    { "NtdLpc", "NXP LPC (In System Programming)" },
    { "NtdStm", "STM32 (USART protocol)" },
    { NULL,     NULL }
};

/* Whether the program has been installed or not before being run.
 * This flag is set by _ntd_init().
 */
static gboolean is_installed = FALSE;

/* The base directory where all files must be referred to
 * (usually left unset by _ntd_init() on POSIX systems).
 */
static gchar *basedir = NULL;


static void
_ntd_init(gint argc, gchar **argv)
{
    /* There are different ways for checking at run-time if this program is
     * run installed or uninstalled. The GNULib relocation module (the
     * relocation de-facto standard) checks the directory the program has
     * been called from by analyzing argv[0]. Here instead we are relying on
     * the fact the uninstalled program has "uninstalled" embedded into its
     * name: a more robust (and simple) approach that only requires a bit of
     * autofoo tweaking.
     */
    gchar *name;

    /* Check for valid arguments */
    if (argc < 1 || argv[0] == NULL)
        return;

    basedir = g_path_get_dirname(argv[0]);

    name = g_path_get_basename(argv[0]);
    if (name == NULL)
        return;

    is_installed = strstr(name, "uninstalled") == NULL;
    g_free(name);
}

static gchar *
_ntd_file(const gchar *file_name)
{
    static gchar *pkg_data_dir = NULL;

    if (! is_installed) {
        /* Running uninstalled: look up the file in BUILDDIR before and
         * in SRCDIR after, returning the first match */
        return ntd_find_file(file_name, BUILDDIR, SRCDIR, NULL);
    }

    /* Otherwise, look up the file only in PKGDATADIR */

    if (pkg_data_dir == NULL) {
#ifdef G_OS_WIN32
        pkg_data_dir = g_build_filename(basedir != NULL ? basedir : "",
                                        PKGDATADIR, NULL);
#else
        pkg_data_dir = g_strdup(PKGDATADIR);
#endif
    }

    return ntd_find_file(file_name, pkg_data_dir, NULL);
}

static GdkPixbuf *
_ntd_pixbuf(const gchar *file_name)
{
    gchar *file = _ntd_file(file_name);

    if (file == NULL)
        return NULL;

    return gdk_pixbuf_new_from_file(file, NULL);
}

static void
_ntd_augment_icon_theme(void)
{
    const gchar *icon_names[] = { "ntdisp-read", "ntdisp-write" };
    const gchar *icon_name;
    gchar *file_name;
    gint n, size;
    GdkPixbuf *pixbuf;

    for (n = 0; n < G_N_ELEMENTS(icon_names); ++n) {
        icon_name = icon_names[n];
        file_name = g_strdup_printf("%s.png", icon_name);
        pixbuf = _ntd_pixbuf(file_name);
        g_free(file_name);

        if (pixbuf != NULL) {
            size = gdk_pixbuf_get_height(pixbuf);
            gtk_icon_theme_add_builtin_icon(icon_name, size, pixbuf);
            g_object_unref(pixbuf);
        }
    }
}

static void
_ntd_use_default_icons(void)
{
    const gchar *file_names[] = {
        "ntdisp-16.png",
        "ntdisp-32.png",
        "ntdisp-48.png",
        "ntdisp-64.png",
        "ntdisp-128.png"
    };
    gint n;
    GdkPixbuf *pixbuf;
    GList *list = NULL;

    for (n = 0; n < G_N_ELEMENTS(file_names); ++n) {
        pixbuf = _ntd_pixbuf(file_names[n]);
        if (pixbuf != NULL)
            list = g_list_append(list, pixbuf);
    }

    if (list != NULL) {
        gtk_window_set_default_icon_list(list);
        g_list_foreach(list, (GFunc) g_object_unref, NULL);
        g_list_free(list);
    }
}

static void
_ntd_refresh_ui(void)
{
    while (gtk_events_pending())
          gtk_main_iteration();
}

static gboolean
_ntd_tree_view_set_action(GtkTreeView *tree_view, const gint action)
{
    GtkTreeSelection *selection;
    GtkTreeModel *model;
    GtkTreeIter iter;
    gint iter_action;

    selection = gtk_tree_view_get_selection(tree_view);
    model = gtk_tree_view_get_model(tree_view);

    if (! gtk_tree_model_get_iter_first(model, &iter))
        return FALSE;

    do {
        gtk_tree_model_get(model, &iter, 2, &iter_action, -1);
        if (iter_action == action) {
            gtk_tree_selection_select_iter(selection, &iter);
            return TRUE;
        }
    } while (gtk_tree_model_iter_next(model, &iter));

    return FALSE;
}

static gboolean
_ntd_tree_view_get_action(GtkTreeView *tree_view, gint *action)
{
    GtkTreeSelection *selection;
    GtkTreeModel *model;
    GtkTreeIter iter;

    selection = gtk_tree_view_get_selection(tree_view);

    if (! gtk_tree_selection_get_selected(selection, &model, &iter))
        return FALSE;

    gtk_tree_model_get(model, &iter, 2, action, -1);
    return TRUE;
}

static void
_ntd_update_progression(GtkProgressBar *progress_bar, gdouble fraction)
{
    gtk_progress_bar_set_fraction(progress_bar, fraction);
    _ntd_refresh_ui();
}

static gboolean
_ntd_quit(void)
{
    gtk_main_quit();
    return TRUE;
}

static void
_ntd_abort(const gchar *format, ...)
{
    va_list var_args;
    gchar *message;
    GtkWidget *dialog;

    va_start(var_args, format);
    message = g_strdup_vprintf(format, var_args);
    va_end(var_args);

    dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL,
                                    GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE,
                                    "%s", message);
    g_free(message);

    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
    gtk_window_set_title(GTK_WINDOW(dialog), "Error");

    /* gtk_dialog_run() stinks of deprecation, so using signals instead */
    g_signal_connect(dialog, "response", G_CALLBACK(gtk_main_quit), NULL);
    g_signal_connect(dialog, "delete-event", G_CALLBACK(_ntd_quit), NULL);

    gtk_widget_show_all(dialog);
    gtk_main();
    gtk_widget_destroy(dialog);

    exit(1);
}


static guint
_ntd_input_to_uint(gboolean flag, const gchar *input)
{
    guint value = 0;

    if (input == NULL) {
    } else if (strspn(input, "0123456789") == strlen(input)) {
        /* Only 0-9 digits present: this is a decimal number */
        sscanf(input, "%u", &value);
    } else {
        /* In all other cases, this is an hexadecimal number */
        const gchar *string;

        /* Skip the optional prefix (0x, $ or &) */
        if (input[0] == '$' || input[0] == '&')
            string = input + 1;
        else if (input[0] == '0' && input[1] == 'x')
            string = input + 2;
        else
            string = input;

        /* Perform the conversion only on valid hexadecimal values */
        if (strspn(string, "0123456789ABCDEFabcdef") == strlen(string))
            sscanf(string, "%x", &value);
    }

    return value;
}

static gchar *
_ntd_input_to_address(gboolean flag, const gchar *input)
{
    return g_strdup_printf("%u", _ntd_input_to_uint(flag, input));
}

typedef enum {
    _NTD_ACTION_NULL = -1,
    _NTD_ACTION_WRITE = 0,
    _NTD_ACTION_READ = 1
} _NtdAction;

typedef struct {
    gchar       *device;
    gint         protocol;
    gboolean     init;
    _NtdAction   action;

    /* READ action */
    gchar       *save_to;
    gboolean     read_from;
    gchar       *read_from_address;
    gboolean     read_size;
    gchar       *read_size_bytes;

    /* WRITE action */
    gchar       *load_from;
    gboolean     erase;
    gboolean     write_to;
    gchar       *write_to_address;
    gboolean     verify;
    gboolean     execute;
} _NtdConfig;


static gboolean
_ntd_settings_schema_exists(const gchar *schema_id)
{
    const gchar * const *schemas;

    schemas = g_settings_list_schemas();
    if (schemas != NULL) {
        const gchar **p_schema;
        for (p_schema = (const gchar **) schemas; *p_schema != NULL; ++p_schema) {
            if (g_strcmp0(schema_id, *p_schema) == 0)
                return TRUE;
        }
    }

    return FALSE;
}

#if GLIB_CHECK_VERSION(2, 32, 0)
static GSettings *
_ntd_settings_new_from(const gchar *schema_id, const gchar *path)
{
    GSettingsSchemaSource *schema_source;
    GSettingsSchema *schema;
    GError *error;

    error = NULL;
    schema_source = g_settings_schema_source_new_from_directory(path, NULL, TRUE, &error);
    if (schema_source == NULL) {
        _ntd_abort("Error while loading the schema source from '%s': %s",
                   path, error->message);
        return NULL;
    }

    schema = g_settings_schema_source_lookup(schema_source, schema_id, FALSE);
    g_settings_schema_source_unref(schema_source);
    if (schema == NULL)
        return NULL;

    return g_settings_new_full(schema, NULL, NULL);
}
#endif

static GSettings *
_ntd_settings(void)
{
    /* If the schema exists: simply return it */
    if (_ntd_settings_schema_exists(NTDISP_SCHEMA)) {
        return g_settings_new(NTDISP_SCHEMA);
    }

#if GLIB_CHECK_VERSION(2, 32, 0)
    /* The schema does not exist but GLib is 2.32.0 or newer is available:
     * try to load it from $builddir if this is an uninstalled run */
    if (! is_installed) {
        GSettings *settings = _ntd_settings_new_from(NTDISP_SCHEMA, BUILDDIR);
        if (settings != NULL)
            return settings;
    }
#endif

    /* No ways to load the schema: raise an error and abort */
    _ntd_abort("Schema '%s' not found", NTDISP_SCHEMA);
    return NULL;
}

static void
_ntd_config_read(_NtdConfig *config)
{
    GSettings *settings = _ntd_settings();

    config->device = g_settings_get_string(settings, "device");
    config->protocol = g_settings_get_enum(settings, "protocol");
    config->init = g_settings_get_boolean(settings, "init");
    config->action = g_settings_get_enum(settings, "action");

    config->save_to = g_settings_get_string(settings, "save-to");
    config->read_from = g_settings_get_boolean(settings, "read-from");
    config->read_from_address = g_settings_get_string(settings, "read-from-address");
    config->read_size = g_settings_get_boolean(settings, "read-size");
    config->read_size_bytes = g_settings_get_string(settings, "read-size-bytes");

    config->load_from = g_settings_get_string(settings, "load-from");
    config->erase = g_settings_get_boolean(settings, "erase");
    config->write_to = g_settings_get_boolean(settings, "write-to");
    config->write_to_address = g_settings_get_string(settings, "write-to-address");
    config->verify = g_settings_get_boolean(settings, "verify");
    config->execute = g_settings_get_boolean(settings, "execute");

    g_object_unref(settings);
}

static void
_ntd_config_write(const _NtdConfig *config)
{
    GSettings *settings = _ntd_settings();

    ntd_settings_set_string(settings, "device", config->device);
    g_settings_set_enum(settings, "protocol", config->protocol);
    g_settings_set_boolean(settings, "init", config->init);
    g_settings_set_enum(settings, "action", config->action);

    ntd_settings_set_string(settings, "save-to", config->save_to);
    g_settings_set_boolean(settings, "read-from", config->read_from);
    ntd_settings_set_string(settings, "read-from-address", config->read_from_address);
    g_settings_set_boolean(settings, "read-size", config->read_size);
    ntd_settings_set_string(settings, "read-size-bytes", config->read_size_bytes);

    ntd_settings_set_string(settings, "load-from", config->load_from);
    g_settings_set_boolean(settings, "erase", config->erase);
    g_settings_set_boolean(settings, "write-to", config->write_to);
    ntd_settings_set_string(settings, "write-to-address", config->write_to_address);
    g_settings_set_boolean(settings, "verify", config->verify);
    g_settings_set_boolean(settings, "execute", config->execute);

    g_settings_sync();
    g_object_unref(settings);
}

static NtdIsp *
_ntd_config_create_isp(const _NtdConfig *config)
{
    GType isp_type;
    NtdIsp *isp;
    NtdDevice *device;

    isp_type = g_type_from_name(_ntd_protocols[config->protocol][0]);
    isp = g_object_new(isp_type, NULL);

    ntd_isp_set_device(isp, config->device);
    device = (NtdDevice *) ntd_isp_get_serial(isp);
    if (! NTD_IS_SERIAL(device)) {
        g_object_unref(isp);
        return NULL;
    }

    return isp;
}

static gboolean
_ntd_config_probe(const _NtdConfig *config, NtdIsp *isp)
{
    NtdDevice *device = (NtdDevice *) ntd_isp_get_serial(isp);

    if (config->init) {
        g_message("Performing the initialization step");
        ntd_isp_connect(isp);
    } else {
        g_message("Opening the connection");
        ntd_device_open(device);
        /* Forcibly set NtdIsp::is-connected to skip the initialization */
        ntd_isp_set_connected(isp, TRUE);
    }

    if (ntd_isp_get_status(isp) != 0) {
        g_message("Timeout: is the target device connected and in ISP mode?");
        return FALSE;
    }

    /* Clearing any previous data in the queues, just in case */
    ntd_device_flush(device, TRUE, TRUE);

    g_message("Probing the connected device");
    if (! ntd_isp_probe(isp)) {
        g_message("Probe failed: target unknown");
        return FALSE;
    }

    return TRUE;
}

static gboolean
_ntd_config_execute(const _NtdConfig *config, NtdIsp *isp)
{
    GError *error = NULL;
    gchar *address;
    gsize size;
    gboolean done;

    switch (config->action) {
    case _NTD_ACTION_READ:
        address = _ntd_input_to_address(config->read_from, config->read_from_address);
        size = _ntd_input_to_uint(config->read_size, config->read_size_bytes);

        g_message("Reading in progress");
        done = ntd_isp_dump_file(isp, address, size, config->save_to, &error);
        if (! done)
            g_message("Read failed: %s", error->message);

        break;

    case _NTD_ACTION_WRITE:
        address = _ntd_input_to_address(config->write_to, config->write_to_address);
        done = TRUE;

        if (done && config->erase) {
            g_message("Wiping in progress");
            done = ntd_isp_wipe(isp, address, 0);
            if (! done)
                g_message("Wiping failed");
        }

        if (done) {
            g_message("Writing in progress");
            done = ntd_isp_download_file(isp, address, config->load_from, &error);
            if (! done)
                g_message("Write failed: %s", error->message);
        }

        if (done && config->verify) {
            g_message("Verification in progress");
            done = ntd_isp_verify_file(isp, address, config->load_from, &error);
            if (! done)
                g_message("Verification failed: %s", error->message);
        }

        if (done && config->execute) {
            g_message("Executing the program");
            done = ntd_isp_run(isp, address);
            /* NB: execution could not return also if performed properly.
             *     No idea on how to handle this. */
            if (! done)
                g_message("Execution failed");
        }

        break;

    default:
        return FALSE;
    }

    if (error != NULL)
        g_error_free(error);
    if (address != NULL)
        g_free(address);

    return done;
}

static void
_ntd_config_free(_NtdConfig *config)
{
    ntd_free_and_nullify((gpointer *) &config->device);
    ntd_free_and_nullify((gpointer *) &config->save_to);
    ntd_free_and_nullify((gpointer *) &config->read_from_address);
    ntd_free_and_nullify((gpointer *) &config->read_size_bytes);
    ntd_free_and_nullify((gpointer *) &config->load_from);
    ntd_free_and_nullify((gpointer *) &config->write_to_address);
}


static void
_ntd_version(void)
{
    g_print("ntdisp-gtk3 " PACKAGE_VERSION "\n");
    exit(0);
}

static void
_ntd_parse_args(gint *p_argc, gchar **p_argv[])
{
    GError *error = NULL;
    GOptionEntry entries[] = {
        {"version", 'V', G_OPTION_FLAG_NO_ARG, G_OPTION_ARG_CALLBACK,
         _ntd_version, "Display version information", NULL},
        {NULL}
    };

    gtk_init_with_args(p_argc, p_argv, "- ntdisp GTK+3 frontend",
                       entries, NULL, &error);

    if (error != NULL) {
        g_error("%s", error->message);
        g_error_free(error);
    }
}


typedef struct {
    _NtdConfig   *config;
    GtkBuilder   *builder;
    GtkWidget    *wnd_main;
    GtkWidget    *txt_device;
    GtkWidget    *lst_protocol;
    GtkWidget    *chk_init;
    GtkWidget    *tv_action;
    GtkWidget    *nb_action;
    GtkWidget    *dlg_save_to;
    GtkWidget    *txt_save_to;
    GtkWidget    *btn_save_to;
    GtkWidget    *chk_read_from;
    GtkWidget    *txt_read_from;
    GtkWidget    *chk_read_size;
    GtkWidget    *txt_read_size;
    GtkWidget    *fc_load_from;
    GtkWidget    *chk_erase;
    GtkWidget    *chk_write_to;
    GtkWidget    *txt_write_to;
    GtkWidget    *chk_verify;
    GtkWidget    *chk_execute;
    GtkWidget    *spn_progress;
    GtkWidget    *pb_progress;
    GtkWidget    *lbl_progress;
    GtkWidget    *btn_run;
    GtkWidget    *btn_stop;
    GtkWidget    *btn_close;
} _NtdUI;


static void
_ntd_ui_nullify(_NtdUI *ui)
{
    /* Set everything to NULL */
    ui->config = NULL;
    ui->builder = NULL;
    ui->wnd_main = NULL;
    ui->txt_device = NULL;
    ui->lst_protocol = NULL;
    ui->chk_init = NULL;
    ui->tv_action = NULL;
    ui->nb_action = NULL;
    ui->dlg_save_to = NULL;
    ui->txt_save_to = NULL;
    ui->btn_save_to = NULL;
    ui->chk_read_from = NULL;
    ui->txt_read_from = NULL;
    ui->chk_read_size = NULL;
    ui->txt_read_size = NULL;
    ui->fc_load_from = NULL;
    ui->chk_erase = NULL;
    ui->chk_write_to = NULL;
    ui->txt_write_to = NULL;
    ui->chk_verify = NULL;
    ui->chk_execute = NULL;
    ui->spn_progress = NULL;
    ui->pb_progress = NULL;
    ui->lbl_progress = NULL;
    ui->btn_run = NULL;
    ui->btn_stop = NULL;
    ui->btn_close = NULL;
}

static void
_ntd_ui_init(_NtdUI *ui)
{
    /* Fallback to NULL values */
    _ntd_ui_nullify(ui);

    /* Used to forcefully register custom GTK+ widgets */
    g_type_name(NTD_TYPE_REGEX_ENTRY);
    g_type_name(NTD_TYPE_LPC);
    g_type_name(NTD_TYPE_STM);

    _ntd_augment_icon_theme();
    _ntd_use_default_icons();
}

static void
_ntd_ui_free_builder(_NtdUI *ui)
{
    if (ui->builder != NULL) {
        g_object_unref(ui->builder);
        ui->builder = NULL;
    }
}

static void
_ntd_ui_free(_NtdUI *ui)
{
    _ntd_ui_free_builder(ui);

    /* Destroying the top levels should free all the underlying widgets */
    if (ui->wnd_main != NULL)
        gtk_widget_destroy(ui->wnd_main);
    if (ui->dlg_save_to != NULL)
        gtk_widget_destroy(ui->dlg_save_to);

    _ntd_ui_nullify(ui);
}

static gboolean
_ntd_ui_set_builder(_NtdUI *ui)
{
    GError *error = NULL;
    gchar *path = _ntd_file("ntdisp.ui");

    if (path == NULL) {
        g_critical("File 'ntdisp.ui' not found");
        return FALSE;
    }

    ui->builder = gtk_builder_new();
    gtk_builder_add_from_file(ui->builder, path, &error);
    g_free(path);

    if (error != NULL) {
        g_critical("%s", error->message);
        g_error_free(error);
        return FALSE;
    }

    return TRUE;
}

/* Helper macro for _ntd_ui_set_objects() */
#define _ntd_ui_set_object(field,name,type) \
    ui-> field = (gpointer) gtk_builder_get_object(ui->builder, name); \
    g_return_val_if_fail(G_TYPE_CHECK_INSTANCE_TYPE(ui-> field, type), FALSE)

static gboolean
_ntd_ui_set_objects(_NtdUI *ui)
{
    _ntd_ui_set_object(wnd_main, "wndMain",  GTK_TYPE_WIDGET);
    _ntd_ui_set_object(txt_device, "txtDevice", GTK_TYPE_WIDGET);
    _ntd_ui_set_object(lst_protocol, "lstProtocol", GTK_TYPE_COMBO_BOX_TEXT);
    _ntd_ui_set_object(chk_init, "chkInit", GTK_TYPE_CHECK_BUTTON);
    _ntd_ui_set_object(tv_action, "tvAction", GTK_TYPE_TREE_VIEW);
    _ntd_ui_set_object(nb_action, "nbAction", GTK_TYPE_NOTEBOOK);
    _ntd_ui_set_object(dlg_save_to, "dlgSaveTo", GTK_TYPE_DIALOG);
    _ntd_ui_set_object(txt_save_to, "txtSaveTo", NTD_TYPE_REGEX_ENTRY);
    _ntd_ui_set_object(btn_save_to, "btnSaveTo", GTK_TYPE_BUTTON);
    _ntd_ui_set_object(chk_read_from, "chkReadFrom", GTK_TYPE_CHECK_BUTTON);
    _ntd_ui_set_object(txt_read_from, "txtReadFrom", NTD_TYPE_REGEX_ENTRY);
    _ntd_ui_set_object(chk_read_size, "chkReadSize", GTK_TYPE_CHECK_BUTTON);
    _ntd_ui_set_object(txt_read_size, "txtReadSize", NTD_TYPE_REGEX_ENTRY);
    _ntd_ui_set_object(fc_load_from, "fcLoadFrom", GTK_TYPE_FILE_CHOOSER_BUTTON);
    _ntd_ui_set_object(chk_erase, "chkErase", GTK_TYPE_CHECK_BUTTON);
    _ntd_ui_set_object(chk_write_to, "chkWriteTo", GTK_TYPE_CHECK_BUTTON);
    _ntd_ui_set_object(txt_write_to, "txtWriteTo", NTD_TYPE_REGEX_ENTRY);
    _ntd_ui_set_object(chk_verify, "chkVerify", GTK_TYPE_CHECK_BUTTON);
    _ntd_ui_set_object(chk_execute, "chkExecute", GTK_TYPE_CHECK_BUTTON);
    _ntd_ui_set_object(spn_progress, "spnProgress", GTK_TYPE_SPINNER);
    _ntd_ui_set_object(pb_progress, "pbProgress", GTK_TYPE_PROGRESS_BAR);
    _ntd_ui_set_object(lbl_progress, "lblProgress", GTK_TYPE_LABEL);
    _ntd_ui_set_object(btn_run, "btnRun", GTK_TYPE_BUTTON);
    _ntd_ui_set_object(btn_stop, "btnStop", GTK_TYPE_BUTTON);
    _ntd_ui_set_object(btn_close, "btnClose", GTK_TYPE_BUTTON);
    return TRUE;
}

static gchar *
_ntd_ui_get_valid_text(GtkWidget *widget)
{
    NtdRegexEntry *entry = NTD_REGEX_ENTRY(widget);

    if (! ntd_regex_entry_get_validity(entry, FALSE))
        return NULL;

    return g_strdup(gtk_entry_get_text(GTK_ENTRY(widget)));
}

static void
_ntd_ui_from_config(_NtdUI *ui)
{
    const _NtdConfig *cfg = ui->config;

    if (cfg->device != NULL) {
        if (GTK_IS_FILE_CHOOSER(ui->txt_device)) {
            gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(ui->txt_device), cfg->device);
        } else {
            gtk_combo_box_set_active_id(GTK_COMBO_BOX(ui->txt_device), cfg->device);
        }
    }
    gtk_combo_box_set_active(GTK_COMBO_BOX(ui->lst_protocol), cfg->protocol);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ui->chk_init), cfg->init);
    _ntd_tree_view_set_action(GTK_TREE_VIEW(ui->tv_action), cfg->action);
    gtk_notebook_set_current_page(GTK_NOTEBOOK(ui->nb_action), cfg->action);

    /* READ action */
    if (cfg->save_to != NULL)
        gtk_entry_set_text(GTK_ENTRY(ui->txt_save_to), cfg->save_to);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ui->chk_read_from), cfg->read_from);
    if (cfg->read_from_address != NULL)
        gtk_entry_set_text(GTK_ENTRY(ui->txt_read_from), cfg->read_from_address);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ui->chk_read_size), cfg->read_size);
    if (cfg->read_size_bytes != NULL)
        gtk_entry_set_text(GTK_ENTRY(ui->txt_read_size), cfg->read_size_bytes);

    /* WRITE action */
    if (cfg->load_from != NULL)
        gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(ui->fc_load_from), cfg->load_from);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ui->chk_erase), cfg->erase);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ui->chk_write_to), cfg->write_to);
    if (cfg->write_to_address != NULL)
        gtk_entry_set_text(GTK_ENTRY(ui->txt_write_to), cfg->write_to_address);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ui->chk_verify), cfg->verify);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ui->chk_execute), cfg->execute);
}

static void
_ntd_ui_to_config(_NtdUI *ui)
{
    _NtdConfig *cfg = ui->config;

    /* Free all the config record and repopulate it */
    _ntd_config_free(cfg);

    if (GTK_IS_FILE_CHOOSER(ui->txt_device)) {
        cfg->device = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(ui->txt_device));
    } else {
        cfg->device = g_strdup(gtk_combo_box_get_active_id(GTK_COMBO_BOX(ui->txt_device)));
    }
    cfg->protocol = gtk_combo_box_get_active(GTK_COMBO_BOX(ui->lst_protocol));
    cfg->init = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(ui->chk_init));
    cfg->action = gtk_notebook_get_current_page(GTK_NOTEBOOK(ui->nb_action));

    /* READ action */
    cfg->save_to = g_strdup(gtk_entry_get_text(GTK_ENTRY(ui->txt_save_to)));
    cfg->read_from = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(ui->chk_read_from));
    cfg->read_from_address = _ntd_ui_get_valid_text(ui->txt_read_from);
    cfg->read_size = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(ui->chk_read_size));
    cfg->read_size_bytes = _ntd_ui_get_valid_text(ui->txt_read_size);

    /* WRITE action */
    cfg->load_from = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(ui->fc_load_from));
    cfg->erase = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(ui->chk_erase));
    cfg->write_to = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(ui->chk_write_to));
    cfg->write_to_address = _ntd_ui_get_valid_text(ui->txt_write_to);
    cfg->verify = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(ui->chk_verify));
    cfg->execute = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(ui->chk_execute));
}

/* Check for a valid READ action */
#define _NTD_VALID_READ(config) \
    (config->action == _NTD_ACTION_READ && \
     config->save_to != NULL && config->save_to[0] != 0 && \
     (! config->read_from || config->read_from_address != NULL) && \
     (! config->read_size || config->read_size_bytes != NULL))

/* Check for a valid WRITE action */
#define _NTD_VALID_WRITE(config) \
    (config->action == _NTD_ACTION_WRITE && \
     config->load_from != NULL && config->load_from[0] != 0 && \
     (! config->write_to || config->write_to_address != NULL))

static void
_ntd_ui_refresh(_NtdUI *ui)
{
    const _NtdConfig *cfg = ui->config;

    gtk_widget_set_sensitive(ui->txt_read_from, cfg->read_from);
    gtk_widget_set_sensitive(ui->txt_read_size, cfg->read_size);
    gtk_widget_set_sensitive(ui->txt_write_to, cfg->write_to);
    gtk_widget_set_sensitive(ui->btn_run,
                             cfg->device != NULL && cfg->protocol >= 0 &&
                             (_NTD_VALID_READ(cfg) || _NTD_VALID_WRITE(cfg)));
}

static void
_ntd_ui_set_protocols(_NtdUI *ui)
{
    GtkComboBoxText *combo_box;
    const gchar *name;
    gint row;

    combo_box = (GtkComboBoxText *) ui->lst_protocol;

    for (row = 0; (name = _ntd_protocols[row][1]) != NULL; ++row) {
        gtk_combo_box_text_append_text(combo_box, name);
    }
}

static void
_ntd_ui_switch_action(_NtdUI *ui)
{
    GtkTreeView *tree_view;
    gint action;

    tree_view = GTK_TREE_VIEW(ui->tv_action);
    action = 0;

    if (! _ntd_tree_view_get_action(tree_view, &action))
        _ntd_tree_view_set_action(tree_view, 0);

    gtk_notebook_set_current_page(GTK_NOTEBOOK(ui->nb_action), action);
}

static void
_ntd_ui_message(const gchar *log_domain, GLogLevelFlags log_level,
                const gchar *message, gpointer user_data)
{
    _NtdUI *ui = (_NtdUI *) user_data;
    gtk_label_set_text(GTK_LABEL(ui->lbl_progress), message);
    _ntd_refresh_ui();
}

static void
_ntd_ui_prepare(_NtdUI *ui)
{
    gboolean done;

    _ntd_ui_init(ui);
    done = _ntd_ui_set_builder(ui) && _ntd_ui_set_objects(ui);
    _ntd_ui_free_builder(ui);

    if (done) {
        _ntd_ui_set_protocols(ui);
        g_log_set_handler(NULL, G_LOG_LEVEL_MESSAGE, _ntd_ui_message, ui);
    } else {
        _ntd_ui_free(ui);
        exit(2);
    }
}

static void
_ntd_ui_save_to_dialog(_NtdUI *ui)
{
    _NtdConfig *cfg = ui->config;
    GtkFileChooser *dialog = (GtkFileChooser *) ui->dlg_save_to;

    if (cfg->save_to != NULL) {
        gchar *name = g_path_get_basename(cfg->save_to);
        gchar *folder = g_path_get_dirname(cfg->save_to);

        gtk_file_chooser_set_current_name(dialog, name);
        gtk_file_chooser_set_current_folder(dialog, folder);

        g_free(name);
        g_free(folder);
    }

    if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
        g_free(cfg->save_to);
        cfg->save_to = gtk_file_chooser_get_filename(dialog);
        gtk_entry_set_text(GTK_ENTRY(ui->txt_save_to), cfg->save_to);
    }

    gtk_widget_hide(GTK_WIDGET(dialog));
}

static void
_ntd_ui_changed(_NtdUI *ui)
{
    _ntd_ui_to_config(ui);
    _ntd_ui_refresh(ui);
}

static gboolean
_ntd_ui_run(gpointer user_data)
{
    const _NtdUI *ui = user_data;
    NtdIsp *isp = _ntd_config_create_isp(ui->config);
    GtkProgressBar *progress_bar = GTK_PROGRESS_BAR(ui->pb_progress);

    gtk_spinner_start(GTK_SPINNER(ui->spn_progress));
    gtk_progress_bar_set_fraction(progress_bar, 0);

    gtk_widget_set_sensitive(ui->btn_run, FALSE);
    gtk_widget_set_sensitive(ui->btn_stop, TRUE);
    gtk_widget_set_sensitive(ui->pb_progress, TRUE);
    gtk_widget_set_sensitive(ui->spn_progress, TRUE);

    if (isp != NULL) {
        g_signal_connect_swapped(isp, "progress-notify",
                                 G_CALLBACK(_ntd_update_progression), progress_bar);
        if (_ntd_config_probe(ui->config, isp) &&
            _ntd_config_execute(ui->config, isp))
        {
            gtk_progress_bar_set_fraction(progress_bar, 1);
            g_message("Operation completed successfully");
        }

        g_object_unref(isp);
    }

    gtk_widget_set_sensitive(ui->btn_run, TRUE);
    gtk_widget_set_sensitive(ui->btn_stop, FALSE);
    gtk_widget_set_sensitive(ui->pb_progress, FALSE);
    gtk_widget_set_sensitive(ui->spn_progress, FALSE);

    gtk_spinner_stop(GTK_SPINNER(ui->spn_progress));
    return FALSE;
}

static void
_ntd_ui_trigger_run(_NtdUI *ui)
{
    _ntd_ui_to_config(ui);
    g_idle_add(_ntd_ui_run, ui);
}

#if !GTK_CHECK_VERSION(3,4,5)
static void
_ntd_update_preview(GtkFileChooser *chooser, _NtdUI *ui)
{
    gchar *filename = gtk_file_chooser_get_filename(chooser);
    if (filename == NULL) {
        /* When cancelling the file selection dialog, the file name
         * is unset *without* emitting GtkFileChooserButton::file-set.
         * Connecting this function to GtkFileChooser:update-preview
         * is a temporary workaround to let be notified anyway:
         * http://mail.gnome.org/archives/gtk-app-devel-list/2012-October/msg00007.html
         */
        _ntd_ui_changed(ui);
    } else {
        g_free(filename);
    }
}
#endif

static void
_ntd_ui_set_behavior(_NtdUI *ui)
{
#if !GTK_CHECK_VERSION(3,4,5)
    if (GTK_IS_FILE_CHOOSER(ui->txt_device)) {
        g_signal_connect(ui->txt_device, "update-preview",
                         G_CALLBACK(_ntd_update_preview), ui);
    }
    g_signal_connect(ui->fc_load_from, "update-preview",
                     G_CALLBACK(_ntd_update_preview), ui);
#endif

    g_signal_connect(ui->wnd_main, "delete-event",
                     G_CALLBACK(_ntd_quit), NULL);
    g_signal_connect_swapped(ui->btn_save_to, "clicked",
                             G_CALLBACK(_ntd_ui_save_to_dialog), ui);

    if (GTK_IS_FILE_CHOOSER(ui->txt_device)) {
        g_signal_connect_swapped(ui->txt_device, "file-set",
                                 G_CALLBACK(_ntd_ui_changed), ui);
    } else {
        g_signal_connect_swapped(ui->txt_device, "changed",
                                 G_CALLBACK(_ntd_ui_changed), ui);
    }
    g_signal_connect_swapped(ui->lst_protocol, "changed",
                             G_CALLBACK(_ntd_ui_changed), ui);
    g_signal_connect_swapped(gtk_tree_view_get_selection(GTK_TREE_VIEW(ui->tv_action)),
                             "changed",
                             G_CALLBACK(_ntd_ui_switch_action), ui);
    g_signal_connect_data(ui->nb_action, "switch-page",
                          G_CALLBACK(_ntd_ui_changed), ui, NULL,
                          G_CONNECT_AFTER | G_CONNECT_SWAPPED);

    g_signal_connect_swapped(ui->txt_save_to, "validity-changed",
                             G_CALLBACK(_ntd_ui_changed), ui);
    g_signal_connect_swapped(ui->chk_read_from, "toggled",
                             G_CALLBACK(_ntd_ui_changed), ui);
    g_signal_connect_swapped(ui->txt_read_from, "validity-changed",
                             G_CALLBACK(_ntd_ui_changed), ui);
    g_signal_connect_swapped(ui->chk_read_size, "toggled",
                             G_CALLBACK(_ntd_ui_changed), ui);
    g_signal_connect_swapped(ui->txt_read_size, "validity-changed",
                             G_CALLBACK(_ntd_ui_changed), ui);
    g_signal_connect_swapped(ui->fc_load_from, "file-set",
                             G_CALLBACK(_ntd_ui_changed), ui);
    g_signal_connect_swapped(ui->chk_write_to, "toggled",
                             G_CALLBACK(_ntd_ui_changed), ui);
    g_signal_connect_swapped(ui->txt_write_to, "validity-changed",
                             G_CALLBACK(_ntd_ui_changed), ui);

    g_signal_connect_swapped(ui->btn_run, "clicked",
                             G_CALLBACK(_ntd_ui_trigger_run), ui);
    g_signal_connect(ui->btn_close, "clicked",
                     G_CALLBACK(_ntd_quit), NULL);
}

static void
_ntd_ui_set_config(_NtdUI *ui, _NtdConfig *config)
{
    ui->config = config;
    _ntd_ui_from_config(ui);
    _ntd_ui_refresh(ui);

    /* Behavior is set after updating the ui with the config record
     * to avoid nasty cyclic dependencies. */
    _ntd_ui_set_behavior(ui);
}


gint
main(gint argc, gchar **argv)
{
    _NtdUI ui;
    _NtdConfig config;

    _ntd_init(argc, argv);

    /* An error in the following two functions is fatal */
    _ntd_parse_args(&argc, &argv);
    _ntd_ui_prepare(&ui);

    _ntd_config_read(&config);
    _ntd_ui_set_config(&ui, &config);

    gtk_widget_show_all(ui.wnd_main);
    gtk_main();

    _ntd_ui_to_config(&ui);
    _ntd_ui_free(&ui);

    _ntd_config_write(&config);
    _ntd_config_free(&config);

    return 0;
}
