/* ntdisp - NtD In System Programming.
 *
 * Copyright (C) 2012  Nicola Fontana <ntd at entidi.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */


#include <glib-object.h>
#include <glib/gprintf.h>
#include <string.h>

#include "../config.h"
#include "ntdisp-intl.h"
#include "ntdisp-utils.h"
#include "ntdisp-marshal.h"
#include "ntdisp-builtins.h"
#include "ntd-device.h"
#include "ntd-device-private.h"



enum {
    PROP_0,
    PROP_FILE
};

enum {
    OPEN,
    CLOSE,
    FLUSH,
    READ,
    WRITE,
    ERROR_SET,
    LAST_SIGNAL
};

static void     ntd_device_class_init           (NtdDeviceClass*    klass);
static void     ntd_device_init                 (NtdDevice*         device);
static void     ntd_device_dispose              (GObject*           object);
static void     ntd_device_get_property         (GObject*           object,
                                                 guint              prop_id,
                                                 GValue*            value,
                                                 GParamSpec*        pspec);
static void     ntd_device_set_property         (GObject*           object,
                                                 guint              prop_id,
                                                 const GValue*      value,
                                                 GParamSpec*        pspec);
static void     ntd_device_real_open            (NtdDevice*         device);
static void     ntd_device_real_close           (NtdDevice*         device);
static void     ntd_device_real_flush           (NtdDevice*         device,
                                                 gboolean           input,
                                                 gboolean           output);
static guint    ntd_device_real_read            (NtdDevice*         device,
                                                 guint              nbytes,
                                                 gpointer           buffer);
static guint    ntd_device_real_write           (NtdDevice*         device,
                                                 guint              nbytes,
                                                 gconstpointer      buffer);
static gchar*   ntd_device_vsprintf             (NtdDevice*         device,
                                                 const gchar*       format,
                                                 va_list            args);
static void     ntd_dump_incoming               (gconstpointer      data,
                                                 gsize              nbytes);
static void     ntd_dump_outcoming              (gconstpointer      data,
                                                 gsize              nbytes);

static guint            signals[LAST_SIGNAL] = { 0 };
static GObjectClass*    parent_class = NULL;


GType
ntd_device_get_type(void)
{
    static GType device_type = 0;

    if G_UNLIKELY(device_type == 0) {
        static const GTypeInfo device_info = {
            sizeof(NtdDeviceClass),
            NULL,                   /* base_init */
            NULL,                   /* base_finalize */
            (GClassInitFunc)        ntd_device_class_init,
            NULL,                   /* class_finalize */
            NULL,                   /* class_data */
            sizeof(NtdDevice),
            0,                      /* n_preallocs */
            (GInstanceInitFunc)     ntd_device_init,
        };

        device_type = g_type_register_static(G_TYPE_OBJECT, "NtdDevice",
                                             &device_info, G_TYPE_FLAG_ABSTRACT);
    }

    return device_type;
}

static void
ntd_device_class_init(NtdDeviceClass *klass)
{
    GObjectClass *gobject_class;
    GParamSpec *param;
    GClosure *closure;
    GType param_types[2];

    parent_class = g_type_class_peek_parent(klass);
    g_type_class_add_private(klass, sizeof(NtdDevicePrivate));

    gobject_class = G_OBJECT_CLASS(klass);
    gobject_class->set_property = ntd_device_set_property;
    gobject_class->get_property = ntd_device_get_property;
    gobject_class->dispose = ntd_device_dispose;

    klass->open = NULL;
    klass->close = NULL;
    klass->flush = NULL;
    klass->raw_read = NULL;
    klass->raw_write = NULL;

    param = g_param_spec_string("file",
                                P_("File"),
                                P_("Device file bound to this device"),
                                NULL,
                                G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_FILE, param);

    /**
     * NtdDevice::open:
     * @device: an #NtdDevice based instance
     *
     * Emitted whenever @device is opened. On success, the internal is-opened
     * state is set to %TRUE. If yet set, that is if @device has been opened
     * before, no actions are taken.
     **/
    closure = g_cclosure_new(G_CALLBACK(ntd_device_real_open), NULL, NULL);
    signals[OPEN] = g_signal_newv("open", NTD_TYPE_DEVICE,
                                  G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE,
                                  closure, NULL, NULL,
                                  ntd_marshal_VOID__VOID,
                                  G_TYPE_NONE, 0, NULL);

    /**
     * NtdDevice::close:
     * @device: an #NtdDevice based instance
     *
     * Emitted whenever @device is closed.
     **/
    closure = g_cclosure_new(G_CALLBACK(ntd_device_real_close), NULL, NULL);
    signals[CLOSE] = g_signal_newv("close", NTD_TYPE_DEVICE,
                                   G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE,
                                   closure, NULL, NULL,
                                   ntd_marshal_VOID__VOID,
                                   G_TYPE_NONE, 0, NULL);

    /**
     * NtdDevice::flush:
     * @device: an #NtdDevice based instance
     * @input: whether to flush the input queue
     * @output: whether to flush the output queue
     *
     * Emitted whenever @device must be flushed. One of @input or @output
     * must be %TRUE otherwise the internal error will be set.
     **/
    closure = g_cclosure_new(G_CALLBACK(ntd_device_real_flush), NULL, NULL);
    param_types[0] = G_TYPE_BOOLEAN;
    param_types[1] = G_TYPE_BOOLEAN;
    signals[FLUSH] = g_signal_newv("flush", NTD_TYPE_DEVICE,
                                   G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE,
                                   closure, NULL, NULL,
                                   ntd_marshal_VOID__BOOLEAN_BOOLEAN,
                                   G_TYPE_NONE, 2, param_types);

    /**
     * NtdDevice::read:
     * @device: an #NtdDevice based instance
     * @nbytes: number of bytes read
     * @buffer: pointer to the destination memory
     *
     * Emitted whenever a read operation, that is a call to the
     * raw_read virtual method, is performed.
     *
     * Returns: the number of bytes effectively read.
     **/
    closure = g_cclosure_new(G_CALLBACK(ntd_device_real_read), NULL, NULL);
    param_types[0] = G_TYPE_UINT;
    param_types[1] = G_TYPE_POINTER;
    signals[READ] = g_signal_newv("read", NTD_TYPE_DEVICE,
                                  G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE,
                                  closure, NULL, NULL,
                                  ntd_marshal_UINT__UINT_POINTER,
                                  G_TYPE_UINT, 2, param_types);

    /**
     * NtdDevice::write:
     * @device: an #NtdDevice based instance
     * @nbytes: number of bytes written
     * @buffer: pointer to the source memory
     *
     * Emitted whenever a write operation, that is a call to the
     * raw_write virtual method, is performed.
     *
     * Returns: the number of bytes effectively written.
     **/
    closure = g_cclosure_new(G_CALLBACK(ntd_device_real_write), NULL, NULL);
    signals[WRITE] = g_signal_newv("write",
                                   NTD_TYPE_DEVICE,
                                   G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE,
                                   closure, NULL, NULL,
                                   ntd_marshal_UINT__UINT_POINTER,
                                   G_TYPE_UINT, 2, param_types);

    /**
     * NtdDevice::error-set:
     * @device: an #NtdDevice based instance
     *
     * Emitted whenever an I/O call fails.
     **/
    signals[ERROR_SET] = g_signal_new("error-set",
                                      NTD_TYPE_DEVICE,
                                      G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE,
                                      G_STRUCT_OFFSET(NtdDeviceClass, error_set),
                                      NULL, NULL,
                                      ntd_marshal_VOID__VOID,
                                      G_TYPE_NONE, 0);
}

static void
ntd_device_init(NtdDevice *device)
{
    NtdDevicePrivate *data = G_TYPE_INSTANCE_GET_PRIVATE(device,
                                                         NTD_TYPE_DEVICE,
                                                         NtdDevicePrivate);
    data->file = NULL;
    data->is_opened = FALSE;
    data->error = NULL;

    device->data = data;
}

static void
ntd_device_dispose(GObject *object)
{
    NtdDevice *device;
    NtdDevicePrivate *data;

    device = (NtdDevice *) object;
    data = device->data;

    if (data->is_opened)
        ntd_device_close(device);

    if (data->file != NULL) {
        g_free(data->file);
        data->file = NULL;
    }

    if (data->error != NULL) {
        g_clear_error(& data->error);
        data->error = NULL;
    }

    if (parent_class->dispose != NULL)
        parent_class->dispose(object);
}


static void
ntd_device_get_property(GObject *object, guint prop_id,
                        GValue *value, GParamSpec *pspec)
{
    NtdDevice *device = (NtdDevice *) object;
    NtdDevicePrivate *data = device->data;

    switch (prop_id) {
    case PROP_FILE:
        g_value_set_string(value, data->file);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
ntd_device_set_property(GObject *object, guint prop_id,
                        const GValue *value, GParamSpec *pspec)
{
    NtdDevice *device = (NtdDevice *) object;

    switch (prop_id) {
    case PROP_FILE:
        ntd_device_set_file(device, g_value_get_string(value));
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}


/**
 * ntd_device_get_error:
 * @device: an #NtdDevice based instance
 *
 * Gets the device internal error. This function is basically called when
 * catching the errors on a device. Be sure to call ntd_device_clear_error()
 * to reset the internal error after the treatment.
 *
 * Returns: a #GError pointer to the internal error, or %NULL if @device
 *          is not a #NtdDevice.
 **/
GError *
ntd_device_get_error(NtdDevice *device)
{
    NtdDevicePrivate *data;

    g_return_val_if_fail(NTD_IS_DEVICE(device), NULL);

    data = device->data;
    return data->error;
}


/**
 * ntd_device_set_errorv:
 * @device: an #NtdDevice based instance
 * @error_code: the #NtdSystemError error code to use
 * @format: printf-style format string
 * @args: parameters for format string
 *
 * Wrapper to ntd_device_set_error() with #va_list argument.
 **/
void
ntd_device_set_errorv(NtdDevice *device, NtdSystemError error_code,
                      const gchar *format, va_list args)
{
    NtdDevicePrivate *data;
    gchar *message;

    g_return_if_fail(NTD_IS_DEVICE(device));

    data = device->data;
    ntd_return_on_pileup(data->error);

    message = ntd_device_vsprintf(device, format, args);

    ntd_system_set_error_with_code(& data->error, error_code, "%s", message);

    g_free(message);

    g_signal_emit(device, signals[ERROR_SET], 0);
}

/**
 * ntd_device_set_error:
 * @device: an #NtdDevice based instance
 * @error_code: an #NtdSystemError error code
 * @format: printf-style format string
 * @...: parameters for format string
 *
 * Sets the device internal error to the specified error code.
 * Optionally, any custom message can be appended to the standard message.
 *
 * At the end, an #NtdDevice::error-set signal will be emitted.
 **/
void
ntd_device_set_error(NtdDevice *device, NtdSystemError error_code,
                     const gchar *format, ...)
{
    va_list args;

    va_start(args, format);
    ntd_device_set_errorv(device, error_code, format, args);
    va_end(args);
}

/**
 * ntd_device_set_error_from_system:
 * @device: an #NtdDevice based instance
 * @format: printf-style format string
 * @...: parameters for format string
 *
 * Checks the system for any error and eventually sets the device internal
 * error.
 *
 * If the device internal error is yet set, the function returns %TRUE
 * immediately.
 *
 * If no system errors are pending, nothing is done and %FALSE is returned.
 *
 * If a system error is present, the device internal error is set accordling
 * to the system error and a ntd_system_clear_error() call is performed to
 * reset the system error. Any custom message can be optionally appended to
 * the standard message.
 *
 * At the end, an #NtdDevice::error-set signal will be emitted.
 *
 * Returns: %TRUE if @device is in error or a system error is pending,
 *          %FALSE otherwise.
 **/
gboolean
ntd_device_set_error_from_system(NtdDevice *device, const gchar *format, ...)
{
    NtdDevicePrivate *data;
    NtdSystemError error_code;
    va_list args;

    error_code = ntd_system_get_error();
    if (error_code == NTD_SYSTEM_ERROR_NONE)
        return FALSE;

    g_return_val_if_fail(NTD_IS_DEVICE(device), TRUE);

    data = device->data;
    ntd_return_val_on_pileup(data->error, TRUE);

    va_start(args, format);
    ntd_device_set_errorv(device, error_code, format, args);
    va_end(args);

    ntd_system_clear_error();

    return TRUE;
}

/**
 * ntd_device_clear_error:
 * @device: an #NtdDevice based instance
 *
 * Resets the device internal error.
 *
 * Returns: %TRUE if there was an internal error, %FALSE otherwise.
 **/
gboolean
ntd_device_clear_error(NtdDevice *device)
{
    NtdDevicePrivate *data;

    g_return_val_if_fail(NTD_IS_DEVICE(device), FALSE);

    data = device->data;

    if (data->error == NULL)
        return FALSE;

    g_clear_error(& data->error);
    return TRUE;
}

/**
 * ntd_device_in_error:
 * @device: an #NtdDevice based instance
 *
 * Checks the presence of internal errors.
 *
 * Returns: %TRUE if the device internal error is set,
 *          %FALSE otherwise.
 **/
gboolean
ntd_device_in_error(NtdDevice *device)
{
    NtdDevicePrivate *data;

    g_return_val_if_fail(NTD_IS_DEVICE(device), FALSE);

    data = device->data;
    return data->error != NULL;
}


/**
 * ntd_device_set_file:
 * @device: an #NtdDevice based instance
 * @file: the file name
 *
 * Sets the file name of the device bound to @device.
 * On posix system, @file is usually a file inside the
 * "/dev/" path.
 **/
void
ntd_device_set_file(NtdDevice *device, const gchar *file)
{
    NtdDevicePrivate *data;

    g_return_if_fail(NTD_IS_DEVICE(device));

    data = device->data;

    if (g_strcmp0(data->file, file) == 0)
        return;

    g_free(data->file);
    data->file = g_strdup(file);

    g_object_notify(G_OBJECT(device), "file");
}


/**
 * ntd_device_get_file:
 * @device: an #NtdDevice based instance
 *
 * Retrieves the file name bound to @device.
 *
 * Returns: a pointer to the device file or %NULL if there is
 *          no associated device. This string points to internally
 *          allocated memory and must not be modified or freed.
 **/
const gchar *
ntd_device_get_file(NtdDevice *device)
{
    NtdDevicePrivate *data;

    g_return_val_if_fail(NTD_IS_DEVICE(device), NULL);

    data = device->data;
    return data->file;
}

/**
 * ntd_device_open:
 * @device: an #NtdDevice based instance
 *
 * Starts the communication over @device. If @device is in error,
 * no operation is performed.
 **/
void
ntd_device_open(NtdDevice *device)
{
    g_return_if_fail(NTD_IS_DEVICE(device));
    g_signal_emit(device, signals[OPEN], 0);
}

/**
 * ntd_device_close:
 * @device: an #NtdDevice based instance
 *
 * Terminates the communication on @device.
 *
 * The device is automatically closed whenever the object is
 * destroyed.
 **/
void
ntd_device_close(NtdDevice *device)
{
    g_return_if_fail(NTD_IS_DEVICE(device));
    g_signal_emit(device, signals[CLOSE], 0);
}

/**
 * ntd_device_is_opened:
 * @device: an #NtdDevice based instance
 *
 * Checks if @device is yet opened. This does not mean @device
 * is really opened, but that a previous call to ntd_device_open()
 * has been performed.
 *
 * Returns: %TRUE if @device is opened, %FALSE otherwise
 **/
gboolean
ntd_device_is_opened(NtdDevice *device)
{
    NtdDevicePrivate *data;

    g_return_val_if_fail(NTD_IS_DEVICE(device), FALSE);

    data = device->data;
    return data->is_opened;
}

/**
 * ntd_device_flush:
 * @device: an #NtdDevice based instance
 * @input: whether to flush the input queue
 * @output: whether to flush the output queue
 *
 * Flushes the input queue (if @input is %TRUE) and/or the output
 * queue (if @output is %TRUE) on @device.
 **/
void
ntd_device_flush(NtdDevice *device, gboolean input, gboolean output)
{
    g_return_if_fail(NTD_IS_DEVICE(device));
    g_signal_emit(device, signals[FLUSH], 0, input, output);
}

/**
 * ntd_device_read:
 * @device: an #NtdDevice based instance
 * @nbytes: number of bytes to read
 *
 * Reads @nbytes bytes of data from @device and returns
 * the result as a newly allocated buffer. If an unexpected
 * condition is encountered before all the @nbytes are read
 * (such as a communication timeout), the data is freed
 * and %NULL is returned.
 *
 * If an unexpected condition is encountered, such as a
 * communication timeout, the buffer is freed and %NULL is
 * returned. Use another function if partial data are expected.
 *
 * Returns: a newly allocated buffer of @nbytes bytes.
 *          Dispose the buffer with g_free() when no longer needed.
 **/
gpointer
ntd_device_read(NtdDevice *device, gsize nbytes)
{
    gpointer buffer;
    guint    nread;

    g_return_val_if_fail(NTD_IS_DEVICE(device), NULL);

    buffer = g_malloc(nbytes);
    nread = 0;
    g_signal_emit(device, signals[READ], 0, nbytes, buffer, &nread);
    if (nread != nbytes) {
        /* Bytes to read and bytes read do not match */
        g_free(buffer);
        buffer = NULL;
    }

    ntd_dump_incoming(buffer, nread);
    return buffer;
}

/**
 * ntd_device_read_until:
 * @device: an #NtdDevice based instance
 * @eol: end-of-line termination character
 *
 * Reads I/O data from @device until the specified @eol terminating
 * character is found. Whatever has been read, including the EOL
 * character, is returned.
 *
 * The returned buffer could possibly embed NUL bytes and must be
 * freed with g_free() when no longer needed.
 *
 * If an unexpected condition is encountered, such as a
 * communication timeout, the buffer is freed and %NULL is
 * returned. Use another function if partial data are expected.
 *
 * Returns: whatever received up to the EOL byte or %NULL on errors.
 *          Dispose the buffer with g_free() when no longer needed.
 **/
gpointer
ntd_device_read_until(NtdDevice *device, gchar eol)
{
    NtdDevicePrivate *data;
    GString *buffer;
    gchar item;
    guint bytes_read;

    g_return_val_if_fail(NTD_IS_DEVICE(device), NULL);

    data = device->data;
    ntd_return_val_on_pileup(data->error, NULL);

    /* g_string_sized_new() initializes buffer to "" */
    buffer = g_string_sized_new(32);
    do {
        g_signal_emit(device, signals[READ], 0, 1, &item, &bytes_read);

        if (bytes_read == 0) {
            /* Timeout or read error */
            break;
        } else if (bytes_read != 1) {
            g_critical("Read more than one byte (%u)", bytes_read);
            break;
        }

        g_string_append_c(buffer, item);
    } while (item != eol);

    ntd_dump_incoming(buffer->str, buffer->len);
    if (item != eol) {
        /* Error: return NULL */
        g_string_free(buffer, TRUE);
        return NULL;
    }

    return g_string_free(buffer, FALSE);
}

/**
 * ntd_device_write:
 * @device: an #NtdDevice based instance
 * @nbytes: number of bytes to write
 * @buffer: the source buffer where data come from
 *
 * Writes @nbytes bytes of data from @buffer to @device.
 *
 * Returns: %TRUE if all the data has been sent, %FALSE otherwise.
 **/
gboolean
ntd_device_write(NtdDevice *device, gsize nbytes, gconstpointer buffer)
{
    guint nwritten;

    g_return_val_if_fail(NTD_IS_DEVICE(device), FALSE);

    g_signal_emit(device, signals[WRITE], 0, nbytes, buffer, &nwritten);

    ntd_dump_outcoming(buffer, nwritten);
    return nwritten == nbytes;
}

/**
 * ntd_device_printf:
 * @device: an #NtdDevice based instance
 * @format: printf-style format string
 * @...:  parameters for format string
 *
 * Provides a convenient way to send complex formatted data
 * throught @device by using a printf style approach.
 *
 * @format and @... are passed directly to g_vasprintf() so
 * check the GLib manual for further details.
 *
 * Returns: %TRUE on success, %FALSE on errors or timeout.
 **/
gboolean
ntd_device_printf(NtdDevice *device, const gchar *format, ...)
{
    va_list var_args;
    gboolean done;

    va_start(var_args, format);
    done = ntd_device_vprintf(device, format, var_args);
    va_end(var_args);

    return done;
}

/**
 * ntd_device_vprintf:
 * @device: an #NtdDevice based instance
 * @format: printf-style format string
 * @var_args: va_list of parameters for format string
 *
 * A variadic version of ntd_device_printf().
 *
 * Returns: %TRUE on success, %FALSE on errors or timeout.
 **/
gboolean
ntd_device_vprintf(NtdDevice *device, const gchar *format, va_list var_args)
{
    gchar *buffer;
    gint nbytes;
    gboolean done;

    g_return_val_if_fail(NTD_IS_DEVICE(device), FALSE);

    nbytes = g_vasprintf(&buffer, format, var_args);
    g_return_val_if_fail(buffer != NULL, FALSE);

    done = ntd_device_write(device, nbytes, buffer);
    g_free(buffer);

    return done;
}


static void
ntd_device_real_open(NtdDevice *device)
{
    NtdDevicePrivate *data;
    NtdDeviceClass *klass;

    data = device->data;
    g_return_if_fail(! data->is_opened);
    ntd_return_on_pileup(data->error);

    klass = NTD_DEVICE_GET_CLASS(device);

    if (klass->open != NULL)
        klass->open(device);

    /* The device is considered opened only if open()
     * has succeeded without errors */
    if (data->error == NULL)
        data->is_opened = TRUE;
}

static void
ntd_device_real_close(NtdDevice *device)
{
    NtdDevicePrivate *data = device->data;
    NtdDeviceClass *klass = NTD_DEVICE_GET_CLASS(device);

    /* No pileup error or is-opened checks: close must close anyway.
     * This is convenient for using paired calls of ntd_device_open()
     * and ntd_device_close() when the former fails.
     */

    if (klass->close != NULL)
        klass->close(device);

    data->is_opened = FALSE;
}

static void
ntd_device_real_flush(NtdDevice *device, gboolean input, gboolean output)
{
    NtdDevicePrivate *data;
    NtdDeviceClass *klass;

    g_return_if_fail(input || output);

    data = device->data;
    g_return_if_fail(data->is_opened);
    ntd_return_on_pileup(data->error);

    klass = NTD_DEVICE_GET_CLASS(device);

    if (klass->flush != NULL)
        klass->flush(device, input, output);
}

static guint
ntd_device_real_read(NtdDevice *device, guint nbytes, gpointer buffer)
{
    NtdDevicePrivate *data;
    NtdDeviceClass *klass;

    data = device->data;
    g_return_val_if_fail(data->is_opened, 0);
    ntd_return_val_on_pileup(data->error, 0);

    if (nbytes == 0)
        return 0;

    g_return_val_if_fail(buffer != NULL, 0);

    klass = NTD_DEVICE_GET_CLASS(device);

    return klass->raw_read ? klass->raw_read(device, nbytes, buffer) : 0;
}

static guint
ntd_device_real_write(NtdDevice* device, guint nbytes, gconstpointer buffer)
{
    NtdDevicePrivate *data;
    NtdDeviceClass *klass;

    data = device->data;
    g_return_val_if_fail(data->is_opened, 0);
    ntd_return_val_on_pileup(data->error, 0);

    if (nbytes == 0)
        return 0;

    g_return_val_if_fail(buffer != NULL, 0);

    klass = NTD_DEVICE_GET_CLASS(device);

    return klass->raw_write ? klass->raw_write(device, nbytes, buffer) : 0;
}

static gchar *
ntd_device_vsprintf(NtdDevice *device, const gchar *format, va_list args)
{
    NtdDevicePrivate *data = device->data;
    gchar *message;

    if (data->file == NULL)
        message = g_strdup(P_("unspecified file"));
    else
        message = g_strdup_printf(P_("file `%s'"), data->file);

    /* Check for no empty format */
    if (format != NULL && format[0] != '\0') {
        gchar *new_message;
        gchar *prefix;

        prefix = g_strdup_vprintf(format, args);
        new_message = g_strdup_printf(P_("%s on %s"), prefix, message);

        g_free(prefix);
        g_free(message);
        message = new_message;
    }

    return message;
}

static void
ntd_dump_incoming(gconstpointer data, gsize nbytes)
{
    gchar *dump;

    /* Do not log anything on empty data */
    if (nbytes == 0)
        return;

    dump = ntd_escape(data, nbytes);
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, " Incoming: %s", dump);
    g_free(dump);
}

static void
ntd_dump_outcoming(gconstpointer data, gsize nbytes)
{
    gchar *dump;

    /* Do not log anything on empty data */
    if (nbytes == 0)
        return;

    dump = ntd_escape(data, nbytes);
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Outcoming: %s", dump);
    g_free(dump);
}
