/* ntdisp - NtD In System Programming.
 *
 * Copyright (C) 2012  Nicola Fontana <ntd at entidi.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * SECTION:ntd-isp
 * @short_description: Abstraction class for the ISP protocol
 *
 **/

/**
 * NtdIsp:
 *
 * All fields are private and should not be used directly.
 * Use its public methods instead.
 **/


#include <glib-object.h>
#include <stdio.h>
#include <string.h>

#include "../config.h"
#include "ntdisp-intl.h"
#include "ntdisp-utils.h"
#include "ntdisp-marshal.h"
#include "ntdisp-enums.h"
#include "ntd-device.h"
#include "ntd-serial.h"
#include "ntd-isp.h"
#include "ntd-isp-private.h"


/* How many bytes to read from stdin at one time */
#define CHUNK_SIZE      1024


G_DEFINE_ABSTRACT_TYPE(NtdIsp, ntd_isp, G_TYPE_OBJECT);

enum {
    PROP_0,
    PROP_SERIAL,
    PROP_STATUS
};

enum {
    STATUS_CHANGED,
    PROGRESS_NOTIFY,
    LAST_SIGNAL
};


static void     ntd_isp_set_property            (GObject*       object,
                                                 guint          prop_id,
                                                 const GValue*  value,
                                                 GParamSpec*    pspec);
static void     ntd_isp_get_property            (GObject*       object,
                                                 guint          prop_id,
                                                 GValue*        value,
                                                 GParamSpec*    pspec);
static guint    _ntd_signals[LAST_SIGNAL] = { 0 };


static void
ntd_isp_class_init(NtdIspClass *klass)
{
    GObjectClass *gobject_class;
    GParamSpec *param;

    gobject_class = (GObjectClass *) klass;

    g_type_class_add_private(klass, sizeof(NtdIspPrivate));

    gobject_class->set_property = ntd_isp_set_property;
    gobject_class->get_property = ntd_isp_get_property;

    param = g_param_spec_object("serial",
                                P_("Serial"),
                                P_("The serial device used by this ISP instance"),
                                NTD_TYPE_SERIAL,
                                G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_SERIAL, param);

    param = g_param_spec_int("status",
                             P_("Status"),
                             P_("Status of the last operation"),
                             -G_MAXINT, G_MAXINT, 0,
                             G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_STATUS, param);

    /**
     * NtdIsp::status-changed
     * @isp: a #NtdIsp
     *
     * Emitted when #NtdIsp:status property has changed.
     **/
    _ntd_signals[STATUS_CHANGED] =
        g_signal_new("status-changed",
                     G_OBJECT_CLASS_TYPE(gobject_class),
                     G_SIGNAL_RUN_FIRST,
                     G_STRUCT_OFFSET(NtdIspClass, status_changed),
                     NULL, NULL,
                     ntd_marshal_VOID__VOID,
                     G_TYPE_NONE, 0);

    /**
     * NtdIsp::progress-notify
     * @isp: a #NtdIsp
     * @fraction: the new fraction value
     *
     * Emitted during the download to notify current progress state.
     **/
    _ntd_signals[PROGRESS_NOTIFY] =
        g_signal_new("progress-notify",
                     G_OBJECT_CLASS_TYPE(gobject_class),
                     G_SIGNAL_RUN_FIRST,
                     G_STRUCT_OFFSET(NtdIspClass, progress_notify),
                     NULL, NULL,
                     ntd_marshal_VOID__DOUBLE,
                     G_TYPE_NONE, 1, G_TYPE_DOUBLE);
}

static void
ntd_isp_init(NtdIsp *isp)
{
    NtdIspPrivate *data = G_TYPE_INSTANCE_GET_PRIVATE(isp,
                                                      NTD_TYPE_ISP,
                                                      NtdIspPrivate);

    data->serial = NULL;
    data->status = 0;
    data->is_connected = FALSE;

    isp->data = data;
}

static void
ntd_isp_set_property(GObject *object, guint prop_id,
                     const GValue *value, GParamSpec *pspec)
{
    NtdIspPrivate *data = ((NtdIsp *) object)->data;

    switch (prop_id) {
    case PROP_SERIAL:
        if (data->serial != g_value_get_object(value)) {
            if (data->serial)
                g_object_unref(data->serial);
            data->serial = g_value_dup_object(value);
        }
        break;
    case PROP_STATUS:
        data->status = g_value_get_int(value);
        g_signal_emit(object, _ntd_signals[STATUS_CHANGED], 0);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
ntd_isp_get_property(GObject *object, guint prop_id,
                     GValue *value, GParamSpec *pspec)
{
    NtdIspPrivate *data = ((NtdIsp *) object)->data;

    switch (prop_id) {
    case PROP_SERIAL:
        g_value_set_object(value, data->serial);
        break;
    case PROP_STATUS:
        g_value_set_int(value, data->status);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}


/**
 * ntd_isp_set_device:
 * @isp: a #NtdIsp instance
 * @file: path to the device file
 *
 * Convenient function to create a new #NtdSerial instance and
 * bind it to @isp (if not yet present) and to initialize it to
 * the default serial values tipically used by this kind of protocol.
 **/
void
ntd_isp_set_device(NtdIsp *isp, const gchar *file)
{
    NtdIspClass *klass;
    NtdIspPrivate *data;

    g_return_if_fail(NTD_IS_ISP(isp));

    /* This method must be defined by the ISP implementation */
    klass = NTD_ISP_GET_CLASS(isp);
    if (klass->set_defaults == NULL) {
        g_warning(_("%s: `set_defaults' method not implemented for type `%s'"),
                  G_STRLOC, g_type_name(G_OBJECT_TYPE(isp)));
        return;
    }

    data = isp->data;

    if (data->serial == NULL) {
        NtdSerial *serial = ntd_serial_new(file);
        g_object_set(isp, "serial", serial, NULL);
    } else {
        g_return_if_fail(NTD_IS_SERIAL(data->serial));
        g_return_if_fail(!data->is_connected);
        ntd_device_set_file(NTD_DEVICE(data->serial), file);
    }

    klass->set_defaults(isp);
}

/**
 * ntd_isp_set_serial:
 * @isp: a #NtdIsp instance
 * @serial: a valid #NtdSerial instance
 *
 * Sets the #NtdIsp:serial property of @isp to @serial.
 **/
void
ntd_isp_set_serial(NtdIsp *isp, NtdSerial *serial)
{
    g_return_if_fail(NTD_IS_ISP(isp));
    g_return_if_fail(NTD_IS_SERIAL(serial));

    g_object_set(isp, "serial", serial, NULL);
}

/**
 * ntd_isp_get_serial:
 * @isp: a #NtdIsp instance
 *
 * Gets the #NtdSerial instance bound to @isp.
 *
 * Returns: the serial object
 **/
NtdSerial *
ntd_isp_get_serial(NtdIsp *isp)
{
    NtdIspPrivate *data;

    g_return_val_if_fail(NTD_IS_ISP(isp), NULL);

    data = isp->data;

    return data->serial;
}

/**
 * ntd_isp_set_status:
 * @isp: a #NtdIsp instance
 * @status: the new status
 *
 * Sets the internal status of @isp to @status.
 **/
void
ntd_isp_set_status(NtdIsp *isp, gint status)
{
    g_return_if_fail(NTD_IS_ISP(isp));
    g_object_set(isp, "status", status, NULL);
}

/**
 * ntd_isp_clear_status:
 * @isp: a #NtdIsp instance
 *
 * Clears the internal status of @isp, that is set
 * the #NtdIsp:status property to 0.
 **/
void
ntd_isp_clear_status(NtdIsp *isp)
{
    ntd_isp_set_status(isp, 0);
}

/**
 * ntd_isp_get_status:
 * @isp: a #NtdIsp instance
 *
 * Gets the internal status of @isp. The internal status is used
 * to hold the result of the last operation executed on @isp.
 *
 * Returns: the current status of @isp
 **/
gint
ntd_isp_get_status(NtdIsp *isp)
{
    NtdIspPrivate *data;

    g_return_val_if_fail(NTD_IS_ISP(isp), 0);

    data = isp->data;

    return data->status;
}

/**
 * ntd_isp_get_message:
 * @isp: a #NtdIsp instance
 *
 * Gets a descriptive string of the current status of @isp. The
 * returned value is owned by @isp and should not be changed or freed.
 *
 * Returns: the status message or %NULL on errors
 **/
const gchar *
ntd_isp_get_message(NtdIsp *isp)
{
    NtdIspClass *klass;
    NtdIspPrivate *data;

    g_return_val_if_fail(NTD_IS_ISP(isp), NULL);

    klass = NTD_ISP_GET_CLASS(isp);
    data = isp->data;

    /* This method must be defined by the ISP implementation */
    if (klass->status_to_message == NULL) {
        g_warning(_("%s: `status_to_message' method not implemented for type `%s'"),
                  G_STRLOC, g_type_name(G_OBJECT_TYPE(isp)));
        return NULL;
    }

    return klass->status_to_message(data->status);
}

/**
 * ntd_isp_status_changed:
 * @isp: a #NtdIsp
 *
 * Emits the #NtdIsp::status-changed signal on @isp.
 * This function is only useful in #NtdIsp implementation.
 **/
void
ntd_isp_status_changed(NtdIsp *isp)
{
    g_return_if_fail(NTD_IS_ISP(isp));

    g_signal_emit(isp, _ntd_signals[STATUS_CHANGED], 0);
}

/**
 * ntd_isp_progress_notify:
 * @isp: a #NtdIsp
 * @fraction: the new fraction value
 *
 * Emits the #NtdIsp::progress-notify signal on @isp.
 **/
void
ntd_isp_progress_notify(NtdIsp *isp, gdouble fraction)
{
    g_return_if_fail(NTD_IS_ISP(isp));

    g_signal_emit(isp, _ntd_signals[PROGRESS_NOTIFY], 0, fraction);
}

/**
 * ntd_isp_connect:
 * @isp: a #NtdIsp instance
 *
 * Connects the host to @isp, performing any initialization procedure
 * requested by the underline protocol.
 *
 * On success, the NtdIsp:status property must be set to 0.
 **/
void
ntd_isp_connect(NtdIsp *isp)
{
    NtdIspPrivate *data;
    NtdIspClass *klass;

    g_return_if_fail(NTD_IS_ISP(isp));

    data = isp->data;
    if (data->is_connected)
        return;

    ntd_isp_clear_status(isp);

    klass = NTD_ISP_GET_CLASS(isp);
    if (klass->connect != NULL)
        klass->connect(isp);

    data->is_connected = data->status == 0;
}

/**
 * ntd_isp_disconnect:
 * @isp: a #NtdIsp instance
 *
 * Disconnects the host from @isp, performing any finalization procedures
 * requested by the protocol.
 **/
void
ntd_isp_disconnect(NtdIsp *isp)
{
    NtdIspPrivate *data;
    NtdIspClass *klass;

    g_return_if_fail(NTD_IS_ISP(isp));

    data = isp->data;
    if (!data->is_connected)
        return;

    klass = NTD_ISP_GET_CLASS(isp);
    if (klass->disconnect != NULL)
        klass->disconnect(isp);

    data->is_connected = FALSE;
}

/**
 * ntd_isp_set_connected:
 * @isp: a #NtdIsp instance
 * @connected: the new flag value
 *
 * Forcibly sets the connection flag of @isp to @connected. This function is
 * reserved for guru use because of its dangerouseness. It is used to skip
 * some initialization steps performed during the connection phase.
 **/
void
ntd_isp_set_connected(NtdIsp *isp, gboolean connected)
{
    NtdIspPrivate *data;

    g_return_if_fail(NTD_IS_ISP(isp));

    data = isp->data;
    data->is_connected = connected;
}

/**
 * ntd_isp_is_connected:
 * @isp: a #NtdIsp instance
 *
 * Checks for the connection status of @isp.
 *
 * Returns: %TRUE if @isp is connected, %FALSE otherwise
 **/
gboolean
ntd_isp_is_connected(NtdIsp *isp)
{
    NtdIspPrivate *data;

    g_return_val_if_fail(NTD_IS_ISP(isp), FALSE);

    data = isp->data;
    return data->is_connected;
}

/**
 * ntd_isp_probe:
 * @isp: a #NtdIsp instance
 *
 * Probes @isp trying to guess the optimal values directly
 * from the target.
 *
 * Returns: %TRUE on success, %FALSE on error
 **/
gboolean
ntd_isp_probe(NtdIsp *isp)
{
    NtdIspClass *klass;

    g_return_val_if_fail(NTD_IS_ISP(isp), FALSE);

    /* This method must be defined by the ISP implementation */
    klass = NTD_ISP_GET_CLASS(isp);
    if (klass->probe == NULL) {
        g_warning(_("%s: `probe' method not implemented for type `%s'"),
                  G_STRLOC, g_type_name(G_OBJECT_TYPE(isp)));
        return FALSE;
    }

    return klass->probe(isp);
}

/**
 * ntd_isp_wipe:
 * @isp: a #NtdIsp instance
 * @address: the address where starting to wipe
 * @nbytes: how many bytes must be cleared
 *
 * Wipes @isp, that is clear the flash on target starting from
 * @address for @nbytes bytes.
 *
 * @address can be %NULL, in which case the starting address of
 * the flash memory will be used.
 *
 * If @nbytes is set to 0, all the flash memory starting from
 * @address will be cleared.
 *
 * Returns: %TRUE on success, %FALSE on error
 **/
gboolean
ntd_isp_wipe(NtdIsp *isp, const gchar *address, gsize nbytes)
{
    NtdIspClass *klass;

    g_return_val_if_fail(NTD_IS_ISP(isp), FALSE);

    /* This method must be defined by the ISP implementation */
    klass = NTD_ISP_GET_CLASS(isp);
    if (klass->wipe == NULL) {
        g_warning(_("%s: `wipe' method not implemented for type `%s'"),
                  G_STRLOC, g_type_name(G_OBJECT_TYPE(isp)));
        return FALSE;
    }

    return klass->wipe(isp, address, nbytes);
}

/**
 * ntd_isp_download:
 * @isp: a #NtdIsp instance
 * @address: the address where starting the download
 * @buffer: the binary data to download
 * @nbytes: how many bytes to download
 *
 * Downloads, that is move from the host (PC) to the target @isp
 * (the embedded device), the provided binary data, starting from
 * @address (an ISP dependent value). When @address is %NULL,
 * this will be the start of the flash memory.
 *
 * Returns: %TRUE on success, %FALSE on error
 **/
gboolean
ntd_isp_download(NtdIsp *isp, const gchar *address,
                 gconstpointer buffer, gsize nbytes)
{
    NtdIspClass *klass;

    g_return_val_if_fail(NTD_IS_ISP(isp), FALSE);

    /* This method must be defined by the ISP implementation */
    klass = NTD_ISP_GET_CLASS(isp);
    if (klass->download == NULL) {
        g_warning(_("%s: `download' method not implemented for type `%s'"),
                  G_STRLOC, g_type_name(G_OBJECT_TYPE(isp)));
        return FALSE;
    }

    return klass->download(isp, address, buffer, nbytes);
}

/**
 * ntd_isp_download_file:
 * @isp: a #NtdIsp instance
 * @address: the address where starting the download
 * @file: the binary file to download
 * @error: return location for a #GError, or %NULL
 *
 * A convenient wrapper around ntd_isp_download() that downloads
 * the content of @file into the @isp chip, starting from @address.
 * The eventual I/O error will be reported in @error.
 *
 * @address can be %NULL, in which case the starting address of the
 * flash memory will be used.
 *
 * @file can be %NULL, in which case the standard input will be used.
 *
 * Returns: %TRUE on success, %FALSE on error
 **/
gboolean
ntd_isp_download_file(NtdIsp *isp, const gchar *address,
                      const gchar *file, GError **error)
{
    gchar *buffer;
    gsize nbytes;
    gboolean done;

    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
    g_return_val_if_fail(NTD_IS_ISP(isp), FALSE);

    buffer = NULL;
    nbytes = 0;

    if (file == NULL) {
        size_t size;

        /* Read stdin at CHUNK_SIZE increments, up to feof() or ferror() */
        clearerr(stdin);
        do {
            nbytes += CHUNK_SIZE;
            buffer = g_realloc(buffer, nbytes);
            size = fread(buffer + nbytes - CHUNK_SIZE, CHUNK_SIZE, 1, stdin);
        } while (size == CHUNK_SIZE);
        done = ! ferror(stdin);

        if (done) {
            /* Remove padding bytes if last read was partial */
            nbytes -= CHUNK_SIZE - size;
        } else {
            g_set_error_literal(error, G_FILE_ERROR, G_FILE_ERROR_IO,
                                _("Failed to read from standard input"));
            clearerr(stdin);
        }
    } else {
        done = g_file_get_contents(file, &buffer, &nbytes, error);
    }

    if (done && nbytes > 0) {
        done = ntd_isp_download(isp, address, buffer, nbytes);
        if (! done && error != NULL) {
            /* Set error to the device error or to a generic error */
            NtdIspPrivate *data;
            GError *device_error;

            data = isp->data;
            device_error = ntd_device_get_error(NTD_DEVICE(data->serial));

            if (device_error != NULL) {
                *error = g_error_copy(device_error);
            } else {
                const gchar *message = ntd_isp_get_message(isp);
                if (message == NULL || message[0] == '\0')
                    message = _("Failed to download");
                g_set_error_literal(error, G_FILE_ERROR, G_FILE_ERROR_IO,
                                    message);
            }
        }
    }

    g_free(buffer);
    return done;
}

/**
 * ntd_isp_dump:
 * @isp: a #NtdIsp instance
 * @address: the address where starting to get
 * @buffer: location to store the data got, use g_free()
 *          to free the returned buffer
 * @nbytes: either specify how many bytes to get and the location
 *          to store the returned buffer size
 *
 * Dumps, that is move from the target @isp (the embedded device) to
 * the host (PC), the specified portion of memory.
 *
 * @address can be %NULL, in which case the starting address of the
 * flash memory will be used.
 *
 * @nbytes specifies the size of the memory to dump. If initialized
 * to 0 or if @nbytes is %NULL the flash memory size will be used.
 *
 * The effective size of the memory dumped will be stored in @nbytes,
 * if not %NULL.
 *
 * Returns: %TRUE on success, %FALSE on error
 **/
gboolean
ntd_isp_dump(NtdIsp *isp, const gchar *address,
             gpointer *buffer, gsize *nbytes)
{
    NtdIspClass *klass;

    g_return_val_if_fail(NTD_IS_ISP(isp), FALSE);
    g_return_val_if_fail(buffer != NULL, FALSE);

    /* This method must be defined by the ISP implementation */
    klass = NTD_ISP_GET_CLASS(isp);
    if (klass->dump == NULL) {
        g_warning(_("%s: `dump' method not implemented for type `%s'"),
                  G_STRLOC, g_type_name(G_OBJECT_TYPE(isp)));
        return FALSE;
    }

    return klass->dump(isp, address, buffer, nbytes);
}

/**
 * ntd_isp_dump_file:
 * @isp: a #NtdIsp instance
 * @address: the address where starting the dump
 * @nbytes: size of the portion to dump, in bytes
 * @file: where to store the binary data
 * @error: return location for a #GError, or %NULL
 *
 * A convenient wrapper around ntd_isp_dump() that dumps @nbytes
 * bytes of memory starting at @address on the target @isp directly
 * to @file, providing optional error checking through @error.
 *
 * @address can be %NULL, in which case the starting address of the
 * flash memory will be used.
 *
 * If @nbytes is 0, the portion of memory starting at @address up
 * to the end of the flash will be dumped.
 *
 * @file can be %NULL, in which case the standard output will be used.
 *
 * Returns: %TRUE on success, %FALSE on error
 **/
gboolean
ntd_isp_dump_file(NtdIsp *isp, const gchar *address, gsize nbytes,
                  const gchar *file, GError **error)
{
    gpointer buffer;
    gboolean done;

    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
    g_return_val_if_fail(NTD_IS_ISP(isp), FALSE);

    if (! ntd_isp_dump(isp, address, &buffer, &nbytes)) {
        g_set_error(error, NTD_SYSTEM_ERROR, NTD_SYSTEM_ERROR_FAILED,
                    _("Failed to dump memory"));
        return FALSE;
    }

    if (file == NULL) {
        fwrite(buffer, nbytes, 1, stdout);
        done = TRUE;
    } else {
        done = g_file_set_contents(file, buffer, nbytes, error);
    }

    g_free(buffer);
    return done;
}

/**
 * ntd_isp_verify:
 * @isp:     an #NtdIsp instance
 * @address: the address where starting to verify
 * @buffer:  location containing the original data
 * @nbytes:  size of @buffer
 *
 * Verifies that @nbytes bytes at @address address of the device
 * bound to @isp are the same as @nbytes bytes at @buffer.
 *
 * This is conceptually equivalent to the following code:
 *
 * |[
 * ntd_isp_dump(isp, address, &real_buffer, &nbytes);
 * return memcmp(real_buffer, buffer, nbytes) == 0;
 * ]|
 *
 * Returns: %TRUE on success, %FALSE on error
 **/
gboolean
ntd_isp_verify(NtdIsp *isp, const gchar *address,
               gconstpointer buffer, gsize nbytes)
{
    gpointer real_buffer;
    gsize real_nbytes;
    gboolean success;

    real_nbytes = nbytes;
    if (! ntd_isp_dump(isp, address, &real_buffer, &real_nbytes))
        return FALSE;

    success = nbytes == real_nbytes &&
        memcmp(buffer, real_buffer, nbytes) == 0;

    g_free(real_buffer);
    return success;
}

/**
 * ntd_isp_verify_file:
 * @isp:     an #NtdIsp instance
 * @address: the address where starting the dump
 * @file:    where to store the binary data
 * @error:   return location for a #GError, or %NULL
 *
 * A convenient wrapper around ntd_isp_verify() that verifies that
 * the content of @file will match the bytes starting at @address
 * on the target @isp, providing optional @error checking.
 *
 * @address can be %NULL, in which case the starting address of the
 * flash memory will be used.
 *
 * Returns: %TRUE on success, %FALSE on error
 **/
gboolean
ntd_isp_verify_file(NtdIsp *isp, const gchar *address,
                    const gchar *file, GError **error)
{
    gchar *buffer;
    gsize nbytes;
    gboolean success;

    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

    if (! g_file_get_contents(file, &buffer, &nbytes, error))
        return FALSE;

    success = ntd_isp_verify(isp, address, buffer, nbytes);

    g_free(buffer);
    return success;
}

/**
 * ntd_isp_run:
 * @isp: a #NtdIsp instance
 * @address: from where the program must be run
 *
 * Puts the @isp in run mode, if possible, running the code at
 * a specific address. @address can be %NULL, in which case the
 * starting address of the flash memory will be used.
 *
 * Returns: %TRUE on success, %FALSE on error
 **/
gboolean
ntd_isp_run(NtdIsp *isp, const gchar *address)
{
    NtdIspClass *klass;

    g_return_val_if_fail(NTD_IS_ISP(isp), FALSE);

    /* This method must be defined by the ISP implementation */
    klass = NTD_ISP_GET_CLASS(isp);
    if (klass->run == NULL) {
        g_warning(_("%s: `run' method not implemented for type `%s'"),
                  G_STRLOC, g_type_name(G_OBJECT_TYPE(isp)));
        return FALSE;
    }

    return klass->run(isp, address);
}
