/* ntdisp - NtD In System Programming.
 *
 * Copyright (C) 2012  Nicola Fontana <ntd at entidi.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */


#include <glib-object.h>
#include <string.h>

#include "../config.h"
#include "ntdisp-utils.h"
#include "ntdisp-enums.h"
#include "ntdisp-builtins.h"
#include "ntd-device.h"
#include "ntd-serial.h"
#include "ntd-isp.h"
#include "ntd-lpc.h"

#define KB              *1024


typedef struct {
    const gchar *name;
    gulong       buffer_address;
    guint        buffer_size;
    guint        chunk_size;
    guint32      device_id;
    const gchar *flash_table;
} NtdLpcConfig;


/* The original data has been adapted from the lpc21isp source code:
 * http://sourceforge.net/projects/lpc21isp/
 *
 * By default, 768 bytes are removed from buffer-size.
 * This because 512 bytes are reserved by the ISP in low memory
 * (hence buffer-address is usually offseted by 0x200 from the
 * SRAM memory starting address) while up to 256 bytes can be
 * used by the ISP stack in the high memory.
 *
 * These assumptions have been made by reading the ISP documentation
 * for devices I used (LPC2141 and LPC2378) but can be possibly
 * wrong for other devices. Just drop me a line if you find such cases.
 */
static const NtdLpcConfig db[] = {
    /* name             buffer-address  buffer-size  chunk-size device-id   flash-table */
    { "LPC1111.../101", 0x10000300UL,    2 KB - 768,    1 KB,   0x041E502B, "\x04\x04" },
    { "LPC1111.../201", 0x10000300UL,    4 KB - 768,    2 KB,   0x0416502B, "\x04\x04" },
    { "LPC1112.../101", 0x10000300UL,    2 KB - 768,    1 KB,   0x042D502B, "\x04\x04\x04\x04" },
    { "LPC1112.../201", 0x10000300UL,    4 KB - 768,    2 KB,   0x0425502B, "\x04\x04\x04\x04" },
    { "LPC1113.../201", 0x10000300UL,    4 KB - 768,    2 KB,   0x0434502B, "\x04\x04\x04\x04\x04\x04" },
    { "LPC1113.../301", 0x10000300UL,    8 KB - 768,    4 KB,   0x0434102B, "\x04\x04\x04\x04\x04\x04" },
    { "LPC1114.../201", 0x10000300UL,    4 KB - 768,    2 KB,   0x0444502B, "\x04\x04\x04\x04\x04\x04\x04\x04" },
    { "LPC1114.../301", 0x10000300UL,    8 KB - 768,    4 KB,   0x0444102B, "\x04\x04\x04\x04\x04\x04\x04\x04" },
    { "LPC1311",        0x10000300UL,    4 KB - 768,    2 KB,   0x2C42502B, "\x04\x04" },
    { "LPC1313",        0x10000300UL,    8 KB - 768,    4 KB,   0x2C40102B, "\x04\x04\x04\x04\x04\x04\x04\x04" },
    { "LPC1342",        0x10000300UL,    4 KB - 768,    2 KB,   0x3D01402B, "\x04\x04" },
    { "LPC1343",        0x10000300UL,    8 KB - 768,    4 KB,   0x3D00002B, "\x04\x04\x04\x04\x04\x04\x04\x04" },
    { "LPC1751",        0x10000200UL,    8 KB - 768,    4 KB,   0x25001118, "\x04\x04\x04\x04\x04\x04\x04\x04" },
    { "LPC1752",        0x10000200UL,   16 KB - 768,    4 KB,   0x25001121, "\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04" },
    { "LPC1754",        0x10000200UL,   32 KB - 768,    4 KB,   0x25011722, "\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20" },
    { "LPC1756",        0x10000200UL,   32 KB - 768,    4 KB,   0x25011723, "\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20" },
    { "LPC1758",        0x10000200UL,   64 KB - 768,    4 KB,   0x25013F37, "\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20" },
    { "LPC1759",        0x10000200UL,   64 KB - 768,    4 KB,   0x25113737, "\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20" },
    { "LPC1764",        0x10000200UL,   32 KB - 768,    4 KB,   0x26011922, "\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20" },
    { "LPC1765",        0x10000200UL,   64 KB - 768,    4 KB,   0x26013733, "\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20" },
    { "LPC1766",        0x10000200UL,   64 KB - 768,    4 KB,   0x26013F33, "\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20" },
    { "LPC1767",        0x10000200UL,   64 KB - 768,    4 KB,   0x26012837, "\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20" },
    { "LPC1768",        0x10000200UL,   64 KB - 768,    4 KB,   0x26013F37, "\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20" },
    { "LPC1769",        0x10000200UL,   64 KB - 768,    4 KB,   0x26113F37, "\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20" },
    { "LPC2103",        0x40000200UL,    8 KB - 768,    4 KB,   0x0004FF11, "\x04\x04\x04\x04\x04\x04\x04\x04" },
    { "LPC2104",        0x40000200UL,   16 KB - 768,    8 KB,   0xFFF0FF12, "\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08" },
    { "LPC2105",        0x40000200UL,   32 KB - 768,    8 KB,   0xFFF0FF22, "\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08" },
    { "LPC2106",        0x40000200UL,   64 KB - 768,    8 KB,   0xFFF0FF32, "\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08" },
    { "LPC2109",        0x40000200UL,    8 KB - 768,    4 KB,   0x0201FF01, "\x08\x08\x08\x08\x08\x08\x08\x08" },
    { "LPC2114",        0x40000200UL,   16 KB - 768,    8 KB,   0x0101FF12, "\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08" },
    { "LPC2119",        0x40000200UL,   16 KB - 768,    8 KB,   0x0201FF12, "\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08" },
    { "LPC2124",        0x40000200UL,   16 KB - 768,    8 KB,   0x0101FF13, "\x08\x08\x08\x08\x08\x08\x08\x08\x40\x40\x08\x08\x08\x08\x08\x08\x08" },
    { "LPC2129",        0x40000200UL,   16 KB - 768,    8 KB,   0x0201FF13, "\x08\x08\x08\x08\x08\x08\x08\x08\x40\x40\x08\x08\x08\x08\x08\x08\x08" },
    { "LPC2131",        0x40000200UL,    8 KB - 768,    4 KB,   0x0002FF01, "\x04\x04\x04\x04\x04\x04\x04\x04" },
    { "LPC2132",        0x40000200UL,   16 KB - 768,    4 KB,   0x0002FF11, "\x04\x04\x04\x04\x04\x04\x04\x04\x20" },
    { "LPC2134",        0x40000200UL,   16 KB - 768,    4 KB,   0x0002FF12, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20" },
    { "LPC2136",        0x40000200UL,   32 KB - 768,    4 KB,   0x0002FF23, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20" },
    { "LPC2138",        0x40000200UL,   32 KB - 768,    4 KB,   0x0002FF25, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x04\x04\x04\x04\x04" },
    { "LPC2141",        0x40000200UL,    8 KB - 768,    4 KB,   0x0402FF01, "\x04\x04\x04\x04\x04\x04\x04\x04" },
    { "LPC2142",        0x40000200UL,   16 KB - 768,    4 KB,   0x0402FF11, "\x04\x04\x04\x04\x04\x04\x04\x04\x20" },
    { "LPC2144",        0x40000200UL,   16 KB - 768,    4 KB,   0x0402FF12, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20" },
    { "LPC2146",        0x40000200UL,   32 KB - 768,    4 KB,   0x0402FF23, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20" },
    { "LPC2148",        0x40000200UL,   32 KB - 768,    4 KB,   0x0402FF25, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x04\x04\x04\x04\x04" },
    { "LPC2194",        0x40000200UL,   16 KB - 768,    8 KB,   0x0301FF13, "\x08\x08\x08\x08\x08\x08\x08\x08\x40\x40\x08\x08\x08\x08\x08\x08\x08" },
    { "LPC2210",        0x40000200UL,   16 KB - 768,    8 KB,   0x0301FF12, "\x08\x08\x08\x08\x08\x08\x08\x08\x40\x40\x08\x08\x08\x08\x08\x08\x08" },
    { "LPC2212",        0x40000200UL,   16 KB - 768,    8 KB,   0x0401FF12, "\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08\x08" },
    { "LPC2214",        0x40000200UL,   16 KB - 768,    8 KB,   0x0601FF13, "\x08\x08\x08\x08\x08\x08\x08\x08\x40\x40\x08\x08\x08\x08\x08\x08\x08" },
    { "LPC2292",        0x40000200UL,   16 KB - 768,    8 KB,   0x0401FF13, "\x08\x08\x08\x08\x08\x08\x08\x08\x40\x40\x08\x08\x08\x08\x08\x08\x08" },
    { "LPC2294",        0x40000200UL,   16 KB - 768,    8 KB,   0x0501FF13, "\x08\x08\x08\x08\x08\x08\x08\x08\x40\x40\x08\x08\x08\x08\x08\x08\x08" },
    { "LPC2361",        0x40000200UL,   32 KB - 768,    4 KB,   0x00000000, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20" },
    { "LPC2364",        0x40000200UL,   32 KB - 768,    4 KB,   0x1600F902, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20" },
    { "LPC2365",        0x40000200UL,   32 KB - 768,    4 KB,   0x1600E823, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20" },
    { "LPC2366",        0x40000200UL,   32 KB - 768,    4 KB,   0x1600F923, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20" },
    { "LPC2367",        0x40000200UL,   32 KB - 768,    4 KB,   0x1600E825, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20" },
    { "LPC2368",        0x40000200UL,   32 KB - 768,    4 KB,   0x1600F925, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x04\x04\x04\x04\x04\x04" },
    { "LPC2377",        0x40000200UL,   32 KB - 768,    4 KB,   0x1700E825, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x04\x04\x04\x04\x04\x04" },
    { "LPC2378",        0x40000200UL,   32 KB - 768,    4 KB,   0x1700FD25, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x04\x04\x04\x04\x04\x04" },
    { "LPC2387",        0x40000200UL,   64 KB - 768,    4 KB,   0x1800F935, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x04\x04\x04\x04\x04\x04" },
    { "LPC2388",        0x40000200UL,   64 KB - 768,    4 KB,   0x1800FF35, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x04\x04\x04\x04\x04\x04" },
    { "LPC2458",        0x40000200UL,   98 KB - 768,    4 KB,   0x1500FF35, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x04\x04\x04\x04\x04\x04" },
    { "LPC2460",        0x40000200UL,   98 KB - 768,    4 KB,   0x1600FF30, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x04\x04\x04\x04\x04\x04" },
    { "LPC2468",        0x40000200UL,   98 KB - 768,    4 KB,   0x1600FF35, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x04\x04\x04\x04\x04\x04" },
    { "LPC2470",        0x40000200UL,   98 KB - 768,    4 KB,   0x1701FF30, "\x04\x04\x04\x04\x04\x04\x04\x04\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x04\x04\x04\x04\x04\x04" },
};


static void
ntd_lpc_config(NtdLpc *lpc, const NtdLpcConfig *config)
{
    gchar *tmp_str;

    g_object_set((GObject *) lpc,
                 "name",           config->name,
                 "buffer-address", config->buffer_address,
                 "buffer-size",    config->buffer_size,
                 "chunk-size",     config->chunk_size,
                 "flash-table",    config->flash_table,
                 NULL);

    tmp_str = ntd_escape(config->flash_table, strlen(config->flash_table));
    g_debug("Configuration: name = '%s', buffer-address = %lx, buffer-size = %u, chunk-size = %u, flash table = '%s'",
            config->name, config->buffer_address, config->buffer_size,
            config->chunk_size, tmp_str);
    g_free(tmp_str);
}


gboolean
ntd_lpc_config_by_device_id(NtdLpc *lpc, guint32 device_id)
{
    gint n;
    const NtdLpcConfig *config;

    for (n = 0; n < G_N_ELEMENTS(db); ++n) {
        config = db + n;
        if (config->device_id == device_id) {
            ntd_lpc_config(lpc, config);
            return TRUE;
        }
    }

    return FALSE;
}

gboolean
ntd_lpc_config_by_name(NtdLpc *lpc, const gchar *name)
{
    gint n;
    const NtdLpcConfig *config;

    for (n = 0; n < G_N_ELEMENTS(db); ++n) {
        config = db + n;
        if (g_strcmp0(config->name, name) == 0) {
            ntd_lpc_config(lpc, config);
            return TRUE;
        }
    }

    return FALSE;
}
