/* ntdisp - NtD In System Programming.
 *
 * Copyright (C) 2012  Nicola Fontana <ntd at entidi.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * SECTION:ntd-lpc
 * @short_description: LPC default bootloader protocol
 *
 * To download a program on the LPC the local SRAM must be used as
 * a mid-step. This complicates things a bit because when downloading
 * big programs more steps are needed. Furthermore, the available
 * RAM buffer can be smaller than a flash sector (!), so multiple
 * download on the same sector can be possibly needed.
 * Another complication is flash sectors don't necessarily have the
 * same size, so the #NtdLpc:flash-table property is provided.
 *
 * Hence, to be able to use it, the buffer must be configured
 * with ntd_lpc_set_buffer_address() and ntd_lpc_set_buffer_size()
 * and the mapping of the flash memory specified with
 * ntd_lpc_set_flash_table().
 *
 * Alternatively, if the device id is yet known by the program,
 * a call to ntd_isp_probe() will do that for you by directly
 * querying the device.
 **/

/**
 * NtdLpc:
 *
 * All fields are private and should not be used directly.
 * Use its public methods instead.
 **/


#include <glib-object.h>
#include <glib/gprintf.h>
#include <stdlib.h>
#include <string.h>

#include "../config.h"
#include "ntdisp-intl.h"
#include "ntdisp-utils.h"
#include "ntdisp-enums.h"
#include "ntdisp-builtins.h"
#include "ntd-device.h"
#include "ntd-serial.h"
#include "ntd-isp.h"
#include "ntd-lpc.h"
#include "ntd-lpc-private.h"


#define MAX_ATTEMPTS    5
#define PROGRESS(size,still)  ((gdouble) ((size) - (still)) / (size))
#define INVALID_ALIGNMENT(object, pspec, subject, value, boundary) \
    G_STMT_START { \
        GParamSpec *_pspec = (GParamSpec*) (pspec); \
        g_warning("%s: %s `%lu' not aligned on a %lu byte boundary for \"%s\" of type `%s' in `%s'", \
                  G_STRLOC, (subject), (gulong) (value), (gulong) (boundary), \
                  _pspec->name, g_type_name(G_PARAM_SPEC_TYPE(_pspec)), \
                  G_OBJECT_TYPE_NAME((GObject *) (object))); \
    } G_STMT_END


G_DEFINE_TYPE(NtdLpc, ntd_lpc, NTD_TYPE_ISP);

enum {
    PROP_0,
    PROP_NAME,
    PROP_BUFFER_ADDRESS,
    PROP_BUFFER_SIZE,
    PROP_CHUNK_SIZE,
    PROP_FLASH_TABLE,
    PROP_FREQUENCY
};


static void             _ntd_set_property       (GObject*       object,
                                                 guint          prop_id,
                                                 const GValue*  value,
                                                 GParamSpec*    pspec);
static void             _ntd_get_property       (GObject*       object,
                                                 guint          prop_id,
                                                 GValue*        value,
                                                 GParamSpec*    pspec);
static void             _ntd_set_defaults       (NtdIsp*        isp);
static const gchar*     _ntd_status_to_message  (gint           status);
static void             _ntd_connect            (NtdIsp*        isp);
static void             _ntd_disconnect         (NtdIsp*        isp);
static gboolean         _ntd_probe              (NtdIsp*        isp);
static gboolean         _ntd_wipe               (NtdIsp*        isp,
                                                 const gchar*   address,
                                                 gsize          nbytes);
static gboolean         _ntd_download           (NtdIsp*        isp,
                                                 const gchar*   address,
                                                 gconstpointer  buffer,
                                                 gsize          nbytes);
static gboolean         _ntd_dump               (NtdIsp*        isp,
                                                 const gchar*   address,
                                                 gpointer*      p_buffer,
                                                 gsize*         p_nbytes);
static gboolean         _ntd_run                (NtdIsp*        isp,
                                                 const gchar*   address);
static gboolean         _ntd_download_chunk     (NtdLpc*        lpc,
                                                 gulong         address,
                                                 gpointer*      buffer,
                                                 gsize          nbytes);
static gboolean         _ntd_real_run           (NtdLpc*        lpc,
                                                 gulong         address);
static gboolean         _ntd_encode             (NtdLpc*        lpc,
                                                 gconstpointer  data,
                                                 guint          nbytes,
                                                 gdouble        from,
                                                 gdouble        to);
static gpointer         _ntd_decode             (NtdLpc*        lpc,
                                                 guint          nbytes,
                                                 gdouble        from,
                                                 gdouble        to);
static guint            _ntd_flash              (NtdLpc*        lpc,
                                                 gulong         ram,
                                                 gulong         flash,
                                                 guint          size,
                                                 gdouble        from,
                                                 gdouble        to);
static void             _ntd_get_vt_code        (gulong         data_address,
                                                 gpointer*      buffer,
                                                 gsize*         size);
static gboolean         _ntd_inject_vt_code     (NtdLpc*        lpc,
                                                 gulong         data_address,
                                                 gulong         code_address);
static gpointer         _ntd_get_vt             (NtdLpc*        lpc);


static void
ntd_lpc_class_init(NtdLpcClass *klass)
{
    GObjectClass *gobject_class;
    NtdIspClass *isp_class;
    GParamSpec *param;

    gobject_class = (GObjectClass *) klass;
    isp_class = (NtdIspClass *) klass;

    g_type_class_add_private(klass, sizeof(NtdLpcPrivate));

    gobject_class->set_property = _ntd_set_property;
    gobject_class->get_property = _ntd_get_property;

    isp_class->set_defaults = _ntd_set_defaults;
    isp_class->status_to_message = _ntd_status_to_message;
    isp_class->connect = _ntd_connect;
    isp_class->disconnect = _ntd_disconnect;
    isp_class->probe = _ntd_probe;
    isp_class->wipe = _ntd_wipe;
    isp_class->download = _ntd_download;
    isp_class->dump = _ntd_dump;
    isp_class->run = _ntd_run;

    param = g_param_spec_string("name",
                                P_("Name"),
                                P_("The informal name of this chip"),
                                NULL,
                                G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_NAME, param);

    param = g_param_spec_ulong("buffer-address",
                               P_("Buffer starting address"),
                               P_("Starting address (usually in the SRAM space) of the biggest contiguous buffer available for downloading"),
                               0, G_MAXULONG, 0,
                               G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_BUFFER_ADDRESS, param);

    param = g_param_spec_uint("buffer-size",
                              P_("Buffer size"),
                              P_("Size, in bytes, of the buffer starting at buffer-address"),
                              0, G_MAXUINT, 0,
                              G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_BUFFER_SIZE, param);

    param = g_param_spec_uint("chunk-size",
                              P_("Chunk Size"),
                              P_("Size, in bytes, of the largest chunk that can be copied to flash in a single command"),
                              0, G_MAXUINT, 0,
                              G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_CHUNK_SIZE, param);

    param = g_param_spec_string("flash-table",
                                P_("Flash Table"),
                                P_("A \\0 terminated array of bytes specifying the size, in KiB, of every flash sector"),
                                NULL,
                                G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_FLASH_TABLE, param);

    param = g_param_spec_uint("frequency",
                              P_("Frequency"),
                              P_("Crystal frequency, in kHz, at which the part is running"),
                              0, G_MAXUINT, 14746,
                              G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_FREQUENCY, param);
}

static void
ntd_lpc_init(NtdLpc *lpc)
{
    NtdLpcPrivate *data = G_TYPE_INSTANCE_GET_PRIVATE(lpc,
                                                      NTD_TYPE_LPC,
                                                      NtdLpcPrivate);

    data->name = NULL;
    data->buffer_address = 0;
    data->buffer_size = 0;
    data->flash_table = NULL;
    data->frequency = 14746;

    lpc->data = data;
}

static void
_ntd_set_property(GObject *object, guint prop_id,
                  const GValue *value, GParamSpec *pspec)
{
    NtdLpcPrivate *data;
    gulong address;
    guint size;

    data = ((NtdLpc *) object)->data;

    switch (prop_id) {
    case PROP_NAME:
        g_free(data->name);
        data->name = g_value_dup_string(value);
        break;
    case PROP_BUFFER_ADDRESS:
        address = g_value_get_ulong(value);
        if ((address & 3) > 0) {
            INVALID_ALIGNMENT(object, pspec, "address", address, 4);
        } else {
            data->buffer_address = address;
        }
        break;
    case PROP_BUFFER_SIZE:
        data->buffer_size = g_value_get_uint(value);
        break;
    case PROP_CHUNK_SIZE:
        size = g_value_get_uint(value);
        if ((size & 0xFF) > 0) {
            INVALID_ALIGNMENT(object, pspec, "size", size, 256);
        } else {
            data->chunk_size = size;
        }
        break;
    case PROP_FLASH_TABLE:
        g_free(data->flash_table);
        data->flash_table = g_value_dup_string(value);
        break;
    case PROP_FREQUENCY:
        data->frequency = g_value_get_uint(value);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
_ntd_get_property(GObject *object, guint prop_id,
                    GValue *value, GParamSpec *pspec)
{
    NtdLpcPrivate *data;

    data = ((NtdLpc *) object)->data;

    switch (prop_id) {
    case PROP_NAME:
        g_value_set_string(value, data->name);
        break;
    case PROP_BUFFER_ADDRESS:
        g_value_set_ulong(value, data->buffer_address);
        break;
    case PROP_BUFFER_SIZE:
        g_value_set_uint(value, data->buffer_size);
        break;
    case PROP_CHUNK_SIZE:
        g_value_set_uint(value, data->chunk_size);
        break;
    case PROP_FLASH_TABLE:
        g_value_set_string(value, (gchar *) data->flash_table);
        break;
    case PROP_FREQUENCY:
        g_value_set_uint(value, data->frequency);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}


/**
 * ntd_lpc_new:
 *
 * Creates a new generic #NtdLpc instance.
 *
 * Returns: the newly created object
 */
NtdLpc *
ntd_lpc_new(void)
{
    return (NtdLpc *) g_object_new(NTD_TYPE_LPC, NULL);
}

/**
 * ntd_lpc_new_with_device:
 * @device: path to the device to use
 *
 * Creates a new default #NtdLpc instance and binds it to a
 * #NtdSerial device automatically configured with sensible defaults.
 *
 * Returns: the newly created object
 */
NtdLpc *
ntd_lpc_new_with_device(const gchar *device)
{
    NtdLpc *lpc = ntd_lpc_new();

    ntd_isp_set_device(NTD_ISP(lpc), device);
    return lpc;
}

/**
 * ntd_lpc_set_name:
 * @lpc: a #NtdLpc
 * @name: device name (such as "LPC2148")
 *
 * Sets a new name on @lpc. If @name is found in the
 * internal (hard-coded) database, @lpc is also
 * configured with the default parameters for that
 * device.
 **/
void
ntd_lpc_set_name(NtdLpc *lpc, const gchar *name)
{
    g_return_if_fail(NTD_IS_LPC(lpc));

    if (! ntd_lpc_config_by_name(lpc, name)) {
        /* "name" is set by ntd_lpc_config_by_name() only if successful */
        g_object_set(lpc, "name", name, NULL);
    }
}

/**
 * ntd_lpc_get_name:
 * @lpc: a #NtdLpc
 *
 * Gets the name bound to @lpc. The string is internally owned
 * and must not be freed or modified.
 *
 * Returns: the name bound to this @lpc instance
 **/
const gchar *
ntd_lpc_get_name(NtdLpc *lpc)
{
    NtdLpcPrivate *data;

    g_return_val_if_fail(NTD_IS_LPC(lpc), NULL);

    data = lpc->data;

    return data->name;
}

/**
 * ntd_lpc_set_buffer_address:
 * @lpc: a #NtdLpc instance
 * @address: the new address
 *
 * Sets the starting address of the biggest buffer available
 * for downloading.
 *
 * While downloading on LPC based devices, data cannot be directly
 * sent to the flash. It must be instead written to a RAM space before
 * (or to whatever available on the @lpc device) by using the ISP write
 * command ("W to size") and then moved to flash with the copy to
 * flash command ("C to from size").
 *
 * This command, in conjunction with ntd_lpc_set_buffer_size(), is
 * used to identify the address and extension of the temporary buffer
 * to use as destination of the "W" command (and, consequently, the
 * source of the "C" command).
 *
 * Some things to keep in mind:
 *
 * - the buffer must be contiguous;
 * - all the memory must be addressable by the "W" command;
 * - check the "C" command documentation to know if some special
 *   restrictions apply to some RAM memory;
 * - the first 512 bytes of the base RAM memory should be
 *   considered reserved by the ISP program itself.
 * - the last bunch of bytes, usually up to 256, can be possibly
 *   used by the ISP as stack for its commands.
 *
 * An example: although LPC2148 is able to address 40 KB of RAM,
 * the last 8 KB are not contiguous and can be reserved for USB
 * purposes. This makes them unavailable for the "W" command.
 * Considering the first 512 bytes and the last 256 bytes are
 * reserved, a proper way to initialize the buffer for this
 * device would be:
 *
 * |[
 * ntd_lpc_set_buffer_address(lpc, 0x40000000UL + 512);
 * ntd_lpc_set_buffer_size(lpc, 40*1024 - 8*1024 - 512 - 256);
 * ]|
 *
 * To be able to use ntd_isp_download(), the flash memory map must
 * be known. This would require a call to ntd_lpc_set_flash_table().
 *
 * ntd_isp_probe() does this for you by querying the device id
 * with the "J" command and setting the properties stored in an
 * hard-coded list.
 **/
void
ntd_lpc_set_buffer_address(NtdLpc *lpc, gulong address)
{
    g_return_if_fail(NTD_IS_LPC(lpc));
    g_object_set(lpc, "buffer-address", address, NULL);
}

/**
 * ntd_lpc_get_buffer_address:
 * @lpc: a #NtdLpc instance
 *
 * Gets the base ram address of @lpc.
 *
 * Returns: the base RAM address
 **/
gulong
ntd_lpc_get_buffer_address(NtdLpc *lpc)
{
    NtdLpcPrivate *data;

    g_return_val_if_fail(NTD_IS_LPC(lpc), 0);

    data = lpc->data;

    return data->buffer_address;
}

/**
 * ntd_lpc_set_buffer_size:
 * @lpc: a #NtdLpc instance
 * @size: size, in bytes, of the biggest RAM buffer available
 *
 * Sets the new size of biggest contiguous buffer available for
 * @lpc to @size.
 *
 * This value is used by write operations ("W ..."): check
 * ntd_lpc_set_buffer_address() for details.
 **/
void
ntd_lpc_set_buffer_size(NtdLpc *lpc, guint size)
{
    g_return_if_fail(NTD_IS_LPC(lpc));
    g_object_set(lpc, "buffer-size", size, NULL);
}

/**
 * ntd_lpc_get_buffer_size:
 * @lpc: a #NtdLpc instance
 *
 * Gets the size, in bytes, of the biggest buffer of @lpc.
 * Check ntd_lpc_set_buffer_address() for details on what this
 * buffer is intended to be.
 *
 * Returns: the buffer size.
 **/
guint
ntd_lpc_get_buffer_size(NtdLpc *lpc)
{
    NtdLpcPrivate *data;

    g_return_val_if_fail(NTD_IS_LPC(lpc), 0);

    data = lpc->data;

    return data->buffer_size;
}

/**
 * ntd_lpc_set_chunk_size:
 * @lpc: a #NtdLpc instance
 * @chunk_size: the new address
 *
 * Sets the chunk size of @lpc to @chunk_size.
 *
 * This option specifies the width (in bytes) of the largest block
 * that can be moved with a single operation by this chip.
 **/
void
ntd_lpc_set_chunk_size(NtdLpc *lpc, guint chunk_size)
{
    g_return_if_fail(NTD_IS_LPC(lpc));
    g_object_set(lpc, "chunk-size", chunk_size, NULL);
}

/**
 * ntd_lpc_get_chunk_size:
 * @lpc: a #NtdLpc instance
 *
 * Gets the chunk size of @lpc.
 *
 * Returns: the chunk size
 **/
guint
ntd_lpc_get_chunk_size(NtdLpc *lpc)
{
    NtdLpcPrivate *data;

    g_return_val_if_fail(NTD_IS_LPC(lpc), 0);

    data = lpc->data;

    return data->chunk_size;
}

/**
 * ntd_lpc_set_frequency:
 * @lpc: a #NtdLpc instance
 * @frequency: the new frequency
 *
 * Sets the frequency of @lpc to @frequency.
 *
 * The frequency is required (or only reccomended?) by the
 * communication protocol and its value will be sent (if not 0)
 * after the synchronization step. No idea what it will be used
 * for, though...
 **/
void
ntd_lpc_set_frequency(NtdLpc *lpc, guint frequency)
{
    g_return_if_fail(NTD_IS_LPC(lpc));
    g_object_set(lpc, "frequency", frequency, NULL);
}

/**
 * ntd_lpc_get_frequency:
 * @lpc: a #NtdLpc instance
 *
 * Gets the current set frequency on @lpc.
 *
 * Returns: the frequency, in kHz
 **/
guint
ntd_lpc_get_frequency(NtdLpc *lpc)
{
    NtdLpcPrivate *data;

    g_return_val_if_fail(NTD_IS_LPC(lpc), 0);

    data = lpc->data;

    return data->frequency;
}

/**
 * ntd_lpc_set_flash_table:
 * @lpc: a #NtdLpc
 * @flash_table: name (informational)
 *
 * Sets a new flash table on @lpc.
 *
 * The flash table is a '\0' terminated array of char (such as normal
 * ASCII text) but where every character specifies the size, in KiB,
 * of the sector at that index. For example, an array such as
 * "\x04\x08\x08" specifies that the @lpc chip has 3 flash sectors,
 * the first one of 4KiB and the others of 8KiB.
 **/
void
ntd_lpc_set_flash_table(NtdLpc *lpc, const gchar *flash_table)
{
    g_return_if_fail(NTD_IS_LPC(lpc));
    g_object_set(lpc, "flash-table", flash_table, NULL);
}

/**
 * ntd_lpc_get_flash_table:
 * @lpc: a #NtdLpc
 *
 * Gets the flash table of @lpc. The array is internally owned
 * and must not be freed or modified.
 *
 * Returns: the flash table array
 **/
const gchar *
ntd_lpc_get_flash_table(NtdLpc *lpc)
{
    NtdLpcPrivate *data;

    g_return_val_if_fail(NTD_IS_LPC(lpc), NULL);

    data = lpc->data;

    return (gchar *) data->flash_table;
}

/**
 * ntd_lpc_get_flash_size:
 * @lpc: a #NtdLpc
 *
 * Gets the flash size of @lpc.
 *
 * Returns: the flash size, in bytes.
 **/
gsize
ntd_lpc_get_flash_size(NtdLpc *lpc)
{
    NtdLpcPrivate *data;
    const gchar *ptr;
    gsize kb;

    g_return_val_if_fail(NTD_IS_LPC(lpc), 0);

    data = lpc->data;
    kb = 0;

    for (ptr = (const gchar *) data->flash_table; ptr && *ptr; ++ptr)
        kb += *ptr;

    return kb * 1024;
}

gboolean
ntd_lpc_command(NtdLpc *lpc, const gchar *format, ...)
{
    va_list var_args;
    gboolean done;

    va_start(var_args, format);
    done = ntd_lpc_command_valist(lpc, format, var_args);
    va_end(var_args);

    return done;
}

gboolean
ntd_lpc_command_valist(NtdLpc *lpc, const gchar *format, va_list var_args)
{
    NtdDevice *device;
    gchar *buffer, *line, *response;
    //gchar **lines, **p_line, *line, *string;

    g_return_val_if_fail(NTD_IS_LPC(lpc), FALSE);

    if (format == NULL)
        return TRUE;

    device = (NtdDevice *) ntd_isp_get_serial((NtdIsp *) lpc);
    g_return_val_if_fail(NTD_IS_SERIAL(device), FALSE);

    g_vasprintf(&buffer, format, var_args);

    for (line = strtok(buffer, "\r\n"); line; line = strtok(NULL, "\r\n")) {
        /* Send one line, always using \r\n as terminator
         * to avoid inconsistencies between ISP implementation */
        if (! ntd_device_printf(device, "%s\r\n", line))
            break;

        response = ntd_device_read_until(device, '\n');
        if (response == NULL)
            break;

        /* Strip trailing "\r\n" before comparing */
        g_strdelimit(response, "\r\n", '\0');

        if (g_strcmp0(line, response) != 0) {
            g_debug("Match failed (sent '%s', received '%s')", line, response);
            g_free(response);
            break;
        }
        g_free(response);
    }

    g_free(buffer);
    return line == NULL;
}

gboolean
ntd_lpc_encode(NtdLpc *lpc, gconstpointer data, guint nbytes)
{
    g_return_val_if_fail(NTD_IS_LPC(lpc), FALSE);

    return _ntd_encode(lpc, data, nbytes, 0, 0);
}

gpointer
ntd_lpc_decode(NtdLpc *lpc, guint nbytes)
{
    g_return_val_if_fail(NTD_IS_LPC(lpc), NULL);
    g_return_val_if_fail(nbytes == 0, NULL);

    return _ntd_decode(lpc, nbytes, 0, 0);
}

gchar *
ntd_lpc_read_line(NtdLpc *lpc)
{
    NtdIsp *isp;
    NtdDevice *device;

    g_return_val_if_fail(NTD_IS_LPC(lpc), NULL);

    isp = (NtdIsp *) lpc;
    device = (NtdDevice *) ntd_isp_get_serial(isp);
    g_return_val_if_fail(NTD_IS_SERIAL(device), NULL);

    return ntd_device_read_until(device, '\n');
}


gboolean
ntd_lpc_answered_OK(NtdLpc *lpc)
{
    gchar *line;
    gboolean done;

    line = ntd_lpc_read_line(lpc);
    if (line == NULL)
        return FALSE;

    done = strncmp(line, "OK", 2) == 0;
    g_free(line);

    return done;
}

gboolean
ntd_lpc_answered_0(NtdLpc *lpc)
{
    gchar *line;
    gboolean is_0;
    gint status;

    line = ntd_lpc_read_line(lpc);
    if (line == NULL)
        return FALSE;

    is_0 = line[0] == '0';
    status = atoi(line);
    g_free(line);

    if (status == 0 && !is_0)
        status = -1;

    ntd_isp_set_status((NtdIsp *) lpc, status);
    return status == 0;
}

gboolean
ntd_lpc_answered_data(NtdLpc *lpc, gchar **p_data, gulong *p_checksum)
{
    GString *string;
    gchar *line;

    string = g_string_new("");

    while ((line = ntd_lpc_read_line(lpc)) != NULL) {
        if (strspn(line, "0123456789\r\n") == strlen(line)) {
            /* This is a checksum line */
            if (p_checksum)
                *p_checksum = atoi(line);
            g_free(line);
            if (p_data)
                *p_data = g_string_free(string, FALSE);
            else
                g_string_free(string, TRUE);
            return TRUE;
        }
        g_string_append(string, line);
        g_free(line);
    }

    g_string_free(string, TRUE);
    return FALSE;
}

/**
 * ntd_lpc_map_address:
 * @lpc: a #NtdLpc instance
 * @address: the address to map
 *
 * Converts @address to a sector number basing the process on the
 * #NtdLpc:flash-table property of @lpc.
 *
 * Returns: the sector number @address belongs to
 *          or -1 if out of range.
 **/
gint
ntd_lpc_map_address(NtdLpc *lpc, gulong address)
{
    NtdLpcPrivate *data;
    guchar *ptr;
    gulong end_address;

    g_return_val_if_fail(NTD_IS_LPC(lpc), -1);

    data = lpc->data;
    g_return_val_if_fail(data->flash_table != NULL, -1);

    end_address = 0;
    for (ptr = (guchar *) data->flash_table; *ptr; ++ptr) {
        end_address += (*ptr) * 1024;
        if (end_address > address)
            return (gchar *) ptr - data->flash_table;
    }

    return -1;
}

/**
 * ntd_lpc_ram_to_flash:
 * @lpc: a #NtdLpc
 * @ram: source address (must be 4 byte boundary)
 * @flash: destination address (must be 256 byte boundary)
 * @size: size of the buffer to copy
 *
 * Copies @size bytes from @ram to @flash.
 *
 * @size should be greather or equal than #NtdLpc:chunk-size,
 * if not it is adjusted accordingly. Furthermore, @size will
 * be forcibly rounded up to a 256 boundary value.
 *
 * Returns: the number of bytes effectively copied, or 0 on errors.
 **/
guint
ntd_lpc_ram_to_flash(NtdLpc *lpc, gulong ram, gulong flash, guint size)
{
    g_return_val_if_fail(NTD_IS_LPC(lpc), 0);
    g_return_val_if_fail((ram & 3) == 0, 0);

    return _ntd_flash(lpc, ram, flash, size, 0, 0);
}


static void
_ntd_set_defaults(NtdIsp *isp)
{
    NtdSerial *serial = ntd_isp_get_serial(isp);

    ntd_serial_set(serial, NTD_BAUD_38400, NTD_PARITY_NONE,
                   NTD_DATA_BITS_8, NTD_STOP_BITS_1);
    ntd_serial_set_software_flow(serial, NTD_SOFTWARE_FLOW_XON_XOFF);
}

static const gchar *
_ntd_status_to_message(gint status)
{
    switch (status) {
    case 0:
        return "";
    case 1:
        return "INVALID_COMMAND";
    case 2:
        return "SRC_ADDR_ERROR: source address is not on word boundary";
    case 3:
        return "DST_ADDR_ERROR: destination address is not on a correct boundary";
    case 4:
        return "SRC_ADDR_NOT_MAPPED: source address is not mapped in the memory map";
    case 5:
        return "DST_ADDR_NOT_MAPPED: destination address is not mapped in the memory map; count value is taken into consideration where applicable";
    case 6:
        return "COUNT_ERROR: byte count is not multiple of 4 or is not a permitted value";
    case 7:
        return "INVALID_SECTOR: sector number is invalid or end sector number is greater than start sector number";
    case 8:
        return "SECTOR_NOT_BLANK";
    case 9:
        return "SECTOR_NOT_PREPARED_FOR_WRITE_OPERATION: command to prepare sector for write operation was not executed";
    case 10:
        return "COMPARE_ERROR: source and destination data not equal";
    case 11:
        return "BUSY: flash programming hardware interface is busy";
    case 12:
        return "PARAM_ERROR: insufficient number of parameters or invalid parameter";
    case 13:
        return "ADDR_ERROR: address is not on word boundary";
    case 14:
        return "ADDR_NOT_MAPPED: address is not mapped in the memory map; count value is taken in to consideration where applicable";
    case  15:
        return "CMD_LOCKED";
    case 16:
        return "INVALID_CODE: unlock code is invalid";
    case 17:
        return "INVALID_BAUD_RATE: invalid baud rate setting";
    case 18:
        return "INVALID_STOP_BIT: invalid stop bit setting";
    case 19:
        return "CODE READ PROTECTION ENABLED";
    case 255:
        return NULL;
    default:
        return "Unkwnow error";
    }
}

static void
_ntd_connect(NtdIsp *isp)
{
    NtdLpc *lpc;
    NtdLpcPrivate *data;
    NtdDevice *device;
    gboolean done;
    gint n;

    lpc = (NtdLpc *) isp;
    data = lpc->data;
    device = (NtdDevice *) ntd_isp_get_serial(isp);
    g_return_if_fail(NTD_IS_SERIAL(device));
    done = FALSE;

    for (n = 0; !done; ++n) {
        gchar *response;

        if (n >= MAX_ATTEMPTS) {
            break;
        } else if (n) {
            ntd_device_close(device);
        }

        g_debug("Synchronization (attempt %d)", n);
        ntd_device_clear_error(device);

        ntd_device_open(device);
        if (ntd_device_in_error(device))
            continue;

        ntd_device_printf(device, "?");
        if (ntd_device_in_error(device))
            continue;

        response = ntd_device_read_until(device, '\n');
        if (response != NULL) {
            done = strncmp(response, "Synchronized", 12) == 0;
            g_free(response);
            response = NULL;
        }
    }
    if (!done)
        goto bailout;

    g_debug("Checking connection");
    done = ntd_lpc_command(lpc, "Synchronized") &&
           ntd_lpc_answered_OK(lpc);
    if (!done)
        goto bailout;

    if (data->frequency > 0) {
        g_debug("Setting up frequency");
        done = ntd_lpc_command(lpc, "%u", data->frequency) &&
               ntd_lpc_answered_OK(lpc);
        if (!done)
            goto bailout;
    }

    g_debug("Unlocking");
    done = ntd_lpc_command(lpc, "U 23130") &&
           ntd_lpc_answered_0(lpc);
    if (!done)
        goto bailout;

    g_debug("connect() successful");
    return;

bailout:
    g_debug("connect() failed");
    ntd_device_close(device);
    ntd_isp_set_status(isp, -1);
}

static void
_ntd_disconnect(NtdIsp *isp)
{
    NtdDevice *device;

    device = (NtdDevice *) ntd_isp_get_serial(isp);
    g_return_if_fail(NTD_IS_SERIAL(device));

    ntd_device_close(device);
}


/* List of typical buffer addresses on a per model basis */

#define BUFFER_ADDRESS_11XX 0x10000300UL
#define BUFFER_ADDRESS_13XX 0x10000300UL
#define BUFFER_ADDRESS_17XX 0x10000200UL
#define BUFFER_ADDRESS_2XXX 0x40000200UL

static gboolean
_ntd_probe(NtdIsp *isp)
{
    NtdLpc    *lpc;
    NtdDevice *device;
    gchar     *response;
    guint32    id;

    ntd_isp_connect(isp);
    if (ntd_isp_get_status(isp) != 0)
        return FALSE;

    g_debug("Querying device id");
    lpc = (NtdLpc *) isp;
    if (!ntd_lpc_command(lpc, "J") ||
        !ntd_lpc_answered_0(lpc))
        goto bailout;

    device = (NtdDevice *) ntd_isp_get_serial(isp);
    response = ntd_device_read_until(device, '\n');
    if (response == NULL)
        goto bailout;

    id = atoi(response);
    g_free(response);

    if (! ntd_lpc_config_by_device_id(lpc, id)) {
        g_debug("probe() failed: device id unknown (%lu)", (gulong) id);
        return FALSE;
    }

    g_debug("probe() successful");
    return TRUE;

bailout:
    g_debug("probe() failed");
    return FALSE;
}

static gboolean
_ntd_wipe(NtdIsp *isp, const gchar *address, gsize nbytes)
{
    NtdLpc *lpc;
    NtdLpcPrivate *data;
    gulong real_address;
    gint from_sector, to_sector;

    lpc = (NtdLpc *) isp;
    data = lpc->data;

    ntd_isp_connect(isp);
    if (!ntd_isp_is_connected(isp))
        return FALSE;

    g_return_val_if_fail(data->flash_table != NULL, FALSE);

    real_address = address == NULL ? 0 : atol(address);
    from_sector = ntd_lpc_map_address(lpc, real_address);

    if (nbytes == 0)
        to_sector = strlen(data->flash_table) - 1;
    else
        to_sector = ntd_lpc_map_address(lpc, real_address + nbytes);

    g_debug("Preparing sectors from %d to %d for writing", from_sector, to_sector);
    if (!ntd_lpc_command(lpc, "P %d %d", from_sector, to_sector) ||
        !ntd_lpc_answered_0(lpc))
        goto bailout;

    g_debug("Erasing sectors from %d to %d", from_sector, to_sector);
    if (!ntd_lpc_command(lpc, "E %d %d", from_sector, to_sector) ||
        !ntd_lpc_answered_0(lpc))
        goto bailout;

    g_debug("wipe() successful");
    return TRUE;

bailout:
    g_debug("wipe() failed");
    return FALSE;
}

static gboolean
_ntd_download(NtdIsp *isp, const gchar *address,
              gconstpointer buffer, gsize nbytes)
{
    NtdLpc *lpc;
    NtdLpcPrivate *data;
    const gchar *ptr;
    gulong buffer_address, flash_address;
    guint remaining, size, buffer_size;

    lpc = (NtdLpc *) isp;
    data = lpc->data;

    g_return_val_if_fail(data->buffer_size >= 256, FALSE);

    ntd_isp_connect(isp);
    if (!ntd_isp_is_connected(isp))
        return FALSE;

    /* Force nbytes to be a multiple of 4. The extra data bytes
     * will be junk but... who cares */
    if ((nbytes & 3) > 0)
        nbytes = (nbytes | 3) + 1;

    buffer_address = ntd_lpc_get_buffer_address(lpc);

    /* buffer_size must be on a 256 bytes boundary to deal with the rounding
     * of size done by the ntd_lpc_ram_to_flash() method. */
    buffer_size = data->buffer_size & ~0xFF;

    size = data->chunk_size;

    if (address == NULL || address[0] == '\0')
        flash_address = 0;
    else
        flash_address = atol(address);

    ptr = buffer;
    for (remaining = nbytes; remaining > 0; remaining -= size) {
        size = MIN(remaining, buffer_size);

        g_debug("Downloading a chunk of %u bytes", size);
        if (!ntd_lpc_command(lpc, "W %lu %u", buffer_address, size) ||
            !ntd_lpc_answered_0(lpc) ||
            !_ntd_encode(lpc, ptr, size,
                         PROGRESS(nbytes, remaining),
                         PROGRESS(nbytes, remaining - size * 9 / 10)))
            goto bailout;

        if (_ntd_flash(lpc, buffer_address, flash_address, size,
                         PROGRESS(nbytes, remaining - size * 9 / 10),
                         PROGRESS(nbytes, remaining - size )) == 0)
            return FALSE;

        flash_address += size;
        ptr += size;
    }

    ntd_isp_progress_notify(isp, PROGRESS(nbytes, 0));
    g_debug("download() successful");
    return TRUE;

bailout:
    g_debug("download() failed");
    return FALSE;
}

static gboolean
_ntd_dump(NtdIsp *isp, const gchar *address,
          gpointer *p_buffer, gsize *p_nbytes)
{
    NtdLpc *lpc;
    GString *string;
    gsize nbytes, isp_nbytes;
    gulong from, isp_from;
    gboolean error;

    lpc = (NtdLpc *) isp;

    ntd_isp_connect(isp);
    if (!ntd_isp_is_connected(isp))
        return FALSE;

    from = address == NULL ? 0 : atol(address);
    nbytes = p_nbytes ? *p_nbytes : 0;
    if (nbytes == 0) {
        /* If not specified, dump up to the end of the flash memory */
        nbytes = ntd_lpc_get_flash_size(lpc);
        g_return_val_if_fail(nbytes > from, FALSE);
        nbytes -= from;
    }

    /* "isp_from" and "isp_nbytes" must be aligned on a 4 bytes boundary */
    isp_from = from & ~3UL;
    isp_nbytes = nbytes + (from - isp_from);
    if ((isp_nbytes & 3) > 0)
        isp_nbytes = (isp_nbytes | 3) + 1;

    ntd_isp_progress_notify(isp, 0);
    string = g_string_sized_new(isp_nbytes);
    error = FALSE;

    if (isp_nbytes > 0 && isp_from < 64) {
        /* The reading overlap the vector table: hacking needed.
         * More details in the following link:
         * http://forum.flashmagictool.com/index.php?topic=3988.0
         */
        gchar *vt = _ntd_get_vt(lpc);
        error = vt == NULL;
        if (! error) {
            g_string_append_len(string, vt + isp_from, 64 - isp_from);
            g_free(vt);
            isp_nbytes -= 64 - isp_from;
            isp_from = 64;
        }
    }

    if (! error && isp_nbytes > 0) {
        g_debug("Reading %" G_GSIZE_FORMAT " bytes starting at address %lu", isp_nbytes, isp_from);
        if (ntd_lpc_command(lpc, "R %lu %" G_GSIZE_FORMAT, isp_from, isp_nbytes) &&
            ntd_lpc_answered_0(lpc)) {
            gchar *data;
            g_debug("Decoding chunk of data");
            data = _ntd_decode(lpc, isp_nbytes, 0, 1);
            if (data == NULL) {
                error = TRUE;
            } else {
                g_string_append_len(string, data, isp_nbytes);
                g_free(data);
            }
        } else {
            error = TRUE;
        }
    }

    g_debug(error ? "dump() failed" : "dump() successful");
    ntd_string_free(string, p_buffer, p_nbytes);
    return ! error;
}

static gboolean
_ntd_run(NtdIsp *isp, const gchar *address)
{
    gulong real_address;

    ntd_isp_connect(isp);
    if (!ntd_isp_is_connected(isp))
        return FALSE;

    real_address = address == NULL ? 0 : atol(address);

    return _ntd_real_run((NtdLpc *) isp, real_address);
}

static gboolean
_ntd_download_chunk(NtdLpc *lpc, gulong address, gpointer *buffer, gsize nbytes)
{
    gchar *encoded_data;
    gulong checksum;
    gboolean done;

    *buffer = NULL;
    encoded_data = NULL;

    g_debug("Reading %" G_GSIZE_FORMAT " bytes starting at address %lu",
            nbytes, address);
    done = ntd_lpc_command(lpc, "R %lu %" G_GSIZE_FORMAT, address, nbytes) &&
           ntd_lpc_answered_0(lpc) &&
           ntd_lpc_answered_data(lpc, &encoded_data, &checksum) &&
           ntd_uudecode(encoded_data, buffer, &nbytes) &&
           ntd_checksum(*buffer, nbytes) == checksum &&
           ntd_lpc_command(lpc, "OK");
    g_free(encoded_data);

    return done;
}

static gboolean
_ntd_real_run(NtdLpc *lpc, gulong address)
{
    gboolean done;

    g_debug("Running code at address %lu", address);
    done = ntd_lpc_command(lpc, "G %lu A", address) &&
           ntd_lpc_answered_0(lpc);
    g_debug(done ? "run() successful" : "run() failed");

    return done;
}

static gboolean
_ntd_encode(NtdLpc *lpc, gconstpointer data, guint nbytes,
            gdouble from, gdouble to)
{
    NtdIsp *isp;
    const gchar *chunk;
    gchar *lines;
    guint remaining, chunk_size;
    gint attempts;
    gdouble delta;

    isp = (NtdIsp *) lpc;

    remaining = nbytes;
    chunk = data;
    attempts = 1;
    delta = to - from;
    while (remaining > 0) {
        if (delta > 0)
            ntd_isp_progress_notify(isp, from + delta * PROGRESS(nbytes, remaining));

        chunk_size = MIN(remaining, 45 * 20);
        lines = ntd_uuencode(chunk, chunk_size);
        g_return_val_if_fail(lines != NULL, FALSE);

        if (!ntd_lpc_command(lpc, "%s%lu", lines,
                             ntd_checksum(chunk, chunk_size)) ||
            !ntd_lpc_answered_OK(lpc)) {
            chunk_size = 0;
        }
        g_free(lines);

        /* Check number of attempts */
        if (chunk_size > 0) {
            attempts = 1;
        } else if (attempts >= MAX_ATTEMPTS) {
            return FALSE;
        } else {
            ++attempts;
        }

        chunk += chunk_size;
        remaining -= chunk_size;
    }

    return TRUE;
}

static gpointer
_ntd_decode(NtdLpc *lpc, guint nbytes, gdouble from, gdouble to)
{
    NtdIsp *isp;
    gpointer buffer, ptr, chunk;
    gsize remaining, chunk_size;
    gchar *data;
    gulong checksum;
    gboolean done;
    gint attempts;
    gdouble delta;

    isp = (NtdIsp *) lpc;
    buffer = g_malloc(nbytes);
    ptr = buffer;
    remaining = nbytes;
    delta = to - from;
    done = TRUE;
    attempts = 1;

    while (remaining > 0 && done) {
        if (delta > 0)
            ntd_isp_progress_notify(isp, from + delta * PROGRESS(nbytes, remaining));

        chunk = NULL;
        if (! ntd_lpc_answered_data(lpc, &data, &checksum)) {
            done = FALSE;
            g_debug("Checksum line not found");
        }

        if (done && ! ntd_uudecode(data, &chunk, &chunk_size)) {
            done = FALSE;
            g_debug("Failed to decode data");
        }

        if (done && (chunk_size <= 0 || chunk_size > remaining)) {
            done = FALSE;
            g_debug("Unexpected chunk size (%" G_GSIZE_FORMAT " of %" G_GSIZE_FORMAT " remaining)",
                    chunk_size, remaining);
        }

        if (done) {
            gulong my_checksum = ntd_checksum(chunk, chunk_size);
            if (my_checksum != checksum) {
                done = FALSE;
                g_debug("Checksums do not match chunk size (got %lu, expected %lu)",
                        checksum, my_checksum);
            }
        }

        g_free(data);
        data = NULL;

        if (done) {
            memcpy(ptr, chunk, chunk_size);
            ptr += chunk_size;
            remaining -= chunk_size;
            attempts = 1;
            done = ntd_lpc_command(lpc, "OK");
        } else if (attempts < MAX_ATTEMPTS) {
            done = ntd_lpc_command(lpc, "RESEND");
            ++attempts;
        }

        g_free(chunk);
        chunk = NULL;
        chunk_size = 0;
    }

    if (!done) {
        g_free(buffer);
        buffer = NULL;
    }

    return buffer;
}

static guint
_ntd_flash(NtdLpc *lpc, gulong ram, gulong flash, guint size,
           gdouble from, gdouble to)
{
    NtdIsp *isp;
    NtdLpcPrivate *data;
    gint from_sector, to_sector;
    guint remaining, chunk_size;
    gdouble delta;

    isp = (NtdIsp *) lpc;
    data = lpc->data;

    g_return_val_if_fail(data->chunk_size >= 256, 0);

    /* Adjust the size value to be at least chunk_size bytes
     * and on a 256 byte boundary */
    if (size == 256) {
        /* Stupid hack to allow the lastsector stuff: do nothing
         * if size is exactly 256 */
    } else if (size < data->chunk_size) {
        size = data->chunk_size;
    } else if ((size & 0xFF) > 0) {
        size = (size | 0xFF) + 1;
    }

    ntd_isp_connect(isp);
    if (!ntd_isp_is_connected(isp))
        return 0;

    delta = to - from;

    for (remaining = size; remaining > 0; remaining -= chunk_size) {
        if (delta > 0)
            ntd_isp_progress_notify(isp, from + delta * PROGRESS(size, remaining));

        /* Adjust addresses and sizes to accomodate the new size.
         * TODO: actually only chunk_size and 256 are used but the
         * LPC devices could accept other values too. Check the
         * "C" command specifications for further details.
         *
         * A chunk of 256 bytes is the only one considered because
         * the lastsector stuff need it! A more serious approach
         * is badly needed...
         */
        if (remaining <= 256)
            chunk_size = 256;
        else
            chunk_size = data->chunk_size;

        if (remaining < chunk_size) {
            flash -= chunk_size - remaining;
            ram -= chunk_size - remaining;
            remaining = chunk_size;
        }

        from_sector = ntd_lpc_map_address(lpc, flash);
        to_sector = ntd_lpc_map_address(lpc, flash + chunk_size - 1);

        g_debug("Preparing sectors from %d to %d", from_sector, to_sector);
        if (!ntd_lpc_command(lpc, "P %d %d", from_sector, to_sector) ||
            !ntd_lpc_answered_0(lpc))
            goto bailout;

        g_debug("Copying %u bytes from %lu to %lu", chunk_size, ram, flash);
        if (!ntd_lpc_command(lpc, "C %lu %lu %u",
                               flash, ram, chunk_size) ||
            !ntd_lpc_answered_0(lpc))
            goto bailout;

        flash += chunk_size;
        ram += chunk_size;
    }

    g_debug("flash() successful");
    return size;

bailout:
    g_debug("flash() failed");
    return 0;
}

static void
_ntd_get_vt_code(gulong data_address, gpointer *buffer, gsize *size)
{
    /* The following code move 64 bytes from the unmapped
     * address 0 to data_address */
    const guint32 code[] = {
        0xE10F5000, /*   CPSR R5 MRS */
        0xE3E000C0, /*   C0 R0 MVN */
        0xE0050000, /*   R0 R5 R0 AND */
        0xE12FF000, /*   R0 PC TEQ */
        0xE59F0030, /*   E01FC000 R0 LDRI */
        0xE3A01001, /*   1 R1 MOV */
        0xE5801040, /*   40 R0+ R1 STR */
        0xE3A01040, /*   40 R1 MOV */
        0xE59F301C, /*   data_address+64 R3 LDRI */
                    /* loop: */
        0xE4114004, /*   4 R1-- R4 LDR */
        0xE4034004, /*   4 R3-- R4 STR */
        0xE1B01001, /*   R1 R1 MOVS */
        0x5AFFFFFB, /*   loop BPL */
        0xE3A01000, /*   0 R1 MOV */
        0xE5801040, /*   40 R0+ R1 STR */
        0xE12FF005, /*   R5 PC TEQ */
        0xE12FFF1E, /*   BX LR */

        data_address + 64,
        0xE01FC000,
    };
    guint32 *p_buffer;
    gint n;

    *size = G_N_ELEMENTS(code) * sizeof(guint32);
    *buffer = p_buffer = g_new(guint32, G_N_ELEMENTS(code));

    for (n = 0; n < G_N_ELEMENTS(code); ++n)
        p_buffer[n] = (code[n]);
}

static gboolean
_ntd_inject_vt_code(NtdLpc *lpc, gulong data_address, gulong code_address)
{
    gpointer code;
    gchar *encoded_code;
    gsize code_size;
    gulong checksum;
    gboolean done;

    _ntd_get_vt_code(data_address, &code, &code_size);
    encoded_code = ntd_uuencode(code, code_size);
    checksum = ntd_checksum(code, code_size);
    g_free(code);

    g_debug("Injecting code to read the vector table");
    done = ntd_lpc_command(lpc, "W %lu %" G_GSIZE_FORMAT, code_address, code_size) &&
           ntd_lpc_answered_0(lpc) &&
           ntd_lpc_command(lpc, "%s%lu", encoded_code, checksum) &&
           ntd_lpc_answered_OK(lpc);
    g_free(encoded_code);

    g_debug(done ? "inject_vt() successful" : "inject_vt() failed");
    return done;
}

static gpointer
_ntd_get_vt(NtdLpc *lpc)
{
    NtdLpcPrivate *data = lpc->data;
    gulong data_address, code_address;
    gpointer vt = NULL;

    data_address = data->buffer_address;
    code_address = data_address + 64;

    if (_ntd_inject_vt_code(lpc, data_address, code_address) &&
        _ntd_real_run(lpc, code_address)) {
        if (!_ntd_download_chunk(lpc, data_address, &vt, 64)) {
            g_free(vt);
            vt = NULL;
        }
    }

    return vt;
}
