/* ntdisp - NtD In System Programming.
 *
 * Copyright (C) 2012  Nicola Fontana <ntd at entidi.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */


#include <glib-object.h>
#include <unistd.h>
#include <termios.h>
#include <fcntl.h>
#include <sys/file.h>

#include "../config.h"
#include "ntdisp-intl.h"
#include "ntdisp-utils.h"
#include "ntdisp-builtins.h"
#include "ntdisp-enums.h"
#include "ntd-device.h"
#include "ntd-serial.h"
#include "ntd-serial-private.h"
#include "ntd-device-private.h"


typedef struct _NtdLibc NtdLibc;

struct _NtdLibc {
    int                 handle;
    struct termios      old_settings;
};


static NtdLibc*     ntd_libc_get            (NtdDevice*     device);
static void         ntd_libc_close          (NtdLibc*       libc);
static void         ntd_libc_destroy        (NtdLibc*       libc);


void
_ntd_serial_open(NtdDevice *device)
{
    NtdSerial *serial;
    NtdSerialPrivate *data;
    NtdLibc *libc;
    const gchar *file;
    struct termios settings;

    libc = ntd_libc_get(device);
    g_return_if_fail(libc->handle <= 0);
    ntd_system_clear_error();

    file = ntd_device_get_file(device);
    if (file == NULL || file[0] == '\0') {
        ntd_device_set_error(device, NTD_SYSTEM_ERROR_NODEV,
                             P_("Device file not specified"));
        return;
    }

    libc->handle = open(file, O_RDWR | O_NOCTTY | O_NONBLOCK);
    if (ntd_device_set_error_from_system(device, "open()"))
        return;

    serial = (NtdSerial *) device;
    data = serial->data;
    if (data->exclusive && flock(libc->handle, LOCK_EX | LOCK_NB) < 0) {
        /* Do not close, otherwise the other process will fail */
        libc->handle = 0;
        ntd_device_set_error(device, NTD_SYSTEM_ERROR_ACCES, "flock()");
        return;
    }

    tcflush(libc->handle, TCIOFLUSH);
    fcntl(libc->handle, F_SETFL, fcntl(libc->handle, F_GETFL) & ~O_NONBLOCK);
    if (ntd_device_set_error_from_system(device, "fcntl()"))
        return;

    tcgetattr(libc->handle, &libc->old_settings);
    if (ntd_device_set_error_from_system(device, "tcgetattr()"))
        return;

    settings = libc->old_settings;
    settings.c_iflag = 0;
    settings.c_oflag = 0;
    settings.c_lflag = 0;
    cfmakeraw(&settings);
    settings.c_cflag |= CLOCAL | CREAD;
    settings.c_iflag |= IGNBRK;

    switch (data->baud) {
    case NTD_BAUD_UNKNOWN:
        break;
    case NTD_BAUD_50:
        cfsetspeed(&settings, B50);
        break;
    case NTD_BAUD_75:
        cfsetspeed(&settings, B75);
        break;
    case NTD_BAUD_110:
        cfsetspeed(&settings, B110);
        break;
    case NTD_BAUD_134:
        cfsetspeed(&settings, B134);
        break;
    case NTD_BAUD_200:
        cfsetspeed(&settings, B200);
        break;
    case NTD_BAUD_300:
        cfsetspeed(&settings, B300);
        break;
    case NTD_BAUD_600:
        cfsetspeed(&settings, B600);
        break;
    case NTD_BAUD_1200:
        cfsetspeed(&settings, B1200);
        break;
    case NTD_BAUD_1800:
        cfsetspeed(&settings, B1800);
        break;
    case NTD_BAUD_2400:
        cfsetspeed(&settings, B2400);
        break;
    case NTD_BAUD_4800:
        cfsetspeed(&settings, B4800);
        break;
    case NTD_BAUD_9600:
        cfsetspeed(&settings, B9600);
        break;
    case NTD_BAUD_19200:
        cfsetspeed(&settings, B19200);
        break;
    case NTD_BAUD_38400:
        cfsetspeed(&settings, B38400);
        break;
    case NTD_BAUD_57600:
        cfsetspeed(&settings, B57600);
        break;
    case NTD_BAUD_115200:
        cfsetspeed(&settings, B115200);
        break;
    case NTD_BAUD_230400:
        cfsetspeed(&settings, B230400);
        break;
    case NTD_BAUD_460800:
        cfsetspeed(&settings, B460800);
        break;
    default:
        ntd_device_set_error(device, NTD_SYSTEM_ERROR_INVAL,
                             "NtdSerial:baud=`%d'",
                             data->baud);
        return;
    }

    switch (data->parity) {
    case NTD_PARITY_UNKNOWN:
        break;
    case NTD_PARITY_NONE:
        settings.c_cflag &= ~PARENB;
        break;
    case NTD_PARITY_ODD:
        settings.c_cflag |= PARENB|PARODD;
        settings.c_iflag |= INPCK;
        break;
    case NTD_PARITY_EVEN:
        settings.c_cflag |= PARENB;
        settings.c_cflag &= ~PARODD;
        settings.c_iflag |= INPCK;
        break;
    case NTD_PARITY_UNCHECK:
        settings.c_cflag |= PARENB;
        settings.c_iflag &= ~INPCK;
        break;
    default:
        ntd_device_set_error(device, NTD_SYSTEM_ERROR_INVAL,
                             "NtdSerial:parity=`%d'",
                             data->parity);
        return;
    }

    settings.c_cflag &= ~CSIZE;
    switch (data->data_bits) {
    case NTD_DATA_BITS_UNKNOWN:
        break;
    case NTD_DATA_BITS_5:
        settings.c_cflag |= CS5;
        break;
    case NTD_DATA_BITS_6:
        settings.c_cflag |= CS6;
        break;
    case NTD_DATA_BITS_7:
        settings.c_cflag |= CS7;
        break;
    case NTD_DATA_BITS_8:
        settings.c_cflag |= CS8;
        break;
    default:
        ntd_device_set_error(device, NTD_SYSTEM_ERROR_INVAL,
                             "NtdSerial:data-bits=`%d'",
                             data->data_bits);
        return;
    }

    switch (data->stop_bits) {
    case NTD_STOP_BITS_UNKNOWN:
        break;
    case NTD_STOP_BITS_1:
    case NTD_STOP_BITS_1_5:
        settings.c_cflag &= ~CSTOPB;
        break;
    case NTD_STOP_BITS_2:
        settings.c_cflag |= CSTOPB;
        break;
    default:
        ntd_device_set_error(device, NTD_SYSTEM_ERROR_INVAL,
                             "NtdSerial:stop-bits=`%d'",
                             data->stop_bits);
        return;
    }

    switch (data->software_flow) {
    case NTD_SOFTWARE_FLOW_NONE:
        settings.c_iflag &= ~(IXON | IXOFF);
        break;
    case NTD_SOFTWARE_FLOW_XON:
        settings.c_iflag |= IXON;
        settings.c_iflag &= ~IXOFF;
        break;
    case NTD_SOFTWARE_FLOW_XOFF:
        settings.c_iflag &= ~IXON;
        settings.c_iflag |= IXOFF;
        break;
    case NTD_SOFTWARE_FLOW_XON_XOFF:
        settings.c_iflag |= IXON | IXOFF;
        break;
    default:
        ntd_device_set_error(device, NTD_SYSTEM_ERROR_INVAL,
                             "NtdSerial:software-flow=`%d'",
                             data->software_flow);
        return;
    }

#ifdef CRTSCTS
    switch (data->hardware_flow) {
    case NTD_HARDWARE_FLOW_NONE:
        settings.c_cflag &= ~CRTSCTS;
        break;
    case NTD_HARDWARE_FLOW_RTS:
    case NTD_HARDWARE_FLOW_CTS:
        /* Not implemented */
        break;
    case NTD_HARDWARE_FLOW_RTS_CTS:
        settings.c_cflag |= CRTSCTS;
        break;
    default:
        ntd_device_set_error(device, NTD_SYSTEM_ERROR_INVAL,
                             "NtdSerial:hardware-flow=`%d'",
                             data->hardware_flow);
        return;
    }
#endif

    settings.c_cc[VTIME] = (cc_t) (data->timeout * 10);
    settings.c_cc[VMIN] = 0;

    tcsetattr(libc->handle, TCSAFLUSH, &settings);
    if (ntd_device_set_error_from_system(device, "tcsetattr()"))
        return;
}

void
_ntd_serial_close(NtdDevice *device)
{
    NtdLibc *libc;

    libc = ntd_libc_get(device);
    g_return_if_fail(libc->handle > 0);
    ntd_system_clear_error();

    ntd_libc_close(libc);
    if (ntd_device_set_error_from_system(device, "close()"))
        return;
}

void
_ntd_serial_flush(NtdDevice *device, gboolean input, gboolean output)
{
    NtdLibc *libc;
    int queue_selector;

    libc = ntd_libc_get(device);
    g_return_if_fail(libc->handle > 0);
    ntd_system_clear_error();

    if (input && output) {
        queue_selector = TCIOFLUSH;
    } else if (input) {
        queue_selector = TCIFLUSH;
    } else {
        queue_selector = TCOFLUSH;
    }

    tcflush(libc->handle, queue_selector);
    if (ntd_device_set_error_from_system(device, "tcflush()"))
        return;
}

gsize
_ntd_serial_raw_read(NtdDevice *device, gsize nbytes, gpointer buffer)
{
    NtdLibc *libc;
    gchar *p_dst;
    gsize nread;
    ssize_t read_result;

    libc = ntd_libc_get(device);
    g_return_val_if_fail(libc->handle > 0, 0);

    p_dst = (gchar *) buffer;
    for (nread = 0; nread < nbytes; nread += read_result) {
        read_result = read(libc->handle, p_dst, nbytes - nread);

        if (read_result < 0) {
            /* Error while reading */
            ntd_device_set_error_from_system(device, "read()");
            break;
        } else if (read_result == 0) {
            /* Read timed out */
            ntd_device_set_error(device, NTD_SYSTEM_ERROR_TIMEDOUT,
                                 P_("Read %zd bytes of %zd"), nread, nbytes);
            break;
        }

        p_dst += read_result;
    }

    return nread;
}

gsize
_ntd_serial_raw_write(NtdDevice *device, gsize nbytes, gconstpointer buffer)
{
    NtdLibc *libc;
    const gchar *p_src;
    guint nwritten;
    ssize_t write_result;

    libc = ntd_libc_get(device);
    g_return_val_if_fail(libc->handle > 0, 0);

    p_src = (const gchar *) buffer;
    for (nwritten = 0; nwritten < nbytes; nwritten += write_result) {
        write_result = write(libc->handle, p_src, nbytes - nwritten);

        if (write_result < 0) {
            ntd_device_set_error_from_system(device, "write()");
            return nwritten;
        } else if (write_result > 0) {
            p_src += write_result;
        }
    }

    /* Be sure the data is transmitted */
    if (tcdrain(libc->handle) == -1)
        ntd_device_set_error_from_system(device, "tcdrain()");

    return nwritten;
}


static NtdLibc *
ntd_libc_get(NtdDevice *device)
{
    NtdSerial *serial;
    NtdSerialPrivate *data;
    NtdLibc *libc;

    serial = (NtdSerial *) device;
    data = serial->data;
    libc = data->implementation;

    if (libc == NULL) {
        data->implementation = libc = g_new0(NtdLibc, 1);
        g_object_weak_ref((GObject *) serial,
                          (GWeakNotify) ntd_libc_destroy, libc);
    }

    return libc;
}

static void
ntd_libc_close(NtdLibc *libc)
{
    tcsetattr(libc->handle, TCSAFLUSH, &libc->old_settings);
    close(libc->handle);
    libc->handle = 0;
}

static void
ntd_libc_destroy(NtdLibc *libc)
{
    g_return_if_fail(libc != NULL);

    if (libc->handle > 0)
        ntd_libc_close(libc);

    g_free(libc);
}
