/* ntdisp - NtD In System Programming.
 *
 * Copyright (C) 2012  Nicola Fontana <ntd at entidi.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */


#include <glib-object.h>
#include <windows.h>

#include "../config.h"
#include "ntdisp-intl.h"
#include "ntdisp-utils.h"
#include "ntdisp-builtins.h"
#include "ntdisp-enums.h"
#include "ntd-device.h"
#include "ntd-serial.h"
#include "ntd-serial-private.h"


typedef struct _NtdWin32 NtdWin32;

struct _NtdWin32 {
    HANDLE              handle;
};


static NtdWin32*    ntd_win32_get            (NtdDevice*     device);
static void         ntd_win32_close          (NtdWin32*      win32);
static void         ntd_win32_destroy        (NtdWin32*      win32);


void
_ntd_serial_open(NtdDevice *device)
{
    NtdWin32 *win32;
    const gchar *file;
    NtdSerial *serial;
    NtdSerialPrivate *data;
    DCB dcb;

    win32 = ntd_win32_get(device);
    g_return_if_fail(win32->handle == INVALID_HANDLE_VALUE);

    file = ntd_device_get_file(device);
    if (file == NULL || file[0] == '\0') {
        ntd_device_set_error(device, NTD_SYSTEM_ERROR_NODEV,
                             P_("Device file not specified"));
        return;
    }

    ntd_system_clear_error();

    serial = (NtdSerial *) device;
    data = serial->data;
    win32->handle = CreateFile(file,
                               GENERIC_READ | GENERIC_WRITE,
                               data->exclusive ?
                                   0 : FILE_SHARE_READ | FILE_SHARE_WRITE,
                               NULL,                       /* Security */
                               OPEN_EXISTING,              /* Creation flags */
                               FILE_FLAG_OVERLAPPED,       /* Attributes */
                               NULL);                      /* Template file */
    if (ntd_device_set_error_from_system(device, "CreateFile()"))
        return;

    GetCommState(win32->handle, &dcb);
    if (ntd_device_set_error_from_system(device, "GetCommState()"))
        return;

    switch (data->baud) {
    case NTD_BAUD_UNKNOWN:
        break;
    case NTD_BAUD_1200:
        dcb.BaudRate = CBR_1200;
        break;
    case NTD_BAUD_2400:
        dcb.BaudRate = CBR_2400;
        break;
    case NTD_BAUD_4800:
        dcb.BaudRate = CBR_4800;
        break;
    case NTD_BAUD_9600:
        dcb.BaudRate = CBR_9600;
        break;
    case NTD_BAUD_19200:
        dcb.BaudRate = CBR_19200;
        break;
    case NTD_BAUD_38400:
        dcb.BaudRate = CBR_38400;
        break;
    case NTD_BAUD_57600:
        dcb.BaudRate = CBR_57600;
        break;
    default:
        ntd_device_set_error(device, NTD_SYSTEM_ERROR_INVAL,
                             "baud=`%d'", data->baud);
        return;
    }

    switch (data->parity) {
    case NTD_PARITY_UNKNOWN:
        break;
    case NTD_PARITY_NONE:
        dcb.Parity = NOPARITY;
        break;
    case NTD_PARITY_ODD:
        dcb.Parity = ODDPARITY;
        break;
    case NTD_PARITY_EVEN:
        dcb.Parity = EVENPARITY;
        break;
    case NTD_PARITY_UNCHECK:
        dcb.Parity = MARKPARITY;
        break;
    default:
        ntd_device_set_error(device, NTD_SYSTEM_ERROR_INVAL,
                             "parity=`%d'", data->parity);
        return;
    }

    switch (data->data_bits) {
    case NTD_DATA_BITS_UNKNOWN:
        break;
    case NTD_DATA_BITS_5:
    case NTD_DATA_BITS_6:
    case NTD_DATA_BITS_7:
    case NTD_DATA_BITS_8:
        dcb.ByteSize = data->data_bits;
        break;
    default:
        ntd_device_set_error(device, NTD_SYSTEM_ERROR_INVAL,
                             "data_bits=`%d'", data->data_bits);
        return;
    }

    switch (data->stop_bits) {
    case NTD_STOP_BITS_UNKNOWN:
        break;
    case NTD_STOP_BITS_1:
        dcb.StopBits = ONESTOPBIT;
        break;
    case NTD_STOP_BITS_1_5:
        dcb.StopBits = ONE5STOPBITS;
        break;
    case NTD_STOP_BITS_2:
        dcb.StopBits = TWOSTOPBITS;
        break;
    default:
        ntd_device_set_error(device, NTD_SYSTEM_ERROR_INVAL,
                             "stop_bits=`%d'", data->stop_bits);
        return;
    }

    SetCommState(win32->handle, &dcb);
    if (ntd_device_set_error_from_system(device, P_("`SetCommState' call")))
        return;

    COMMTIMEOUTS TimeOuts = {
        MAXDWORD, MAXDWORD, data->timeout * 1000000,
        0, 0
    };

    SetCommTimeouts(win32->handle, &TimeOuts);
    if (ntd_device_set_error_from_system(device, P_("`SetCommTimeouts' call")))
        return;
}

void
_ntd_serial_close(NtdDevice *device)
{
    NtdWin32 *win32 = ntd_win32_get(device);
    g_return_if_fail(win32->handle != INVALID_HANDLE_VALUE);

    ntd_system_clear_error();

    ntd_win32_close(win32);
    if (ntd_device_set_error_from_system(device, "CloseHandle()"))
        return;
}

void
_ntd_serial_flush(NtdDevice *device, gboolean input, gboolean output)
{
    NtdWin32 *win32 = ntd_win32_get(device);
    g_return_if_fail(win32->handle != INVALID_HANDLE_VALUE);

    ntd_system_clear_error();

    if (output) {
        /* TODO: citing MSDN (http://msdn.microsoft.com/aa364439(v=vs.85).aspx),
         *       "If hFile is a handle to a communications device, the
         *        function only flushes the transmit buffer". I did not found
         *        a reliable way to flush the input/receiving queue. */
        FlushFileBuffers(win32->handle);
        if (ntd_device_set_error_from_system(device, "FlushFileBuffers()"))
            return;
    }
}

gsize
_ntd_serial_raw_read(NtdDevice *device, gsize nbytes, gpointer buffer)
{
    NtdWin32 *win32;
    OVERLAPPED overlap;
    DWORD nread;

    win32 = ntd_win32_get(device);
    g_return_val_if_fail(win32->handle != INVALID_HANDLE_VALUE, 0);

    ntd_system_clear_error();

    overlap.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
    if (ntd_device_set_error_from_system(device, "CreateEvent()"))
        return 0;

    overlap.Offset = 0;
    overlap.OffsetHigh = 0;

    if (! ReadFile(win32->handle, buffer, nbytes, &nread, &overlap) &&
        GetLastError() == ERROR_IO_PENDING)
    {
        SetLastError(0);
        GetOverlappedResult(win32->handle, &overlap, &nread, TRUE);
    }

    CloseHandle(overlap.hEvent);

    if (ntd_device_set_error_from_system(device, P_("ReadFile()"))) {
        /* Error while reading */
    } else if (nbytes != nread) {
        /* Partial read: timeout? */
        ntd_device_set_error(device, NTD_SYSTEM_ERROR_TIMEDOUT,
                             P_("Read %" G_GSIZE_FORMAT " bytes of %" G_GSIZE_FORMAT),
                             (gsize) nread, nbytes);
    }

    return nread;
}

gsize
_ntd_serial_raw_write(NtdDevice *device, gsize nbytes, gconstpointer buffer)
{
    NtdWin32 *win32;
    OVERLAPPED overlap;
    DWORD nwritten;

    win32 = ntd_win32_get(device);
    g_return_val_if_fail(win32->handle != INVALID_HANDLE_VALUE, 0);

    ntd_system_clear_error();

    overlap.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
    if (ntd_device_set_error_from_system(device, "CreateEvent()"))
        return 0;

    overlap.Offset = 0;
    overlap.OffsetHigh = 0;

    if (! WriteFile(win32->handle, buffer, nbytes, &nwritten, &overlap) &&
        GetLastError() == ERROR_IO_PENDING)
    {
        SetLastError(0);
        GetOverlappedResult(win32->handle, &overlap, &nwritten, TRUE);
    }

    CloseHandle(overlap.hEvent);

    if (ntd_device_set_error_from_system(device, "WriteFile()")) {
        /* Error while writing */
    } else if (nbytes != nwritten) {
        /* Partial write: timeout? */
        ntd_device_set_error(device, NTD_SYSTEM_ERROR_TIMEDOUT,
                             P_("Wrote %" G_GSIZE_FORMAT " bytes of %" G_GSIZE_FORMAT),
                             (gsize) nwritten, nbytes);
    }

    return nwritten;
}


static NtdWin32 *
ntd_win32_get(NtdDevice *device)
{
    NtdSerial *serial;
    NtdSerialPrivate *data;
    NtdWin32 *win32;

    serial = (NtdSerial *) device;
    data = serial->data;
    win32 = data->implementation;

    if (win32 == NULL) {
        data->implementation = win32 = g_new0(NtdWin32, 1);
        win32->handle = INVALID_HANDLE_VALUE;
        g_object_weak_ref((GObject *) serial,
                          (GWeakNotify) ntd_win32_destroy, win32);
    }

    return win32;
}

static void
ntd_win32_close(NtdWin32 *win32)
{
    CloseHandle(win32->handle);
    win32->handle = INVALID_HANDLE_VALUE;
}

static void
ntd_win32_destroy(NtdWin32 *win32)
{
    g_return_if_fail(win32 != NULL);

    if (win32->handle != INVALID_HANDLE_VALUE)
        ntd_win32_close(win32);

    g_free(win32);
}
