/* ntdisp - NtD In System Programming.
 *
 * Copyright (C) 2012  Nicola Fontana <ntd at entidi.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * SECTION:ntd-serial
 * @short_description: Serial communications
 *
 **/

/**
 * NtdSerial:
 *
 * All fields are private and should not be used directly.
 * Use its public methods instead.
 **/


#include <glib-object.h>

#include "../config.h"
#include "ntdisp-intl.h"
#include "ntdisp-utils.h"
#include "ntdisp-builtins.h"
#include "ntdisp-enums.h"
#include "ntd-device.h"
#include "ntd-serial.h"
#include "ntd-serial-private.h"


G_DEFINE_TYPE(NtdSerial, ntd_serial, NTD_TYPE_DEVICE);

enum {
    PROP_0,
    PROP_EXCLUSIVE,
    PROP_BAUD,
    PROP_PARITY,
    PROP_DATA_BITS,
    PROP_STOP_BITS,
    PROP_SOFTWARE_FLOW,
    PROP_HARDWARE_FLOW,
    PROP_TIMEOUT
};


static void     ntd_serial_set_property         (GObject*       object,
                                                 guint          prop_id,
                                                 const GValue*  value,
                                                 GParamSpec*    pspec);
static void     ntd_serial_get_property         (GObject*       object,
                                                 guint          prop_id,
                                                 GValue*        value,
                                                 GParamSpec*    pspec);


static void
ntd_serial_class_init(NtdSerialClass *klass)
{
    GObjectClass *gobject_class;
    NtdDeviceClass *device_class;
    GParamSpec *param;

    gobject_class = G_OBJECT_CLASS(klass);
    device_class = NTD_DEVICE_CLASS(klass);

    g_type_class_add_private(klass, sizeof(NtdSerialPrivate));

    gobject_class->set_property = ntd_serial_set_property;
    gobject_class->get_property = ntd_serial_get_property;

    device_class->open = _ntd_serial_open;
    device_class->close = _ntd_serial_close;
    device_class->flush = _ntd_serial_flush;
    device_class->raw_read = _ntd_serial_raw_read;
    device_class->raw_write = _ntd_serial_raw_write;

    param = g_param_spec_boolean("exclusive",
                                 P_("Exclusive Access"),
                                 P_("Define whether this process must access the device exclusively."),
                                 FALSE,
                                 G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_EXCLUSIVE, param);

    param = g_param_spec_enum("baud",
                              P_("Baud"),
                              P_("Serial communication speed"),
                              NTD_TYPE_BAUD, NTD_BAUD_9600,
                              G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_BAUD, param);

    param = g_param_spec_enum("parity",
                              P_("Parity"),
                              P_("Serial communication parity"),
                              NTD_TYPE_PARITY, NTD_PARITY_NONE,
                              G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_PARITY, param);

    param = g_param_spec_enum("data-bits",
                              P_("Data Bits"),
                              P_("Data bits"),
                              NTD_TYPE_DATA_BITS, NTD_DATA_BITS_8,
                              G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_DATA_BITS, param);

    param = g_param_spec_enum("stop-bits",
                              P_("Stop Bits"),
                              P_("Stop bits"),
                              NTD_TYPE_STOP_BITS, NTD_STOP_BITS_1,
                              G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_STOP_BITS, param);

    param = g_param_spec_enum("software-flow",
                              P_("Software Flow"),
                              P_("Software flow control to use"),
                              NTD_TYPE_SOFTWARE_FLOW,
                              NTD_SOFTWARE_FLOW_NONE,
                              G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_SOFTWARE_FLOW, param);

    param = g_param_spec_enum("hardware-flow",
                              P_("Hadware Flow"),
                              P_("Hardware flow control to use"),
                              NTD_TYPE_HARDWARE_FLOW,
                              NTD_HARDWARE_FLOW_NONE,
                              G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_HARDWARE_FLOW, param);

    param = g_param_spec_double("timeout",
                                P_("Timeout"),
                                P_("Maximum time (in seconds) waiting for a response before generating a timeout. 0 will wait forever."),
                                0, G_MAXDOUBLE, 2.,
                                G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_TIMEOUT, param);
}

static void
ntd_serial_init(NtdSerial *serial)
{
    NtdSerialPrivate *data = G_TYPE_INSTANCE_GET_PRIVATE(serial,
                                                           NTD_TYPE_SERIAL,
                                                           NtdSerialPrivate);

    data->implementation = NULL;
    data->exclusive = FALSE;
    data->baud = NTD_BAUD_9600;
    data->parity = NTD_PARITY_NONE;
    data->data_bits = NTD_DATA_BITS_8;
    data->stop_bits = NTD_STOP_BITS_1;
    data->software_flow = NTD_SOFTWARE_FLOW_NONE;
    data->hardware_flow = NTD_HARDWARE_FLOW_NONE;
    data->timeout = 1.;

    serial->data = data;
}

static void
ntd_serial_set_property(GObject *object, guint prop_id,
                          const GValue *value, GParamSpec *pspec)
{
    NtdSerialPrivate *data;

    data = ((NtdSerial *) object)->data;

    switch (prop_id) {
    case PROP_EXCLUSIVE:
        data->exclusive = g_value_get_boolean(value);
        break;
    case PROP_BAUD:
        data->baud = g_value_get_enum(value);
        break;
    case PROP_PARITY:
        data->parity = g_value_get_enum(value);
        break;
    case PROP_DATA_BITS:
        data->data_bits = g_value_get_enum(value);
        break;
    case PROP_STOP_BITS:
        data->stop_bits = g_value_get_enum(value);
        break;
    case PROP_SOFTWARE_FLOW:
        data->software_flow = g_value_get_enum(value);
        break;
    case PROP_HARDWARE_FLOW:
        data->hardware_flow = g_value_get_enum(value);
        break;
    case PROP_TIMEOUT:
        data->timeout = g_value_get_double(value);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
ntd_serial_get_property(GObject *object, guint prop_id,
                          GValue *value, GParamSpec *pspec)
{
    NtdSerialPrivate *data;

    data = ((NtdSerial *) object)->data;

    switch (prop_id) {
    case PROP_EXCLUSIVE:
        g_value_set_boolean(value, data->exclusive);
        break;
    case PROP_BAUD:
        g_value_set_enum(value, data->baud);
        break;
    case PROP_PARITY:
        g_value_set_enum(value, data->parity);
        break;
    case PROP_DATA_BITS:
        g_value_set_enum(value, data->data_bits);
        break;
    case PROP_STOP_BITS:
        g_value_set_enum(value, data->stop_bits);
        break;
    case PROP_SOFTWARE_FLOW:
        g_value_set_enum(value, data->software_flow);
        break;
    case PROP_HARDWARE_FLOW:
        g_value_set_enum(value, data->hardware_flow);
        break;
    case PROP_TIMEOUT:
        g_value_set_double(value, data->timeout);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}


/**
 * ntd_serial_new:
 * @file: device file, such as "/dev/ttyS0" on GNU/Linux
 *
 * Creates a new #NtdSerial instance.
 *
 * Returns: the newly created object.
 */
NtdSerial *
ntd_serial_new(const gchar *file)
{
    return (NtdSerial *) g_object_new(NTD_TYPE_SERIAL, "file", file, NULL);
}

/**
 * ntd_serial_set:
 * @serial: an #NtdSerial device
 * @baud: the new speed
 * @parity: the new parity
 * @data_bits: the new bits of data
 * @stop_bits: the new bits of stop
 *
 * Sets in one shot all the most common parameters of @serial.
 *
 * To enable the new configuration, the @serial device must be closed and
 * reopened (ntd_device_close() and ntd_device_open()).
 */
void
ntd_serial_set(NtdSerial *serial, NtdBaud baud, NtdParity parity,
               NtdDataBits data_bits, NtdStopBits stop_bits)
{
    g_return_if_fail(NTD_IS_SERIAL(serial));

    g_object_set(serial, "baud", baud, "parity", parity,
                 "data-bits", data_bits, "stop-bits", stop_bits, NULL);
}

/**
 * ntd_serial_set_exclusive:
 * @serial: an #NtdSerial device
 * @exclusive: whether this device will be accessed exclusively
 *
 * Sets the state of the exclusive flag of @serial to @exclusive.
 *
 * An exclusive state means the device is opened with flock(), that
 * is if enabled only the current process will be able to access
 * this resource.
 */
void
ntd_serial_set_exclusive(NtdSerial *serial, gboolean exclusive)
{
    g_return_if_fail(NTD_IS_SERIAL(serial));
    g_object_set(serial, "exclusive", exclusive, NULL);
}

/**
 * ntd_serial_get_exclusive:
 * @serial: an #NtdSerial device
 *
 * Gets the exclusive state of @serial.
 *
 * Returns: the value of the exclusive flag.
 */
gboolean
ntd_serial_get_exclusive(NtdSerial *serial)
{
    NtdSerialPrivate *data;

    g_return_val_if_fail(NTD_IS_SERIAL(serial), FALSE);

    data = serial->data;

    return data->exclusive;
}

/**
 * ntd_serial_set_baud:
 * @serial: an #NtdSerial device
 * @baud: the new speed
 *
 * Sets the baud-rate to be used on @serial.
 * To enable the new configuration, the @serial device must be closed and
 * reopened (ntd_device_close() and ntd_device_open()).
 *
 * To configure all the serial parameters in one shot, consider using
 * ntd_serial_set().
 */
void
ntd_serial_set_baud(NtdSerial *serial, NtdBaud baud)
{
    g_return_if_fail(NTD_IS_SERIAL(serial));
    g_object_set(serial, "baud", baud, NULL);
}

/**
 * ntd_serial_get_baud:
 * @serial: an #NtdSerial device
 *
 * Gets the serial speed of @serial.
 *
 * Returns: the requested serial speed or %NTD_BAUD_UNKNOWN if @serial is not
 *          an #NtdSerial object.
 */
NtdBaud
ntd_serial_get_baud(NtdSerial *serial)
{
    NtdSerialPrivate *data;

    g_return_val_if_fail(NTD_IS_SERIAL(serial), NTD_BAUD_UNKNOWN);

    data = serial->data;

    return data->baud;
}

/**
 * ntd_serial_set_parity:
 * @serial: an #NtdSerial device
 * @parity: the new parity
 *
 * Sets the parity to be used on @serial.
 * To enable the new configuration, the @serial device must be closed and
 * reopened (ntd_device_close() and ntd_device_open()).
 *
 * To configure all the serial parameters in one shot, consider using
 * ntd_serial_set().
 */
void
ntd_serial_set_parity(NtdSerial *serial, NtdParity parity)
{
    g_return_if_fail(NTD_IS_SERIAL(serial));
    g_object_set(serial, "parity", parity, NULL);
}

/**
 * ntd_serial_get_parity:
 * @serial: an #NtdSerial device
 *
 * Gets the parity used on @serial.
 *
 * Returns: the requested parity or %NTD_PARITY_UNKNOWN
 *          if @serial is not an #NtdSerial object.
 */
NtdParity
ntd_serial_get_parity(NtdSerial *serial)
{
    NtdSerialPrivate *data;

    g_return_val_if_fail(NTD_IS_SERIAL(serial), NTD_PARITY_UNKNOWN);

    data = serial->data;

    return data->parity;
}

/**
 * ntd_serial_set_data_bits:
 * @serial: an #NtdSerial device
 * @data_bits: the new bits of data
 *
 * Sets the number of data bits to be used on @serial.
 * To enable the new configuration, the @serial device must be closed and
 * reopened (ntd_device_close() and ntd_device_open()).
 *
 * To configure all the serial parameters in one shot, consider using
 * ntd_serial_set().
 */
void
ntd_serial_set_data_bits(NtdSerial *serial, NtdDataBits data_bits)
{
    g_return_if_fail(NTD_IS_SERIAL(serial));
    g_object_set(serial, "data-bits", data_bits, NULL);
}

/**
 * ntd_serial_get_data_bits:
 * @serial: an #NtdSerial device
 *
 * Gets the number of bits used as data on @serial.
 *
 * Returns: the requested data bits value or %NTD_DATA_BITS_UNKNOWN
 *          if @serial is not an #NtdSerial object.
 */
NtdDataBits
ntd_serial_get_data_bits(NtdSerial *serial)
{
    NtdSerialPrivate *data;

    g_return_val_if_fail(NTD_IS_SERIAL(serial), NTD_DATA_BITS_UNKNOWN);

    data = serial->data;

    return data->data_bits;
}

/**
 * ntd_serial_set_stop_bits:
 * @serial: an #NtdSerial device
 * @stop_bits: the new bits of stop
 *
 * Sets the number of null bits to use as stop on @serial.
 * To enable the new configuration, the @serial device must be closed and
 * reopened (ntd_device_close() and ntd_device_open()).
 *
 * To configure all the serial parameters in one shot, consider using
 * ntd_serial_set().
 */
void
ntd_serial_set_stop_bits(NtdSerial *serial, NtdStopBits stop_bits)
{
    g_return_if_fail(NTD_IS_SERIAL(serial));
    g_object_set(serial, "stop-bits", stop_bits, NULL);
}

/**
 * ntd_serial_get_stop_bits:
 * @serial: an #NtdSerial device
 *
 * Gets the number of bits used as stop on @serial.
 *
 * Returns: the requested stop bits value or %NTD_STOP_BITS_UNKNOWN
 *          if @serial is not an #NtdSerial object.
 */
NtdStopBits
ntd_serial_get_stop_bits(NtdSerial *serial)
{
    NtdSerialPrivate *data;

    g_return_val_if_fail(NTD_IS_SERIAL(serial), NTD_STOP_BITS_UNKNOWN);

    data = serial->data;

    return data->stop_bits;
}

/**
 * ntd_serial_set_software_flow:
 * @serial: an #NtdSerial device
 * @software_flow: the software flow control to use
 *
 * Sets the software flow control on @serial to @software_flow.
 *
 * To configure all the serial parameters in one shot, consider using
 * ntd_serial_set().
 */
void
ntd_serial_set_software_flow(NtdSerial *serial,
                               NtdSoftwareFlow software_flow)
{
    g_return_if_fail(NTD_IS_SERIAL(serial));
    g_object_set(serial, "software-flow", software_flow, NULL);
}

/**
 * ntd_serial_get_software_flow:
 * @serial: an #NtdSerial device
 *
 * Gets the software flow control used by @serial.
 *
 * Returns: the current software flow control
 */
NtdSoftwareFlow
ntd_serial_get_software_flow(NtdSerial *serial)
{
    NtdSerialPrivate *data;

    g_return_val_if_fail(NTD_IS_SERIAL(serial), NTD_STOP_BITS_UNKNOWN);

    data = serial->data;

    return data->software_flow;
}

/**
 * ntd_serial_set_hardware_flow:
 * @serial: an #NtdSerial device
 * @hardware_flow: the hardware flow control to use
 *
 * Sets the hardware flow control on @serial to @hardware_flow.
 *
 * To configure all the serial parameters in one shot, consider using
 * ntd_serial_set().
 */
void
ntd_serial_set_hardware_flow(NtdSerial *serial,
                               NtdHardwareFlow hardware_flow)
{
    g_return_if_fail(NTD_IS_SERIAL(serial));
    g_object_set(serial, "hardware-flow", hardware_flow, NULL);
}

/**
 * ntd_serial_get_hardware_flow:
 * @serial: an #NtdSerial device
 *
 * Gets the hardware flow control used by @serial.
 *
 * Returns: the current hardware flow control
 */
NtdHardwareFlow
ntd_serial_get_hardware_flow(NtdSerial *serial)
{
    NtdSerialPrivate *data;

    g_return_val_if_fail(NTD_IS_SERIAL(serial), NTD_STOP_BITS_UNKNOWN);

    data = serial->data;

    return data->hardware_flow;
}

/**
 * ntd_serial_set_timeout:
 * @serial: an #NtdSerial device
 * @timeout: the new timeout time (in seconds)
 *
 * Sets #NtdSerial:timeout on @serial to @timeout. To enable the new
 * value, the @serial device must be closed and reopened
 * (ntd_device_close() and ntd_device_open()).
 *
 * To configure all the serial parameters in one shot, consider using
 * ntd_serial_set() instead.
 */
void
ntd_serial_set_timeout(NtdSerial *serial, gdouble timeout)
{
    g_return_if_fail(NTD_IS_SERIAL(serial));
    g_object_set(serial, "timeout", timeout, NULL);
}

/**
 * ntd_serial_get_timeout:
 * @serial: an #NtdSerial instance
 *
 * Gets the time (in seconds) to wait before generating a timeout
 * error on @serial. A 0 value means no timeout value set,
 * that is wait forever.
 *
 * Returns: the current timeout value.
 */
gdouble
ntd_serial_get_timeout(NtdSerial *serial)
{
    NtdSerialPrivate *data;

    g_return_val_if_fail(NTD_IS_SERIAL(serial), 0);

    data = serial->data;

    return data->timeout;
}
