/* ntdisp - NtD In System Programming.
 *
 * Copyright (C) 2012  Nicola Fontana <ntd at entidi.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * SECTION:ntd-stm
 * @short_description: STM default bootloader protocol
 *
 **/

/**
 * NtdStm:
 *
 * All fields are private and should not be used directly.
 * Use its public methods instead.
 **/


#include <glib-object.h>
#include <stdlib.h>
#include <string.h>

#include "../config.h"
#include "ntdisp-intl.h"
#include "ntdisp-utils.h"
#include "ntdisp-builtins.h"
#include "ntdisp-enums.h"
#include "ntd-device.h"
#include "ntd-serial.h"
#include "ntd-isp.h"
#include "ntd-stm.h"
#include "ntd-stm-private.h"


#define MAX_ATTEMPTS    5
#define PROGRESS(size,still)  ((gdouble) ((size) - (still)) / (size))


G_DEFINE_TYPE(NtdStm, ntd_stm, NTD_TYPE_ISP);

enum {
    PROP_0,
    PROP_PID,
    PROP_VERSION,
    PROP_CODE_TABLE,
    PROP_FLASH_BASE
};


static void         _ntd_set_property           (GObject*       object,
                                                 guint          prop_id,
                                                 const GValue*  value,
                                                 GParamSpec*    pspec);
static void         _ntd_get_property           (GObject*       object,
                                                 guint          prop_id,
                                                 GValue*        value,
                                                 GParamSpec*    pspec);
static void         _ntd_set_defaults           (NtdIsp*        isp);
static const gchar* _ntd_status_to_message      (gint           status);
static void         _ntd_connect                (NtdIsp*        isp);
static void         _ntd_disconnect             (NtdIsp*        isp);
static gboolean     _ntd_probe                  (NtdIsp*        isp);
static gboolean     _ntd_wipe                   (NtdIsp*        isp,
                                                 const gchar*   address,
                                                 gsize          nbytes);
static gboolean     _ntd_download               (NtdIsp*        isp,
                                                 const gchar*   address,
                                                 gconstpointer  buffer,
                                                 gsize          nbytes);
static gboolean     _ntd_dump                   (NtdIsp*        isp,
                                                 const gchar*   address,
                                                 gpointer*      p_buffer,
                                                 gsize*         p_nbytes);
static gboolean     _ntd_run                    (NtdIsp*        isp,
                                                 const gchar*   address);
static gboolean     _ntd_download_chunk         (NtdStm*        stm,
                                                 gulong         address,
                                                 gconstpointer  buffer,
                                                 gsize          nbytes);
static gboolean     _ntd_dump_chunk             (NtdStm*        stm,
                                                 gulong         address,
                                                 gpointer       buffer,
                                                 gsize          nbytes);


static void
ntd_stm_class_init(NtdStmClass *klass)
{
    GObjectClass *gobject_class;
    NtdIspClass *isp_class;
    GParamSpec *param;

    gobject_class = (GObjectClass *) klass;
    isp_class = (NtdIspClass *) klass;

    g_type_class_add_private(klass, sizeof(NtdStmPrivate));

    gobject_class->set_property = _ntd_set_property;
    gobject_class->get_property = _ntd_get_property;

    isp_class->set_defaults = _ntd_set_defaults;
    isp_class->status_to_message = _ntd_status_to_message;
    isp_class->connect = _ntd_connect;
    isp_class->disconnect = _ntd_disconnect;
    isp_class->probe = _ntd_probe;
    isp_class->wipe = _ntd_wipe;
    isp_class->download = _ntd_download;
    isp_class->dump = _ntd_dump;
    isp_class->run = _ntd_run;

    param = g_param_spec_uint("pid",
                              P_("Product Id"),
                              P_("Version of the chip, also known as product id"),
                              0, G_MAXUINT, 0,
                              G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_PID, param);

    param = g_param_spec_string("version",
                                P_("Version"),
                                P_("Bootloader version"),
                                NULL,
                                G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_VERSION, param);

    param = g_param_spec_pointer("code-table",
                                 P_("Code Table"),
                                 P_("An array of bytes specifying the codes of the available commands: the first byte, instead of being the GET code (always '\\0'), is used to specify the length of the array"),
                                 G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_CODE_TABLE, param);

    param = g_param_spec_ulong("flash-base",
                               P_("Flash Base"),
                               P_("Starting address of the flash memory"),
                               0, G_MAXULONG, 0,
                               G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_FLASH_BASE, param);
}

static void
ntd_stm_init(NtdStm *stm)
{
    NtdStmPrivate *data = G_TYPE_INSTANCE_GET_PRIVATE(stm,
                                                        NTD_TYPE_STM,
                                                        NtdStmPrivate);

    data->pid = 0;
    data->version = NULL;
    data->code_table = NULL;
    data->flash_base = 0;

    stm->data = data;
}

static void
_ntd_set_property(GObject *object, guint prop_id,
                    const GValue *value, GParamSpec *pspec)
{
    NtdStmPrivate *data;
    guchar *table;

    data = ((NtdStm *) object)->data;

    switch (prop_id) {
    case PROP_PID:
        data->pid = g_value_get_uint(value);
        break;
    case PROP_VERSION:
        g_free(data->version);
        data->version = g_value_dup_string(value);
        break;
    case PROP_CODE_TABLE:
        g_free(data->code_table);
        table = g_value_get_pointer(value);
        data->code_table = g_memdup(table, table[0] + 1);
        break;
    case PROP_FLASH_BASE:
        data->flash_base = g_value_get_ulong(value);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
_ntd_get_property(GObject *object, guint prop_id,
                    GValue *value, GParamSpec *pspec)
{
    NtdStmPrivate *data;

    data = ((NtdStm *) object)->data;

    switch (prop_id) {
    case PROP_PID:
        g_value_set_uint(value, data->pid);
        break;
    case PROP_VERSION:
        g_value_set_string(value, data->version);
        break;
    case PROP_CODE_TABLE:
        g_value_set_pointer(value, data->code_table);
        break;
    case PROP_FLASH_BASE:
        g_value_set_ulong(value, data->flash_base);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}


/**
 * ntd_stm_new:
 *
 * Creates a new generic #NtdStm instance.
 *
 * Returns: the newly created object
 */
NtdStm *
ntd_stm_new(void)
{
    return (NtdStm *) g_object_new(NTD_TYPE_STM, NULL);
}

/**
 * ntd_stm_new_with_device:
 * @device: path to the device to use
 *
 * Creates a new default #NtdStm instance and binds it to a
 * #NtdSerial device automatically configured with sensible defaults.
 *
 * Returns: the newly created object
 */
NtdStm *
ntd_stm_new_with_device(const gchar *device)
{
    NtdStm *stm = ntd_stm_new();

    ntd_isp_set_device(NTD_ISP(stm), device);
    return stm;
}

/**
 * ntd_stm_set_pid:
 * @stm: a #NtdStm instance
 * @pid: the new product id
 *
 * Sets the product id of @stm to @pid. This is a pure informational
 * data that can be used for device specific operations.
 **/
void
ntd_stm_set_pid(NtdStm *stm, guint pid)
{
    g_return_if_fail(NTD_IS_STM(stm));
    g_object_set(stm, "pid", pid, NULL);
}

/**
 * ntd_stm_get_pid:
 * @stm: a #NtdStm instance
 *
 * Gets the product id of @stm.
 *
 * Returns: a product id
 **/
guint
ntd_stm_get_pid(NtdStm *stm)
{
    NtdStmPrivate *data;

    g_return_val_if_fail(NTD_IS_STM(stm), 0);

    data = stm->data;

    return data->pid;
}

/**
 * ntd_stm_set_version:
 * @stm: a #NtdStm
 * @version: version (informational)
 *
 * Sets a new version on @stm.
 **/
void
ntd_stm_set_version(NtdStm *stm, const gchar *version)
{
    g_return_if_fail(NTD_IS_STM(stm));
    g_object_set(stm, "version", version, NULL);
}

/**
 * ntd_stm_get_version:
 * @stm: a #NtdStm
 *
 * Gets the version of @stm. This is an informational string.
 *
 * Returns: the version
 **/
const gchar *
ntd_stm_get_version(NtdStm *stm)
{
    NtdStmPrivate *data;

    g_return_val_if_fail(NTD_IS_STM(stm), NULL);

    data = stm->data;

    return (gchar *) data->version;
}

/**
 * ntd_stm_set_code_table:
 * @stm: a #NtdStm
 * @code_table: table of command codes
 *
 * Sets a new code table on @stm.
 *
 * @code_table must be an array of bytes where every character specifies
 * the code of a command. The command sequence is specified by the
 * #NtdStmCommand enum type.
 *
 * <important>
 * The first byte, instead of representing the code of a
 * %NTD_STM_COMMAND_GET comand (always represented by a %'\0' byte),
 * must contain the length of the table, itself included.
 * </important>
 **/
void
ntd_stm_set_code_table(NtdStm *stm, const guchar *code_table)
{
    g_return_if_fail(NTD_IS_STM(stm));
    g_object_set(stm, "code-table", code_table, NULL);
}

/**
 * ntd_stm_get_code_table:
 * @stm: a #NtdStm
 *
 * Gets the code table of @stm. The table is internally owned
 * and must not be freed or modified.
 *
 * Returns: the code table array
 **/
const guchar *
ntd_stm_get_code_table(NtdStm *stm)
{
    NtdStmPrivate *data;

    g_return_val_if_fail(NTD_IS_STM(stm), NULL);

    data = stm->data;

    return data->code_table;
}

/**
 * ntd_stm_set_flash_base:
 * @stm: a #NtdStm instance
 * @flash_base: the new address
 *
 * Sets the start address of the flash on @stm to @flash_base.
 *
 * This number is used to compute the available flash for write
 * operations:
 *
 * |[
 * flash(available) = "flash-size" + "flash-base" - "write-base"
 * ]|
 **/
void
ntd_stm_set_flash_base(NtdStm *stm, gulong flash_base)
{
    g_return_if_fail(NTD_IS_STM(stm));
    g_object_set(stm, "flash-base", flash_base, NULL);
}

/**
 * ntd_stm_get_flash_base:
 * @stm: a #NtdStm instance
 *
 * Gets the base flash address of @stm.
 *
 * Returns: the base flash address
 **/
gulong
ntd_stm_get_flash_base(NtdStm *stm)
{
    NtdStmPrivate *data;

    g_return_val_if_fail(NTD_IS_STM(stm), 0);

    data = stm->data;

    return data->flash_base;
}

/**
 * ntd_stm_command:
 * @stm: a #NtdStm
 * @command: the command
 *
 * Sends @command, together with the needed protocol stuff, to the
 * device identified by @stm.
 *
 * Returns: %TRUE on success, %FALSE otherwise
 **/
gboolean
ntd_stm_command(NtdStm *stm, NtdStmCommand command)
{
    NtdStmPrivate *data;
    guchar code;

    g_return_val_if_fail(NTD_IS_STM(stm), FALSE);

    data = stm->data;

    if (command == NTD_STM_COMMAND_GET) {
        /* The GET command does not need a valid code table:
         * instead it is used to retrieve the code table.
         */
        code = '\0';
    } else {
        g_return_val_if_fail(data->code_table != NULL, FALSE);
        g_return_val_if_fail(command >= 0, FALSE);
        g_return_val_if_fail(command < data->code_table[0], FALSE);

        code = data->code_table[command];
    }

    return ntd_stm_data(stm, &code, 1);
}

/**
 * ntd_stm_data:
 * @stm: a #NtdStm
 * @data: a buffer of data
 * @data_size: size of the buffer
 *
 * Sends a chunk of generic data to @stm: this method basically adds
 * the XOR checksum, as required by the STM protocol.
 *
 * Returns: %TRUE on success, %FALSE otherwise
 **/
gboolean
ntd_stm_data(NtdStm *stm, gconstpointer data, guint data_size)
{
    NtdIsp *isp;
    NtdDevice *device;
    guint n;
    const guchar *ptr;
    guchar chksum;

    g_return_val_if_fail(NTD_IS_STM(stm), FALSE);

    isp = (NtdIsp *) stm;
    device = (NtdDevice *) ntd_isp_get_serial(isp);

    g_return_val_if_fail(NTD_IS_SERIAL(device), FALSE);
    g_return_val_if_fail(ntd_isp_is_connected(isp), FALSE);

    chksum = data_size < 2 ? 0xFF : 0;
    ptr = data;
    for (n = 0; n < data_size; ++ n) {
        chksum ^= *ptr;
        ++ ptr;
    }

    return
        ntd_device_write(device, data_size, data) &&
        ntd_device_write(device, 1, &chksum);
}

/**
 * ntd_stm_answered:
 * @stm: a #NtdStm
 *
 * Checks the response of a packet.
 *
 * Returns: a constant identifing the response type: %NTD_STM_ANSWER_NONE
 *          if timed out, %NTD_STM_ANSWER_ACK if 0x79 is received,
 *          %NTD_STM_ANSWER_NAK when 0x1F is received,
 *          %NTD_STM_ANSWER_UNKNOWN when unknows data is received.
 **/
NtdStmAnswer
ntd_stm_answered(NtdStm *stm)
{
    NtdIsp *isp;
    NtdDevice *device;
    guchar *response;
    gchar code;

    g_return_val_if_fail(NTD_IS_STM(stm), FALSE);

    isp = (NtdIsp *) stm;
    device = (NtdDevice *) ntd_isp_get_serial(isp);
    g_return_val_if_fail(NTD_IS_SERIAL(device), FALSE);

    response = ntd_device_read(device, 1);

    if (response == NULL)
        return NTD_STM_ANSWER_NONE;

    code = *response;
    g_free(response);

    switch (code) {
    case 0x79:
        return NTD_STM_ANSWER_ACK;
    case 0x1F:
        return NTD_STM_ANSWER_NAK;
    }

    return NTD_STM_ANSWER_UNKNOWN;
}


static void
_ntd_set_defaults(NtdIsp *isp)
{
    NtdSerial *serial = ntd_isp_get_serial(isp);

    ntd_serial_set(serial, NTD_BAUD_57600, NTD_PARITY_EVEN,
                   NTD_DATA_BITS_8, NTD_STOP_BITS_1);
}

static const gchar *
_ntd_status_to_message(gint status)
{
    return "";
}

static void
_ntd_connect(NtdIsp *isp)
{
    NtdStm *stm;
    NtdDevice *device;
    gint n;

    /* Undefined error by default */
    ntd_isp_set_status(isp, -1);

    stm = (NtdStm *) isp;
    device = (NtdDevice *) ntd_isp_get_serial(isp);
    g_return_if_fail(NTD_IS_SERIAL(device));

    ntd_device_open(device);

    for (n = 0; n < MAX_ATTEMPTS; ++n) {
        g_debug("Autobaud request (attempt %d)", n);
        ntd_device_clear_error(device);
        if (ntd_device_write(device, 1, "\x7F")) {
            switch (ntd_stm_answered(stm)) {
            case NTD_STM_ANSWER_ACK:
                g_debug("connect() successful: device returned ACK");
                ntd_isp_set_status(isp, 0);
                return;
            case NTD_STM_ANSWER_NAK:
                g_debug("connect() successful: device returned NACK (autobaud not needed)");
                ntd_isp_set_status(isp, 0);
                return;
            default:
                break;
            }
        }
    }

    g_debug("connect() failed");
    ntd_device_close(device);
}

static void
_ntd_disconnect(NtdIsp *isp)
{
    NtdDevice *device;

    device = (NtdDevice *) ntd_isp_get_serial(isp);
    g_return_if_fail(NTD_IS_SERIAL(device));

    ntd_device_close(device);
}

static gboolean
_ntd_probe(NtdIsp *isp)
{
    NtdStm *stm;
    NtdDevice *device;
    guchar *response;
    guchar len, hex_version, digit;
    gchar version[] = "0.0";

    ntd_isp_connect(isp);
    if (!ntd_isp_is_connected(isp))
        return FALSE;

    stm = (NtdStm *) isp;
    device = (NtdDevice *) ntd_isp_get_serial(isp);
    response = NULL;

    g_debug("Sending GET request");
    if (!ntd_stm_command(stm, NTD_STM_COMMAND_GET) ||
        ntd_stm_answered(stm) != NTD_STM_ANSWER_ACK)
        goto bailout;

    g_debug("Getting version");
    response = ntd_device_read(device, 2);
    if (response != NULL) {
        len = response[0];
        hex_version = response[1];
        g_free(response);
        g_debug("Getting ACK");
        response = ntd_device_read(device, len);
    }
    if (response == NULL)
        goto bailout;

    if (ntd_stm_answered(stm) != NTD_STM_ANSWER_ACK)
        goto bailout;

    /* Force the first byte of the table to be its length */
    response[0] = len;

    /* Convert the hexadecimal version to its string representation */
    digit = hex_version >> 4;
    version[0] = (digit > 9 ? 'A' - 10 : '0') + digit;
    digit = hex_version & 0x0F;
    version[2] = (digit > 9 ? 'A' - 10 : '0') + digit;

    ntd_stm_set_version(stm, version);
    ntd_stm_set_code_table(stm, response);

    /* A more correct way would be to check the pid and set accordingly */
    ntd_stm_set_flash_base(stm, 0x08000000);

    g_free(response);
    g_debug("probe() successful");
    return TRUE;

bailout:
    g_debug("probe() failed");
    g_free(response);
    return FALSE;
}

static gboolean
_ntd_wipe(NtdIsp *isp, const gchar *address, gsize nbytes)
{
    NtdStm *stm;
    NtdStmPrivate *data;
    NtdDevice *device;
    gboolean done;

    ntd_isp_connect(isp);
    if (!ntd_isp_is_connected(isp))
        return FALSE;

    stm = (NtdStm *) isp;
    device = (NtdDevice *) ntd_isp_get_serial(isp);
    data = stm->data;

    if (! ntd_empty(address) && data->flash_base != atol(address)) {
        g_warning("Partial wiping not yet implemented");
        return FALSE;
    }

    g_debug("Sending erase command");
    if (!ntd_stm_command(stm, NTD_STM_COMMAND_ERASE) ||
        ntd_stm_answered(stm) != NTD_STM_ANSWER_ACK)
        goto bailout;

    g_debug("Specifying all sectors");
    switch (data->code_table[NTD_STM_COMMAND_ERASE]) {
    case '\x43':
        done = ntd_device_write(device, 2, "\xFF\x00");
        break;
    case '\x44':
        done = ntd_device_write(device, 3, "\xFF\xFF\x00");
        break;
    default:
        goto bailout;
    }
    if (!done || ntd_stm_answered(stm) != NTD_STM_ANSWER_ACK)
        goto bailout;

    g_debug("wipe() successful");
    return TRUE;

bailout:
    g_debug("wipe() failed");
    return FALSE;
}

static gboolean
_ntd_download(NtdIsp *isp, const gchar *address,
              gconstpointer buffer, gsize nbytes)
{
    NtdStm *stm;
    NtdStmPrivate *data;
    const guchar *chunk;
    gsize chunk_size, remaining;
    gulong real_address;

    ntd_isp_connect(isp);
    if (!ntd_isp_is_connected(isp))
        return FALSE;

    stm = (NtdStm *) isp;
    data = stm->data;
    chunk = buffer;
    real_address = ntd_empty(address) ? data->flash_base : atol(address);

    for (remaining = nbytes; remaining > 0; remaining -= chunk_size) {
        ntd_isp_progress_notify(isp, PROGRESS(nbytes, remaining));
        chunk_size = MIN(remaining, 256);

        if (!_ntd_download_chunk(stm, real_address, chunk, chunk_size)) {
            g_debug("download(): failed");
            return FALSE;
        }

        chunk += chunk_size;
        real_address += chunk_size;
    }

    ntd_isp_progress_notify(isp, PROGRESS(nbytes, 0));
    g_debug("download(): successful");
    return TRUE;
}

static gboolean
_ntd_dump(NtdIsp *isp, const gchar *address,
          gpointer *p_buffer, gsize *p_nbytes)
{
    NtdStm *stm;
    NtdStmPrivate *data;
    gchar *chunk;
    gulong real_address;
    gsize nbytes, chunk_size, remaining;

    if (p_nbytes == NULL || *p_nbytes == 0) {
        g_warning("Getting the flash size is not implemented by the STM ISP");
        return FALSE;
    }

    stm = (NtdStm *) isp;
    data = stm->data;
    real_address = ntd_empty(address) ? data->flash_base : atol(address);
    nbytes = *p_nbytes;
    chunk = *p_buffer = g_malloc(nbytes);

    ntd_isp_connect(isp);
    if (!ntd_isp_is_connected(isp))
        return FALSE;

    ntd_isp_progress_notify(isp, 0);

    for (remaining = nbytes; remaining > 0; remaining -= chunk_size) {
        ntd_isp_progress_notify(isp, PROGRESS(nbytes, remaining));
        chunk_size = MIN(remaining, 256);

        if (!_ntd_dump_chunk(stm, real_address, chunk, chunk_size)) {
            g_free(*p_buffer);
            *p_buffer = NULL;
            g_debug("dump(): failed");
            return FALSE;
        }

        chunk += chunk_size;
        real_address += chunk_size;
    }

    ntd_isp_progress_notify(isp, 1);
    g_debug("dump(): successful");
    return TRUE;
}

static gboolean
_ntd_run(NtdIsp *isp, const gchar *address)
{
    NtdStm *stm;
    NtdStmPrivate *data;
    gulong raw_address;

    ntd_isp_connect(isp);
    if (!ntd_isp_is_connected(isp))
        return FALSE;

    stm = (NtdStm *) isp;
    data = stm->data;

    g_debug("Sending GO command");
    if (!ntd_stm_command(stm, NTD_STM_COMMAND_GO) ||
        ntd_stm_answered(stm) != NTD_STM_ANSWER_ACK)
        goto bailout;

    raw_address = ntd_empty(address) ? data->flash_base : atol(address);
    g_debug("Sending destination address (%lu)", raw_address);
    raw_address = GUINT32_TO_BE(raw_address);
    if (!ntd_stm_data(stm, &raw_address, 4) ||
        ntd_stm_answered(stm) != NTD_STM_ANSWER_ACK ||
        ntd_stm_answered(stm) != NTD_STM_ANSWER_ACK)
        return FALSE;

    g_debug("run() successful");
    return TRUE;

bailout:
    g_debug("run() failed");
    return FALSE;
}

static gboolean
_ntd_download_chunk(NtdStm *stm, gulong address,
                    gconstpointer buffer, gsize nbytes)
{
    guchar chunk[256+1];

    g_return_val_if_fail(nbytes <= 256, FALSE);

    g_debug("Sending WRITE operation");
    if (!ntd_stm_command(stm, NTD_STM_COMMAND_WRITE) ||
        ntd_stm_answered(stm) != NTD_STM_ANSWER_ACK)
        return FALSE;

    g_debug("Setting download address (%lu)", address);
    address = GUINT32_TO_BE(address);
    if (!ntd_stm_data(stm, &address, 4) ||
        ntd_stm_answered(stm) != NTD_STM_ANSWER_ACK)
        return FALSE;

    g_debug("Sending data");
    chunk[0] = nbytes - 1;
    memcpy(chunk + 1, buffer, nbytes);
    if (!ntd_stm_data(stm, chunk, nbytes + 1) ||
        ntd_stm_answered(stm) != NTD_STM_ANSWER_ACK)
        return FALSE;

    return TRUE;
}

static gboolean
_ntd_dump_chunk(NtdStm *stm, gulong address,
                gpointer buffer, gsize nbytes)
{
    NtdDevice *device;
    guchar n;
    gpointer response;

    g_debug("Sending READ operation");
    if (!ntd_stm_command(stm, NTD_STM_COMMAND_READ) ||
        ntd_stm_answered(stm) != NTD_STM_ANSWER_ACK)
        return FALSE;

    g_debug("Setting dump address (%lu)", address);
    address = GUINT32_TO_BE(address);
    if (!ntd_stm_data(stm, &address, 4) ||
        ntd_stm_answered(stm) != NTD_STM_ANSWER_ACK)
        return FALSE;

    n = nbytes - 1;

    g_debug("Setting number of bytes to dump (%u)", n);
    if (!ntd_stm_data(stm, &n, 1) ||
        ntd_stm_answered(stm) != NTD_STM_ANSWER_ACK)
        return FALSE;

    g_debug("Reading data");
    device = (NtdDevice *) ntd_isp_get_serial((NtdIsp *) stm);
    response = ntd_device_read(device, nbytes);
    if (response == NULL)
        return FALSE;

    memcpy(buffer, response, nbytes);
    g_free(response);
    return TRUE;
}
