/* ntdisp - NtD In System Programming.
 *
 * Copyright (C) 2012  Nicola Fontana <ntd at entidi.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */


#include <glib.h>
#include <errno.h>

#include "../config.h"
#include "ntdisp-intl.h"
#include "ntdisp-utils.h"

static int              NtdSystemError_to_errno(NtdSystemError  error_code);
static NtdSystemError   errno_to_NtdSystemError(int             error_code);


/**
 * ntd_system_get_error:
 *
 * Gets the current #NtdSystemError error code
 * from the system.
 *
 * Returns: the #NtdSystemError of the current system error,
 *   or %NTD_SYSTEM_ERROR_NONE if there are no system errors.
 **/
NtdSystemError
ntd_system_get_error(void)
{
    return errno_to_NtdSystemError(errno);
}

/**
 * ntd_system_get_error_message:
 * @error_code: the error code to convert.
 *
 * Gets the description of the specified system error.
 * If the @error_code has no meaning for the operating system, this
 * function returns %NULL.
 *
 * Returns: a newly allocated string containig the error message.
 *          This string must be freed when no longer used.
 *
 * TODO: Implement a reentrant version.
 **/
gchar *
ntd_system_get_error_message(NtdSystemError error_code)
{
    const gchar *message;

    switch (error_code) {
    case NTD_SYSTEM_ERROR_NONE:
        break;
    case NTD_SYSTEM_ERROR_FAILED:
        return g_strdup(P_("Generic system error"));
    default:
        message = g_strerror(NtdSystemError_to_errno(error_code));
        if (message)
            return g_strdup(message);
    }

    return NULL;
}

/**
 * ntd_system_clear_error:
 *
 * Resets the system for any pending error.
 **/
void
ntd_system_clear_error(void)
{
    errno = 0;
}


static NtdSystemError
errno_to_NtdSystemError(int error_code)
{
    NtdSystemError result;

    if (! error_code)
        return NTD_SYSTEM_ERROR_NONE;

    switch (error_code) {
    case E2BIG:
        result = NTD_SYSTEM_ERROR_2BIG;
        break;
    case EACCES:
        result = NTD_SYSTEM_ERROR_ACCES;
        break;
    case EADDRINUSE:
        result = NTD_SYSTEM_ERROR_ADDRINUSE;
        break;
    case EADDRNOTAVAIL:
        result = NTD_SYSTEM_ERROR_ADDRNOTAVAIL;
        break;
    case EAFNOSUPPORT:
        result = NTD_SYSTEM_ERROR_AFNOSUPPORT;
        break;
    case EAGAIN:
        result = NTD_SYSTEM_ERROR_AGAIN;
        break;
    case EALREADY:
        result = NTD_SYSTEM_ERROR_ALREADY;
        break;
    case EBADF:
        result = NTD_SYSTEM_ERROR_BADF;
        break;
    case EBADMSG:
        result = NTD_SYSTEM_ERROR_BADMSG;
        break;
    case EBUSY:
        result = NTD_SYSTEM_ERROR_BUSY;
        break;
    case ECANCELED:
        result = NTD_SYSTEM_ERROR_CANCELED;
        break;
    case ECHILD:
        result = NTD_SYSTEM_ERROR_CHILD;
        break;
    case ECONNABORTED:
        result = NTD_SYSTEM_ERROR_CONNABORTED;
        break;
    case ECONNREFUSED:
        result = NTD_SYSTEM_ERROR_CONNREFUSED;
        break;
    case ECONNRESET:
        result = NTD_SYSTEM_ERROR_CONNRESET;
        break;
    case EDEADLK:
        result = NTD_SYSTEM_ERROR_DEADLK;
        break;
    case EDESTADDRREQ:
        result = NTD_SYSTEM_ERROR_DESTADDRREQ;
        break;
    case EDOM:
        result = NTD_SYSTEM_ERROR_DOM;
        break;
    case EEXIST:
        result = NTD_SYSTEM_ERROR_EXIST;
        break;
    case EFAULT:
        result = NTD_SYSTEM_ERROR_FAULT;
        break;
    case EFBIG:
        result = NTD_SYSTEM_ERROR_FBIG;
        break;
    case EHOSTUNREACH:
        result = NTD_SYSTEM_ERROR_HOSTUNREACH;
        break;
    case EIDRM:
        result = NTD_SYSTEM_ERROR_IDRM;
        break;
    case EILSEQ:
        result = NTD_SYSTEM_ERROR_ILSEQ;
        break;
    case EINPROGRESS:
        result = NTD_SYSTEM_ERROR_INPROGRESS;
        break;
    case EINTR:
        result = NTD_SYSTEM_ERROR_INTR;
        break;
    case EINVAL:
        result = NTD_SYSTEM_ERROR_INVAL;
        break;
    case EIO:
        result = NTD_SYSTEM_ERROR_IO;
        break;
    case EISCONN:
        result = NTD_SYSTEM_ERROR_ISCONN;
        break;
    case EISDIR:
        result = NTD_SYSTEM_ERROR_ISDIR;
        break;
    case ELOOP:
        result = NTD_SYSTEM_ERROR_LOOP;
        break;
    case EMFILE:
        result = NTD_SYSTEM_ERROR_MFILE;
        break;
    case EMLINK:
        result = NTD_SYSTEM_ERROR_MLINK;
        break;
    case EMSGSIZE:
        result = NTD_SYSTEM_ERROR_MSGSIZE;
        break;
    case ENAMETOOLONG:
        result = NTD_SYSTEM_ERROR_NAMETOOLONG;
        break;
    case ENETDOWN:
        result = NTD_SYSTEM_ERROR_NETDOWN;
        break;
    case ENETRESET:
        result = NTD_SYSTEM_ERROR_NETRESET;
        break;
    case ENETUNREACH:
        result = NTD_SYSTEM_ERROR_NETUNREACH;
        break;
    case ENFILE:
        result = NTD_SYSTEM_ERROR_NFILE;
        break;
    case ENOBUFS:
        result = NTD_SYSTEM_ERROR_NOBUFS;
        break;
    case ENODATA:
        result = NTD_SYSTEM_ERROR_NODATA;
        break;
    case ENODEV:
        result = NTD_SYSTEM_ERROR_NODEV;
        break;
    case ENOENT:
        result = NTD_SYSTEM_ERROR_NOENT;
        break;
    case ENOEXEC:
        result = NTD_SYSTEM_ERROR_NOEXEC;
        break;
    case ENOLCK:
        result = NTD_SYSTEM_ERROR_NOLCK;
        break;
    case ENOMEM:
        result = NTD_SYSTEM_ERROR_NOMEM;
        break;
    case ENOMSG:
        result = NTD_SYSTEM_ERROR_NOMSG;
        break;
    case ENOPROTOOPT:
        result = NTD_SYSTEM_ERROR_NOPROTOOPT;
        break;
    case ENOSPC:
        result = NTD_SYSTEM_ERROR_NOSPC;
        break;
    case ENOSR:
        result = NTD_SYSTEM_ERROR_NOSR;
        break;
    case ENOSTR:
        result = NTD_SYSTEM_ERROR_NOSTR;
        break;
    case ENOSYS:
        result = NTD_SYSTEM_ERROR_NOSYS;
        break;
#ifdef ENOTCONN
    case ENOTCONN:
        result = NTD_SYSTEM_ERROR_NOTCONN;
        break;
#endif
    case ENOTDIR:
        result = NTD_SYSTEM_ERROR_NOTDIR;
        break;
    case ENOTEMPTY:
        result = NTD_SYSTEM_ERROR_NOTEMPTY;
        break;
    case ENOTSOCK:
        result = NTD_SYSTEM_ERROR_NOTSOCK;
        break;
    case ENOTSUP:
        result = NTD_SYSTEM_ERROR_NOTSUP;
        break;
    case ENOTTY:
        result = NTD_SYSTEM_ERROR_NOTTY;
        break;
    case ENXIO:
        result = NTD_SYSTEM_ERROR_NXIO;
        break;
#if EOPNOTSUPP != ENOTSUP
    case EOPNOTSUPP:
        result = NTD_SYSTEM_ERROR_OPNOTSUPP;
        break;
#endif
    case EOVERFLOW:
        result = NTD_SYSTEM_ERROR_OVERFLOW;
        break;
    case EPERM:
        result = NTD_SYSTEM_ERROR_PERM;
        break;
    case EPIPE:
        result = NTD_SYSTEM_ERROR_PIPE;
        break;
    case EPROTO:
        result = NTD_SYSTEM_ERROR_PROTO;
        break;
    case EPROTONOSUPPORT:
        result = NTD_SYSTEM_ERROR_PROTONOSUPPORT;
        break;
    case EPROTOTYPE:
        result = NTD_SYSTEM_ERROR_PROTOTYPE;
        break;
    case ERANGE:
        result = NTD_SYSTEM_ERROR_RANGE;
        break;
    case EROFS:
        result = NTD_SYSTEM_ERROR_ROFS;
        break;
    case ESPIPE:
        result = NTD_SYSTEM_ERROR_SPIPE;
        break;
    case ESRCH:
        result = NTD_SYSTEM_ERROR_SRCH;
        break;
    case ETIME:
        result = NTD_SYSTEM_ERROR_TIME;
        break;
    case ETIMEDOUT:
        result = NTD_SYSTEM_ERROR_TIMEDOUT;
        break;
    case ETXTBSY:
        result = NTD_SYSTEM_ERROR_TXTBSY;
        break;
    case EXDEV:
        result = NTD_SYSTEM_ERROR_XDEV;
        break;
    default:
        result = NTD_SYSTEM_ERROR_FAILED;
        break;
    }

    return result;
}

static int
NtdSystemError_to_errno(NtdSystemError error_code)
{
    int result;

    if (error_code == NTD_SYSTEM_ERROR_NONE)
        return 0;

    switch (error_code) {
    case NTD_SYSTEM_ERROR_2BIG:
        result = E2BIG;
        break;
    case NTD_SYSTEM_ERROR_ACCES:
        result = EACCES;
        break;
    case NTD_SYSTEM_ERROR_ADDRINUSE:
        result = EADDRINUSE;
        break;
    case NTD_SYSTEM_ERROR_ADDRNOTAVAIL:
        result = EADDRNOTAVAIL;
        break;
    case NTD_SYSTEM_ERROR_AFNOSUPPORT:
        result = EAFNOSUPPORT;
        break;
    case NTD_SYSTEM_ERROR_AGAIN:
        result = EAGAIN;
        break;
    case NTD_SYSTEM_ERROR_ALREADY:
        result = EALREADY;
        break;
    case NTD_SYSTEM_ERROR_BADF:
        result = EBADF;
        break;
    case NTD_SYSTEM_ERROR_BADMSG:
        result = EBADMSG;
        break;
    case NTD_SYSTEM_ERROR_BUSY:
        result = EBUSY;
        break;
    case NTD_SYSTEM_ERROR_CANCELED:
        result = ECANCELED;
        break;
    case NTD_SYSTEM_ERROR_CHILD:
        result = ECHILD;
        break;
    case NTD_SYSTEM_ERROR_CONNABORTED:
        result = ECONNABORTED;
        break;
    case NTD_SYSTEM_ERROR_CONNREFUSED:
        result = ECONNREFUSED;
        break;
    case NTD_SYSTEM_ERROR_CONNRESET:
        result = ECONNRESET;
        break;
    case NTD_SYSTEM_ERROR_DEADLK:
        result = EDEADLK;
        break;
    case NTD_SYSTEM_ERROR_DESTADDRREQ:
        result = EDESTADDRREQ;
        break;
    case NTD_SYSTEM_ERROR_DOM:
        result = EDOM;
        break;
    case NTD_SYSTEM_ERROR_EXIST:
        result = EEXIST;
        break;
    case NTD_SYSTEM_ERROR_FAULT:
        result = EFAULT;
        break;
    case NTD_SYSTEM_ERROR_FBIG:
        result = EFBIG;
        break;
    case NTD_SYSTEM_ERROR_HOSTUNREACH:
        result = EHOSTUNREACH;
        break;
    case NTD_SYSTEM_ERROR_IDRM:
        result = EIDRM;
        break;
    case NTD_SYSTEM_ERROR_ILSEQ:
        result = EILSEQ;
        break;
    case NTD_SYSTEM_ERROR_INPROGRESS:
        result = EINPROGRESS;
        break;
    case NTD_SYSTEM_ERROR_INTR:
        result = EINTR;
        break;
    case NTD_SYSTEM_ERROR_INVAL:
        result = EINVAL;
        break;
    case NTD_SYSTEM_ERROR_IO:
        result = EIO;
        break;
    case NTD_SYSTEM_ERROR_ISCONN:
        result = EISCONN;
        break;
    case NTD_SYSTEM_ERROR_ISDIR:
        result = EISDIR;
        break;
    case NTD_SYSTEM_ERROR_LOOP:
        result = ELOOP;
        break;
    case NTD_SYSTEM_ERROR_MFILE:
        result = EMFILE;
        break;
    case NTD_SYSTEM_ERROR_MLINK:
        result = EMLINK;
        break;
    case NTD_SYSTEM_ERROR_MSGSIZE:
        result = EMSGSIZE;
        break;
    case NTD_SYSTEM_ERROR_NAMETOOLONG:
        result = ENAMETOOLONG;
        break;
    case NTD_SYSTEM_ERROR_NETDOWN:
        result = ENETDOWN;
        break;
    case NTD_SYSTEM_ERROR_NETRESET:
        result = ENETRESET;
        break;
    case NTD_SYSTEM_ERROR_NETUNREACH:
        result = ENETUNREACH;
        break;
    case NTD_SYSTEM_ERROR_NFILE:
        result = ENFILE;
        break;
    case NTD_SYSTEM_ERROR_NOBUFS:
        result = ENOBUFS;
        break;
    case NTD_SYSTEM_ERROR_NODATA:
        result = ENODATA;
        break;
    case NTD_SYSTEM_ERROR_NODEV:
        result = ENODEV;
        break;
    case NTD_SYSTEM_ERROR_NOENT:
        result = ENOENT;
        break;
    case NTD_SYSTEM_ERROR_NOEXEC:
        result = ENOEXEC;
        break;
    case NTD_SYSTEM_ERROR_NOLCK:
        result = ENOLCK;
        break;
    case NTD_SYSTEM_ERROR_NOMEM:
        result = ENOMEM;
        break;
    case NTD_SYSTEM_ERROR_NOMSG:
        result = ENOMSG;
        break;
    case NTD_SYSTEM_ERROR_NOPROTOOPT:
        result = ENOPROTOOPT;
        break;
    case NTD_SYSTEM_ERROR_NOSPC:
        result = ENOSPC;
        break;
    case NTD_SYSTEM_ERROR_NOSR:
        result = ENOSR;
        break;
    case NTD_SYSTEM_ERROR_NOSTR:
        result = ENOSTR;
        break;
    case NTD_SYSTEM_ERROR_NOSYS:
        result = ENOSYS;
        break;
#ifdef ENOTCONN
    case NTD_SYSTEM_ERROR_NOTCONN:
        result = ENOTCONN;
        break;
#endif
    case NTD_SYSTEM_ERROR_NOTDIR:
        result = ENOTDIR;
        break;
    case NTD_SYSTEM_ERROR_NOTEMPTY:
        result = ENOTEMPTY;
        break;
    case NTD_SYSTEM_ERROR_NOTSOCK:
        result = ENOTSOCK;
        break;
    case NTD_SYSTEM_ERROR_NOTSUP:
        result = ENOTSUP;
        break;
    case NTD_SYSTEM_ERROR_NOTTY:
        result = ENOTTY;
        break;
    case NTD_SYSTEM_ERROR_NXIO:
        result = ENXIO;
        break;
#if EOPNOTSUPP != ENOTSUP
    case NTD_SYSTEM_ERROR_OPNOTSUPP:
        result = EOPNOTSUPP;
        break;
#endif
    case NTD_SYSTEM_ERROR_OVERFLOW:
        result = EOVERFLOW;
        break;
    case NTD_SYSTEM_ERROR_PERM:
        result = EPERM;
        break;
    case NTD_SYSTEM_ERROR_PIPE:
        result = EPIPE;
        break;
    case NTD_SYSTEM_ERROR_PROTO:
        result = EPROTO;
        break;
    case NTD_SYSTEM_ERROR_PROTONOSUPPORT:
        result = EPROTONOSUPPORT;
        break;
    case NTD_SYSTEM_ERROR_PROTOTYPE:
        result = EPROTOTYPE;
        break;
    case NTD_SYSTEM_ERROR_RANGE:
        result = ERANGE;
        break;
    case NTD_SYSTEM_ERROR_ROFS:
        result = EROFS;
        break;
    case NTD_SYSTEM_ERROR_SPIPE:
        result = ESPIPE;
        break;
    case NTD_SYSTEM_ERROR_SRCH:
        result = ESRCH;
        break;
    case NTD_SYSTEM_ERROR_TIME:
        result = ETIME;
        break;
    case NTD_SYSTEM_ERROR_TIMEDOUT:
        result = ETIMEDOUT;
        break;
    case NTD_SYSTEM_ERROR_TXTBSY:
        result = ETXTBSY;
        break;
    case NTD_SYSTEM_ERROR_XDEV:
        result = EXDEV;
        break;
    default:
        result = ENOTSUP;
        break;
    }

    return result;
}

