/* ntdisp - NtD In System Programming.
 *
 * Copyright (C) 2012  Nicola Fontana <ntd at entidi.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */


#include <glib.h>
#include <windows.h>

#include "../config.h"
#include "ntdisp-intl.h"
#include "ntdisp-utils.h"


#ifndef ERROR_FILE_TOO_LARGE
#define ERROR_FILE_TOO_LARGE                223
#endif

#ifndef ERROR_OPERATION_IN_PROGRESS
#define ERROR_OPERATION_IN_PROGRESS         329
#endif

#ifndef ERROR_DIRECTORY_NOT_SUPPORTED
#define ERROR_DIRECTORY_NOT_SUPPORTED       336
#endif

#ifndef ERROR_ILLEGAL_CHARACTER
#define ERROR_ILLEGAL_CHARACTER             582
#endif

#ifndef ERROR_UNIDENTIFIED_ERROR
#define ERROR_UNIDENTIFIED_ERROR            1287
#endif

#ifndef ERROR_EVT_MESSAGE_NOT_FOUND
#define ERROR_EVT_MESSAGE_NOT_FOUND         15027
#endif

#ifndef ERROR_EVT_MESSAGE_ID_NOT_FOUND
#define ERROR_EVT_MESSAGE_ID_NOT_FOUND      15028
#endif


static DWORD            NtdSystemError_to_LastError(NtdSystemError  error_code);
static NtdSystemError   LastError_to_NtdSystemError(DWORD           error_code);


/**
 * ntd_system_get_error:
 *
 * Gets the current #NtdSystemError error code
 * from the system.
 *
 * returns: the #NtdSystemError of the current system error,
 *   or %NTD_SYSTEM_ERROR_NONE if there are no system errors.
 **/
NtdSystemError
ntd_system_get_error(void)
{
    return LastError_to_NtdSystemError(GetLastError());
}

/**
 * ntd_system_get_error_message:
 * @error_code: the error code to convert.
 *
 * Gets the description of the specified system error.
 * If the @error_code has no meaning for the operating system, this function
 * returns %NULL.
 *
 * returns: a newly allocated string containig the error message.
 *          This string must be freed when no longer used.
 **/
gchar *
ntd_system_get_error_message(NtdSystemError error_code)
{
    const gchar *message;

    switch (error_code) {
    case NTD_SYSTEM_ERROR_NONE:
        break;
    case NTD_SYSTEM_ERROR_FAILED:
        return g_strdup(P_("Generic system error"));
    default:
        message = g_strerror(NtdSystemError_to_LastError(error_code));
        if (message)
            return g_strdup(message);
    }

    return NULL;
}

/**
 * ntd_system_clear_error:
 *
 * Resets the system for any pending error.
 **/
void
ntd_system_clear_error(void)
{
    SetLastError(ERROR_SUCCESS);
}


static NtdSystemError
LastError_to_NtdSystemError(DWORD error_code)
{
    NtdSystemError result;

    if (error_code == ERROR_SUCCESS)
        return NTD_SYSTEM_ERROR_NONE;

    switch (error_code) {
    case ERROR_BAD_ENVIRONMENT:
        result = NTD_SYSTEM_ERROR_2BIG;
        break;
    case ERROR_ACCESS_DENIED:
    case ERROR_INVALID_ACCESS:
    case ERROR_SHARING_VIOLATION:
    case ERROR_LOCK_VIOLATION:
        result = NTD_SYSTEM_ERROR_ACCES;
        break;
    case ERROR_ADDRESS_ALREADY_ASSOCIATED:
        result = NTD_SYSTEM_ERROR_ADDRINUSE;
        break;
    case ERROR_INVALID_ADDRESS:
        result = NTD_SYSTEM_ERROR_ADDRNOTAVAIL;
        break;
    case ERROR_INCORRECT_ADDRESS:
        result = NTD_SYSTEM_ERROR_AFNOSUPPORT;
        break;
    case ERROR_RETRY:
    case ERROR_MORE_DATA:
    case ERROR_NO_PROC_SLOTS:
    case ERROR_MAX_THRDS_REACHED:
    case ERROR_DEVICE_IN_USE:
    case ERROR_OPEN_FILES:
    case ERROR_ACTIVE_CONNECTIONS:
        result = NTD_SYSTEM_ERROR_AGAIN;
        break;
    case ERROR_CONNECTION_ACTIVE:
        result = NTD_SYSTEM_ERROR_ALREADY;
        break;
    case ERROR_INVALID_HANDLE:
        result = NTD_SYSTEM_ERROR_BADF;
        break;
    case ERROR_INVALID_MESSAGE:
    case ERROR_INVALID_MESSAGENAME:
    case ERROR_INVALID_MESSAGEDEST:
        result = NTD_SYSTEM_ERROR_BADMSG;
        break;
    case ERROR_BUSY:
    case ERROR_CHILD_NOT_COMPLETE:
    case ERROR_SIGNAL_PENDING:
    case ERROR_PIPE_BUSY:
    case ERROR_PIPE_CONNECTED:
        result = NTD_SYSTEM_ERROR_BUSY;
        break;
    case ERROR_CANCELLED:
        result = NTD_SYSTEM_ERROR_CANCELED;
        break;
    case ERROR_WAIT_NO_CHILDREN:
        result = NTD_SYSTEM_ERROR_CHILD;
        break;
    case ERROR_CONNECTION_ABORTED:
        result = NTD_SYSTEM_ERROR_CONNABORTED;
        break;
    case ERROR_CONNECTION_REFUSED:
        result = NTD_SYSTEM_ERROR_CONNREFUSED;
        break;
    case ERROR_CONNECTION_UNAVAIL:
        result = NTD_SYSTEM_ERROR_CONNRESET;
        break;
    case ERROR_POSSIBLE_DEADLOCK:
        result = NTD_SYSTEM_ERROR_DEADLK;
        break;
    case ERROR_ALREADY_EXISTS:
    case ERROR_FILE_EXISTS:
        result = NTD_SYSTEM_ERROR_EXIST;
        break;
    case ERROR_NOACCESS:
    case ERROR_PROCESS_ABORTED:
        result = NTD_SYSTEM_ERROR_FAULT;
        break;
    case ERROR_FILE_TOO_LARGE:
        result = NTD_SYSTEM_ERROR_FBIG;
        break;
    case ERROR_HOST_UNREACHABLE:
        result = NTD_SYSTEM_ERROR_HOSTUNREACH;
        break;
    case ERROR_ILLEGAL_CHARACTER:
        result = NTD_SYSTEM_ERROR_ILSEQ;
        break;
    case ERROR_OPERATION_IN_PROGRESS:
        result = NTD_SYSTEM_ERROR_INPROGRESS;
        break;
    case ERROR_INVALID_AT_INTERRUPT_TIME:
        result = NTD_SYSTEM_ERROR_INTR;
        break;
    case ERROR_INVALID_DATA:
    case ERROR_INVALID_PARAMETER:
    case ERROR_BAD_ARGUMENTS:
    case ERROR_INVALID_SIGNAL_NUMBER:
    case ERROR_THREAD_1_INACTIVE:
    case ERROR_BAD_PIPE:
    case ERROR_NO_TOKEN:
    case ERROR_BAD_USERNAME:
    case ERROR_NEGATIVE_SEEK:
        result = NTD_SYSTEM_ERROR_INVAL;
        break;
    case ERROR_IO_DEVICE:
    case ERROR_OPEN_FAILED:
    case ERROR_SIGNAL_REFUSED:
    case ERROR_NO_SIGNAL_SENT:
    case ERROR_CRC:
        result = NTD_SYSTEM_ERROR_IO;
        break;
    case ERROR_DIRECTORY_NOT_SUPPORTED:
        result = NTD_SYSTEM_ERROR_ISDIR;
        break;
    case ERROR_CIRCULAR_DEPENDENCY:
        result = NTD_SYSTEM_ERROR_LOOP;
        break;
    case ERROR_TOO_MANY_OPEN_FILES:
        result = NTD_SYSTEM_ERROR_MFILE;
        break;
    case ERROR_TOO_MANY_LINKS:
        result = NTD_SYSTEM_ERROR_MLINK;
        break;
    case ERROR_MESSAGE_EXCEEDS_MAX_SIZE:
        result = NTD_SYSTEM_ERROR_MSGSIZE;
        break;
    case ERROR_FILENAME_EXCED_RANGE:
        result = NTD_SYSTEM_ERROR_NAMETOOLONG;
        break;
    case ERROR_NO_NETWORK:
        result = NTD_SYSTEM_ERROR_NETDOWN;
        break;
    case ERROR_NETWORK_BUSY:
        result = NTD_SYSTEM_ERROR_NETRESET;
        break;
    case ERROR_NETWORK_UNREACHABLE:
        result = NTD_SYSTEM_ERROR_NETUNREACH;
        break;
    case ERROR_NO_MORE_SEARCH_HANDLES:
        result = NTD_SYSTEM_ERROR_NFILE;
        break;
    case ERROR_BUFFER_OVERFLOW:
        result = NTD_SYSTEM_ERROR_NOBUFS;
        break;
    case ERROR_HANDLE_EOF:
        result = NTD_SYSTEM_ERROR_NODATA;
        break;
    case ERROR_INVALID_DRIVE:
    case ERROR_BAD_UNIT:
    case ERROR_BAD_DEVICE:
        result = NTD_SYSTEM_ERROR_NODEV;
        break;
    case ERROR_FILE_NOT_FOUND:
    case ERROR_PATH_NOT_FOUND:
    case ERROR_INVALID_NAME:
    case ERROR_BAD_PATHNAME:
        result = NTD_SYSTEM_ERROR_NOENT;
        break;
    case ERROR_BAD_EXE_FORMAT:
        result = NTD_SYSTEM_ERROR_NOEXEC;
        break;
    case ERROR_SHARING_BUFFER_EXCEEDED:
        result = NTD_SYSTEM_ERROR_NOLCK;
        break;
    case ERROR_NOT_ENOUGH_MEMORY:
    case ERROR_OUTOFMEMORY:
        result = NTD_SYSTEM_ERROR_NOMEM;
        break;
    case ERROR_EVT_MESSAGE_NOT_FOUND:
    case ERROR_EVT_MESSAGE_ID_NOT_FOUND:
        result = NTD_SYSTEM_ERROR_NOMSG;
        break;
    case ERROR_DISK_FULL:
    case ERROR_HANDLE_DISK_FULL:
    case ERROR_END_OF_MEDIA:
    case ERROR_EOM_OVERFLOW:
    case ERROR_NO_DATA_DETECTED:
        result = NTD_SYSTEM_ERROR_NOSPC;
        break;
    case ERROR_NOT_SUPPORTED:
    case ERROR_EAS_NOT_SUPPORTED:
    case ERROR_ATOMIC_LOCKS_NOT_SUPPORTED:
    case ERROR_CALL_NOT_IMPLEMENTED:
        result = NTD_SYSTEM_ERROR_NOSYS;
        break;
    case ERROR_DIRECTORY:
        result = NTD_SYSTEM_ERROR_NOTDIR;
        break;
    case ERROR_DIR_NOT_EMPTY:
        result = NTD_SYSTEM_ERROR_NOTEMPTY;
        break;
    case ERROR_UNIDENTIFIED_ERROR:
        result = NTD_SYSTEM_ERROR_NOTSUP;
        break;
    case ERROR_INVALID_SERVICE_CONTROL:
        result = NTD_SYSTEM_ERROR_NOTTY;
        break;
    case ERROR_FILE_INVALID:
        result = NTD_SYSTEM_ERROR_NXIO;
        break;
    case ERROR_ARITHMETIC_OVERFLOW:
        result = NTD_SYSTEM_ERROR_OVERFLOW;
        break;
    case ERROR_CANNOT_MAKE:
    case ERROR_NOT_OWNER:
        result = NTD_SYSTEM_ERROR_PERM;
        break;
    case ERROR_BROKEN_PIPE:
    case ERROR_NO_DATA:
        result = NTD_SYSTEM_ERROR_PIPE;
        break;
    case ERROR_META_EXPANSION_TOO_LONG:
    case ERROR_LABEL_TOO_LONG:
        result = NTD_SYSTEM_ERROR_RANGE;
        break;
    case ERROR_WRITE_PROTECT:
        result = NTD_SYSTEM_ERROR_ROFS;
        break;
    case ERROR_SETMARK_DETECTED:
    case ERROR_BEGINNING_OF_MEDIA:
        result = NTD_SYSTEM_ERROR_SPIPE;
        break;
    case ERROR_NOT_FOUND:
        result = NTD_SYSTEM_ERROR_SRCH;
        break;
    case ERROR_SEM_TIMEOUT:
        result = NTD_SYSTEM_ERROR_TIME;
        break;
    case ERROR_TIMEOUT:
    case WAIT_TIMEOUT:
    case ERROR_COUNTER_TIMEOUT:
        result = NTD_SYSTEM_ERROR_TIMEDOUT;
        break;
    case ERROR_NOT_SAME_DEVICE:
        result = NTD_SYSTEM_ERROR_XDEV;
        break;
    default:
        result = NTD_SYSTEM_ERROR_FAILED;
        break;
    }

    return result;
}

static DWORD
NtdSystemError_to_LastError(NtdSystemError error_code)
{
    DWORD result;

    if (error_code == NTD_SYSTEM_ERROR_NONE)
        return ERROR_SUCCESS;

    switch (error_code) {
    case NTD_SYSTEM_ERROR_2BIG:
        result = ERROR_BAD_ENVIRONMENT;
        break;
    case NTD_SYSTEM_ERROR_ACCES:
        result = ERROR_ACCESS_DENIED;
        break;
    case NTD_SYSTEM_ERROR_ADDRINUSE:
        result = ERROR_ADDRESS_ALREADY_ASSOCIATED;
        break;
    case NTD_SYSTEM_ERROR_ADDRNOTAVAIL:
        result = ERROR_INVALID_ADDRESS;
        break;
    case NTD_SYSTEM_ERROR_AFNOSUPPORT:
        result = ERROR_INCORRECT_ADDRESS;
        break;
    case NTD_SYSTEM_ERROR_AGAIN:
        result = ERROR_RETRY;
        break;
    case NTD_SYSTEM_ERROR_ALREADY:
        result = ERROR_CONNECTION_ACTIVE;
        break;
    case NTD_SYSTEM_ERROR_BADF:
        result = ERROR_INVALID_HANDLE;
        break;
    case NTD_SYSTEM_ERROR_BADMSG:
        result = ERROR_INVALID_MESSAGE;
        break;
    case NTD_SYSTEM_ERROR_BUSY:
        result = ERROR_BUSY;
        break;
    case NTD_SYSTEM_ERROR_CANCELED:
        result = ERROR_CANCELLED;
        break;
    case NTD_SYSTEM_ERROR_CHILD:
        result = ERROR_WAIT_NO_CHILDREN;
        break;
    case NTD_SYSTEM_ERROR_CONNABORTED:
        result = ERROR_CONNECTION_ABORTED;
        break;
    case NTD_SYSTEM_ERROR_CONNREFUSED:
        result = ERROR_CONNECTION_REFUSED;
        break;
    case NTD_SYSTEM_ERROR_CONNRESET:
        result = ERROR_CONNECTION_UNAVAIL;
        break;
    case NTD_SYSTEM_ERROR_DEADLK:
        result = ERROR_POSSIBLE_DEADLOCK;
        break;
    case NTD_SYSTEM_ERROR_EXIST:
        result = ERROR_ALREADY_EXISTS;
        break;
    case NTD_SYSTEM_ERROR_FAULT:
        result = ERROR_NOACCESS;
        break;
    case NTD_SYSTEM_ERROR_FBIG:
        result = ERROR_FILE_TOO_LARGE;
        break;
    case NTD_SYSTEM_ERROR_HOSTUNREACH:
        result = ERROR_HOST_UNREACHABLE;
        break;
    case NTD_SYSTEM_ERROR_ILSEQ:
        result = ERROR_ILLEGAL_CHARACTER;
        break;
    case NTD_SYSTEM_ERROR_INPROGRESS:
        result = ERROR_OPERATION_IN_PROGRESS;
        break;
    case NTD_SYSTEM_ERROR_INTR:
        result = ERROR_INVALID_AT_INTERRUPT_TIME;
        break;
    case NTD_SYSTEM_ERROR_INVAL:
        result = ERROR_INVALID_DATA;
        break;
    case NTD_SYSTEM_ERROR_IO:
        result = ERROR_IO_DEVICE;
        break;
    case NTD_SYSTEM_ERROR_ISDIR:
        result = ERROR_DIRECTORY_NOT_SUPPORTED;
        break;
    case NTD_SYSTEM_ERROR_LOOP:
        result = ERROR_CIRCULAR_DEPENDENCY;
        break;
    case NTD_SYSTEM_ERROR_MFILE:
        result = ERROR_TOO_MANY_OPEN_FILES;
        break;
    case NTD_SYSTEM_ERROR_MLINK:
        result = ERROR_TOO_MANY_LINKS;
        break;
    case NTD_SYSTEM_ERROR_MSGSIZE:
        result = ERROR_MESSAGE_EXCEEDS_MAX_SIZE;
        break;
    case NTD_SYSTEM_ERROR_NAMETOOLONG:
        result = ERROR_FILENAME_EXCED_RANGE;
        break;
    case NTD_SYSTEM_ERROR_NETDOWN:
        result = ERROR_NO_NETWORK;
        break;
    case NTD_SYSTEM_ERROR_NETRESET:
        result = ERROR_NETWORK_BUSY;
        break;
    case NTD_SYSTEM_ERROR_NETUNREACH:
        result = ERROR_NETWORK_UNREACHABLE;
        break;
    case NTD_SYSTEM_ERROR_NFILE:
        result = ERROR_NO_MORE_SEARCH_HANDLES;
        break;
    case NTD_SYSTEM_ERROR_NOBUFS:
        result = ERROR_BUFFER_OVERFLOW;
        break;
    case NTD_SYSTEM_ERROR_NODATA:
        result = ERROR_HANDLE_EOF;
        break;
    case NTD_SYSTEM_ERROR_NODEV:
        result = ERROR_BAD_DEVICE;
        break;
    case NTD_SYSTEM_ERROR_NOENT:
        result = ERROR_FILE_NOT_FOUND;
        break;
    case NTD_SYSTEM_ERROR_NOEXEC:
        result = ERROR_BAD_EXE_FORMAT;
        break;
    case NTD_SYSTEM_ERROR_NOLCK:
        result = ERROR_SHARING_BUFFER_EXCEEDED;
        break;
    case NTD_SYSTEM_ERROR_NOMEM:
        result = ERROR_NOT_ENOUGH_MEMORY;
        break;
    case NTD_SYSTEM_ERROR_NOMSG:
        result = ERROR_EVT_MESSAGE_NOT_FOUND;
        break;
    case NTD_SYSTEM_ERROR_NOSPC:
        result = ERROR_DISK_FULL;
        break;
    case NTD_SYSTEM_ERROR_NOSYS:
        result = ERROR_NOT_SUPPORTED;
        break;
    case NTD_SYSTEM_ERROR_NOTDIR:
        result = ERROR_DIRECTORY;
        break;
    case NTD_SYSTEM_ERROR_NOTEMPTY:
        result = ERROR_DIR_NOT_EMPTY;
        break;
    case NTD_SYSTEM_ERROR_NOTSUP:
        result = ERROR_UNIDENTIFIED_ERROR;
        break;
    case NTD_SYSTEM_ERROR_NOTTY:
        result = ERROR_INVALID_SERVICE_CONTROL;
        break;
    case NTD_SYSTEM_ERROR_NXIO:
        result = ERROR_FILE_INVALID;
        break;
    case NTD_SYSTEM_ERROR_OVERFLOW:
        result = ERROR_ARITHMETIC_OVERFLOW;
        break;
    case NTD_SYSTEM_ERROR_PERM:
        result = ERROR_CANNOT_MAKE;
        break;
    case NTD_SYSTEM_ERROR_PIPE:
        result = ERROR_BROKEN_PIPE;
        break;
    case NTD_SYSTEM_ERROR_RANGE:
        result = ERROR_META_EXPANSION_TOO_LONG;
        break;
    case NTD_SYSTEM_ERROR_ROFS:
        result = ERROR_WRITE_PROTECT;
        break;
    case NTD_SYSTEM_ERROR_SPIPE:
        result = ERROR_SETMARK_DETECTED;
        break;
    case NTD_SYSTEM_ERROR_SRCH:
        result = ERROR_NOT_FOUND;
        break;
    case NTD_SYSTEM_ERROR_TIME:
        result = ERROR_SEM_TIMEOUT;
        break;
    case NTD_SYSTEM_ERROR_TIMEDOUT:
        result = ERROR_TIMEOUT;
        break;
    case NTD_SYSTEM_ERROR_XDEV:
        result = ERROR_NOT_SAME_DEVICE;
        break;
    case NTD_SYSTEM_ERROR_DESTADDRREQ:
    case NTD_SYSTEM_ERROR_IDRM:
    case NTD_SYSTEM_ERROR_ISCONN:
    case NTD_SYSTEM_ERROR_TXTBSY:
    case NTD_SYSTEM_ERROR_DOM:
    case NTD_SYSTEM_ERROR_PROTO:
    case NTD_SYSTEM_ERROR_PROTONOSUPPORT:
    case NTD_SYSTEM_ERROR_PROTOTYPE:
    case NTD_SYSTEM_ERROR_NOPROTOOPT:
    case NTD_SYSTEM_ERROR_NOSR:
    case NTD_SYSTEM_ERROR_NOSTR:
    case NTD_SYSTEM_ERROR_NOTSOCK:
    default:
        result = ERROR_UNIDENTIFIED_ERROR;
        break;
    }

    return result;
}
